/*
 * Copyright (c) 2005-2014, BearWare.dk
 * 
 * Contact Information:
 *
 * Bjoern D. Rasmussen
 * Skanderborgvej 40 4-2
 * DK-8000 Aarhus C
 * Denmark
 * Email: contact@bearware.dk
 * Phone: +45 20 20 54 59
 * Web: http://www.bearware.dk
 *
 * This source code is part of the TeamTalk 4 SDK owned by
 * BearWare.dk. All copyright statements may not be removed 
 * or altered from any source distribution. If you use this
 * software in a product, an acknowledgment in the product 
 * documentation is required.
 *
 */

#include "channeldlg.h"
#include "mainwindow.h"

#include <QPushButton>
#include <QMessageBox>

extern TTInstance* ttInst;

ChannelDlg::ChannelDlg(ChannelDlgType type, const Channel& chan, QWidget * parent/* = 0*/)
    : QDialog(parent, Qt::Dialog | Qt::WindowTitleHint)
    , m_channel(chan)
{
    ui.setupUi(this);
    setWindowIcon(QIcon(APPICON));

    connect(ui.audiocodecBox, SIGNAL(currentIndexChanged(int)), 
            SLOT(slotAudioCodecChanged(int)));

    connect(ui.spx_qualitySlider, SIGNAL(valueChanged(int)),
            SLOT(slotUpdateSliderLabels()));
    connect(ui.spxvbr_qualitySlider, SIGNAL(valueChanged(int)),
            SLOT(slotUpdateSliderLabels()));
    connect(ui.nameEdit, SIGNAL(textChanged(const QString&)),
            SLOT(slotUpdateChannelPath(const QString&)));
    
    connect(ui.agcBox, SIGNAL(toggled(bool)),
            ui.gainlevelSlider, SLOT(setEnabled(bool)));
    connect(ui.gainlevelSlider, SIGNAL(valueChanged(int)),
            SLOT(slotUpdateSliderLabels()));

    ServerProperties prop;
    ZERO_STRUCT(prop);
    TT_GetServerProperties(ttInst, &prop);

    ui.audiocodecBox->addItem(tr("No Audio"), NO_CODEC);
    ui.audiocodecBox->addItem("Speex", SPEEX_CODEC);
    if(m_channel.codec.nCodec == CELT_0_5_2_OBSOLETE_CODEC)
        ui.audiocodecBox->addItem("CELT 0.5.2 (obsolete)", CELT_0_5_2_OBSOLETE_CODEC);

    if(VersionSameOrLater(_Q(prop.szServerProtocolVersion), "4.3"))
    {
        ui.audiocodecBox->addItem("Speex Variable Bit Rate", SPEEX_VBR_CODEC);
        ui.audiocodecBox->addItem("CELT", CELT_CODEC);
        ui.audiocodecBox->addItem("CELT Variable Bit Rate", CELT_VBR_CODEC);
    }
    ui.celt_channelsBox->addItem(tr("Mono"));
    ui.celt_channelsBox->addItem(tr("Stereo"));
    ui.celtvbr_channelsBox->addItem(tr("Mono"));
    ui.celtvbr_channelsBox->addItem(tr("Stereo"));

    if( (TT_GetMyUserType(ttInst) & USERTYPE_ADMIN) == 0)
    {
        ui.staticchanBox->setEnabled(false);
        ui.diskquotaSpinBox->setEnabled(false);
        ui.maxusersSpinBox->setEnabled(false);
    }

    ui.classroomchanBox->setEnabled(VersionSameOrLater(_Q(prop.szServerProtocolVersion), "4.2"));
    ui.agcBox->setEnabled(VersionSameOrLater(_Q(prop.szServerProtocolVersion), "4.2"));
    ui.gainlevelSlider->setEnabled(VersionSameOrLater(_Q(prop.szServerProtocolVersion), "4.2"));
    ui.denoiseBox->setEnabled(VersionSameOrLater(_Q(prop.szServerProtocolVersion), "4.2"));
    ui.echovideoBox->setEnabled(VersionSameOrLater(_Q(prop.szServerProtocolVersion), "4.4"));
    ui.echodesktopBox->setEnabled(VersionSameOrLater(_Q(prop.szServerProtocolVersion), "4.4"));
    ui.oprecvonlychanBox->setEnabled(VersionSameOrLater(_Q(prop.szServerProtocolVersion), "4.5"));

    switch(type)
    {
    case CHANNEL_CREATE :
        setWindowTitle(tr("Create Channel"));
        //default settings for Speex
        m_channel.codec.nCodec = SPEEX_CODEC;
        m_channel.codec.speex.nBandmode = DEFAULT_SPEEX_BANDMODE;
        m_channel.codec.speex.nQuality = DEFAULT_SPEEX_QUALITY;
        m_channel.codec.speex.nMSecPerPacket = DEFAULT_SPEEX_DELAY;
        m_channel.codec.speex.bUseJitterBuffer = DEFAULT_SPEEX_JITTERBUF;
        m_channel.codec.speex.bStereoPlayback = DEFAULT_SPEEX_SIMSTEREO;
        //default settings for Speex VBR
        ui.spxvbr_khz16Button->setChecked(true);
        ui.spxvbr_qualitySlider->setValue(DEFAULT_SPEEX_VBR_QUALITY);
        ui.spxvbr_maxbpsSpinBox->setValue(DEFAULT_SPEEX_VBR_MAXBITRATE);
        ui.spxvbr_dtxBox->setChecked(true);
        ui.spxvbr_txdelaySpinBox->setValue(DEFAULT_SPEEX_VBR_DELAY);
        //default settings for CELT
        ui.celt_khz32Button->setChecked(true);
        ui.celt_bpsSpinBox->setValue(DEFAULT_CELT_BITRATE);
        ui.celt_channelsBox->setCurrentIndex(0);
        ui.celt_txdelaySpinBox->setValue(DEFAULT_CELT_DELAY);
        //default settings for CELT VBR
        ui.celtvbr_khz32Button->setChecked(true);
        ui.celtvbr_bpsSpinBox->setValue(DEFAULT_CELT_VBR_BITRATE);
        ui.celtvbr_channelsBox->setCurrentIndex(0);
        ui.celtvbr_txdelaySpinBox->setValue(DEFAULT_CELT_VBR_DELAY);

        //set default channel options
        m_channel.nMaxUsers = prop.nMaxUsers;
        m_channel.uChannelType = CHANNEL_DEFAULT;
        //set AGC
        m_channel.audiocfg.bEnableAGC = DEFAULT_AGC_ENABLE;
        m_channel.audiocfg.nGainLevel = DEFAULT_AGC_GAINLEVEL;
        m_channel.audiocfg.nMaxIncDBSec = DEFAULT_AGC_INC_MAXDB;
        m_channel.audiocfg.nMaxDecDBSec = DEFAULT_AGC_DEC_MAXDB;
        m_channel.audiocfg.nMaxGainDB = DEFAULT_AGC_GAINMAXDB;
        m_channel.audiocfg.bEnableDenoise = DEFAULT_DENOISE_ENABLE;
        m_channel.audiocfg.nMaxNoiseSuppressDB = DEFAULT_DENOISE_SUPPRESS;
        break;
    case CHANNEL_UPDATE :
        {
            setWindowTitle(tr("Update Channel"));
            ui.nameEdit->setReadOnly(true);
            int count = 0;
            TT_GetChannelUsers(ttInst, chan.nChannelID, NULL, &count);
            if(count>0)
            {
                ui.audiocodecBox->setEnabled(false);
                ui.spx_khz8Button->setEnabled(false);
                ui.spx_khz16Button->setEnabled(false);
                ui.spx_khz32Button->setEnabled(false);
                ui.spx_qualitySlider->setEnabled(false);
                ui.spx_txdelaySpinBox->setEnabled(false);

                ui.spxvbr_khz8Button->setEnabled(false);
                ui.spxvbr_khz16Button->setEnabled(false);
                ui.spxvbr_khz32Button->setEnabled(false);
                ui.spxvbr_qualitySlider->setEnabled(false);
                ui.spxvbr_maxbpsSpinBox->setEnabled(false);
                ui.spxvbr_dtxBox->setEnabled(false);
                ui.spxvbr_txdelaySpinBox->setEnabled(false);

                ui.celt_khz32Button->setEnabled(false);
                ui.celt_khz441Button->setEnabled(false);
                ui.celt_channelsBox->setEnabled(false);
                ui.celt_bpsSpinBox->setEnabled(false);
                ui.celt_txdelaySpinBox->setEnabled(false);

                ui.celtvbr_khz32Button->setEnabled(false);
                ui.celtvbr_khz441Button->setEnabled(false);
                ui.celtvbr_channelsBox->setEnabled(false);
                ui.celtvbr_bpsSpinBox->setEnabled(false);
                ui.celtvbr_txdelaySpinBox->setEnabled(false);
            }
            break;
        }
    case CHANNEL_READONLY :
        setWindowTitle(tr("View Channel Information"));
        ui.nameEdit->setReadOnly(true);
        ui.topicTextEdit->setReadOnly(true);
        ui.chanpasswdEdit->setReadOnly(true);
        ui.oppasswdEdit->setReadOnly(true);
        ui.maxusersSpinBox->setReadOnly(true);
        ui.diskquotaSpinBox->setReadOnly(true);
        ui.staticchanBox->setEnabled(false);
        ui.echochanBox->setEnabled(false);
        ui.echovideoBox->setEnabled(false);
        ui.echodesktopBox->setEnabled(false);
        ui.singletxchanBox->setEnabled(false);
        ui.classroomchanBox->setEnabled(false);
        ui.oprecvonlychanBox->setEnabled(false);

        ui.audiocodecBox->setEnabled(false);

        ui.spx_khz8Button->setEnabled(false);
        ui.spx_khz16Button->setEnabled(false);
        ui.spx_khz32Button->setEnabled(false);
        ui.spx_qualitySlider->setEnabled(false);
        ui.spx_txdelaySpinBox->setEnabled(false);

        ui.spxvbr_khz8Button->setEnabled(false);
        ui.spxvbr_khz16Button->setEnabled(false);
        ui.spxvbr_khz32Button->setEnabled(false);
        ui.spxvbr_qualitySlider->setEnabled(false);
        ui.spxvbr_maxbpsSpinBox->setEnabled(false);
        ui.spxvbr_dtxBox->setEnabled(false);
        ui.spxvbr_txdelaySpinBox->setEnabled(false);

        ui.celt_khz32Button->setEnabled(false);
        ui.celt_khz441Button->setEnabled(false);
        ui.celt_channelsBox->setEnabled(false);
        ui.celt_bpsSpinBox->setEnabled(false);
        ui.celt_txdelaySpinBox->setEnabled(false);

        ui.celtvbr_khz32Button->setEnabled(false);
        ui.celtvbr_khz441Button->setEnabled(false);
        ui.celtvbr_channelsBox->setEnabled(false);
        ui.celtvbr_bpsSpinBox->setEnabled(false);
        ui.celtvbr_txdelaySpinBox->setEnabled(false);

        ui.agcBox->setEnabled(false);
        ui.gainlevelSlider->setEnabled(false);
        ui.denoiseBox->setEnabled(false);
        ui.buttonBox->setStandardButtons(QDialogButtonBox::Close);
        break;
    }

    ui.nameEdit->setText(_Q(m_channel.szName));
    ui.topicTextEdit->setPlainText(_Q(m_channel.szTopic));
    ui.chanpasswdEdit->setText(_Q(m_channel.szPassword));
    ui.oppasswdEdit->setText(_Q(m_channel.szOpPassword));
    ui.maxusersSpinBox->setValue(m_channel.nMaxUsers);
    ui.diskquotaSpinBox->setValue(m_channel.nDiskQuota/1024);
    ui.staticchanBox->setChecked(m_channel.uChannelType & CHANNEL_STATIC);
    ui.echochanBox->setChecked(m_channel.uChannelType & CHANNEL_ECHO);
    ui.echovideoBox->setChecked(m_channel.uChannelType & CHANNEL_ECHO_VIDEO);
    ui.echodesktopBox->setChecked(m_channel.uChannelType & CHANNEL_ECHO_DESKTOP);
    ui.singletxchanBox->setChecked(m_channel.uChannelType & CHANNEL_SOLO_TRANSMIT);
    ui.classroomchanBox->setChecked(m_channel.uChannelType & CHANNEL_CLASSROOM);
    ui.oprecvonlychanBox->setChecked(m_channel.uChannelType & CHANNEL_OPERATOR_RECVONLY);
    //codec
    switch(m_channel.codec.nCodec)
    {
    case SPEEX_CODEC :
        ui.spx_khz8Button->setChecked(m_channel.codec.speex.nBandmode == 0);
        ui.spx_khz16Button->setChecked(m_channel.codec.speex.nBandmode == 1);
        ui.spx_khz32Button->setChecked(m_channel.codec.speex.nBandmode == 2);
        ui.spx_qualitySlider->setValue(m_channel.codec.speex.nQuality);
        ui.spx_txdelaySpinBox->setValue(m_channel.codec.speex.nMSecPerPacket);
        break;
    case SPEEX_VBR_CODEC :
        ui.spxvbr_khz8Button->setChecked(m_channel.codec.speex_vbr.nBandmode == 0);
        ui.spxvbr_khz16Button->setChecked(m_channel.codec.speex_vbr.nBandmode == 1);
        ui.spxvbr_khz32Button->setChecked(m_channel.codec.speex_vbr.nBandmode == 2);
        ui.spxvbr_qualitySlider->setValue(m_channel.codec.speex_vbr.nQualityVBR);
        ui.spxvbr_maxbpsSpinBox->setValue(m_channel.codec.speex_vbr.nMaxBitRate);
        ui.spxvbr_dtxBox->setChecked(m_channel.codec.speex_vbr.bDTX);
        ui.spxvbr_txdelaySpinBox->setValue(m_channel.codec.speex_vbr.nMSecPerPacket);
        break;
    case CELT_CODEC :
        ui.celt_khz32Button->setChecked(m_channel.codec.celt.nSampleRate == 32000);
        ui.celt_khz441Button->setChecked(m_channel.codec.celt.nSampleRate == 44100);
        ui.celt_channelsBox->setCurrentIndex(m_channel.codec.celt.nChannels-1);
        ui.celt_bpsSpinBox->setValue(m_channel.codec.celt.nBitRate);
        ui.celt_txdelaySpinBox->setValue(m_channel.codec.celt.nMSecPerPacket);
        break;
    case CELT_VBR_CODEC :
        ui.celtvbr_khz32Button->setChecked(m_channel.codec.celt_vbr.nSampleRate == 32000);
        ui.celtvbr_khz441Button->setChecked(m_channel.codec.celt_vbr.nSampleRate == 44100);
        ui.celtvbr_channelsBox->setCurrentIndex(m_channel.codec.celt_vbr.nChannels-1);
        ui.celtvbr_bpsSpinBox->setValue(m_channel.codec.celt_vbr.nBitRate);
        ui.celtvbr_txdelaySpinBox->setValue(m_channel.codec.celt_vbr.nMSecPerPacket);
        break;
    default :
        break;
    }
    ui.audiocodecBox->setCurrentIndex(ui.audiocodecBox->findData(m_channel.codec.nCodec));
    //AGC
    ui.agcBox->setChecked(m_channel.audiocfg.bEnableAGC);
    ui.gainlevelSlider->setEnabled(m_channel.audiocfg.bEnableAGC);
    ui.gainlevelSlider->setValue(m_channel.audiocfg.nGainLevel / 1000);
    ui.denoiseBox->setChecked(m_channel.audiocfg.bEnableDenoise);

    slotUpdateSliderLabels();
    slotUpdateChannelPath(_Q(m_channel.szName));

    //connect here so there's no event triggered when the channels are set
    connect(ui.celt_channelsBox, SIGNAL(currentIndexChanged(int)),
            SLOT(slotAudioChannelChanged(int)));
    connect(ui.celtvbr_channelsBox, SIGNAL(currentIndexChanged(int)),
            SLOT(slotAudioChannelChanged(int)));
}

Channel ChannelDlg::GetChannel() const
{
    Channel newchannel = m_channel;
    COPY_TTSTR(newchannel.szName, ui.nameEdit->text());
    COPY_TTSTR(newchannel.szTopic, ui.topicTextEdit->toPlainText());
    COPY_TTSTR(newchannel.szPassword, ui.chanpasswdEdit->text());
    COPY_TTSTR(newchannel.szOpPassword, ui.oppasswdEdit->text());
    newchannel.nMaxUsers = ui.maxusersSpinBox->value();
    newchannel.nDiskQuota = (quint64)ui.diskquotaSpinBox->value()*1024;

    if(ui.staticchanBox->isChecked())
        newchannel.uChannelType |= CHANNEL_STATIC;
    else
        newchannel.uChannelType &= ~CHANNEL_STATIC;

    if(ui.echochanBox->isChecked())
        newchannel.uChannelType |= CHANNEL_ECHO;
    else
        newchannel.uChannelType &= ~CHANNEL_ECHO;

    if(ui.singletxchanBox->isChecked())
        newchannel.uChannelType |= CHANNEL_SOLO_TRANSMIT;
    else
        newchannel.uChannelType &= ~CHANNEL_SOLO_TRANSMIT;

    if(ui.classroomchanBox->isChecked())
        newchannel.uChannelType |= CHANNEL_CLASSROOM;
    else
        newchannel.uChannelType &= ~CHANNEL_CLASSROOM;

    if(ui.echovideoBox->isChecked())
        newchannel.uChannelType |= CHANNEL_ECHO_VIDEO;
    else
        newchannel.uChannelType &= ~CHANNEL_ECHO_VIDEO;

    if(ui.echodesktopBox->isChecked())
        newchannel.uChannelType |= CHANNEL_ECHO_DESKTOP;
    else
        newchannel.uChannelType &= ~CHANNEL_ECHO_DESKTOP;

    if(ui.oprecvonlychanBox->isChecked())
        newchannel.uChannelType |= CHANNEL_OPERATOR_RECVONLY;
    else
        newchannel.uChannelType &= ~CHANNEL_OPERATOR_RECVONLY;

    newchannel.codec.nCodec = (Codec)ui.audiocodecBox->itemData(ui.audiocodecBox->currentIndex()).toInt();
    switch(newchannel.codec.nCodec)
    {
    case SPEEX_CODEC :
        if(ui.spx_khz8Button->isChecked())
            newchannel.codec.speex.nBandmode = 0;
        if(ui.spx_khz16Button->isChecked())
            newchannel.codec.speex.nBandmode = 1;
        if(ui.spx_khz32Button->isChecked())
            newchannel.codec.speex.nBandmode = 2;
        newchannel.codec.speex.nQuality = ui.spx_qualitySlider->value();
        newchannel.codec.speex.nMSecPerPacket = ui.spx_txdelaySpinBox->value();
        newchannel.codec.speex.bUseJitterBuffer = DEFAULT_SPEEX_JITTERBUF;
        newchannel.codec.speex.bStereoPlayback = DEFAULT_SPEEX_SIMSTEREO;
        break;
    case SPEEX_VBR_CODEC :
        if(ui.spxvbr_khz8Button->isChecked())
            newchannel.codec.speex_vbr.nBandmode = 0;
        if(ui.spxvbr_khz16Button->isChecked())
            newchannel.codec.speex_vbr.nBandmode = 1;
        if(ui.spxvbr_khz32Button->isChecked())
            newchannel.codec.speex_vbr.nBandmode = 2;
        newchannel.codec.speex_vbr.nQualityVBR = ui.spxvbr_qualitySlider->value();
        newchannel.codec.speex_vbr.nBitRate = DEFAULT_SPEEX_VBR_BITRATE;
        newchannel.codec.speex_vbr.nMaxBitRate = ui.spxvbr_maxbpsSpinBox->value();
        newchannel.codec.speex_vbr.bDTX = ui.spxvbr_dtxBox->isChecked();
        newchannel.codec.speex_vbr.nMSecPerPacket = ui.spxvbr_txdelaySpinBox->value();
        newchannel.codec.speex_vbr.bUseJitterBuffer = DEFAULT_SPEEX_VBR_JITTERBUF;
        newchannel.codec.speex_vbr.bStereoPlayback = DEFAULT_SPEEX_VBR_SIMSTEREO;
        break;
    case CELT_CODEC :
        if(ui.celt_khz32Button->isChecked())
            newchannel.codec.celt.nSampleRate = 32000;
        if(ui.celt_khz441Button->isChecked())
            newchannel.codec.celt.nSampleRate = 44100;

        newchannel.codec.celt.nBitRate = ui.celt_bpsSpinBox->value();
        newchannel.codec.celt.nMSecPerPacket = ui.celt_txdelaySpinBox->value();
        newchannel.codec.celt.nChannels = ui.celt_channelsBox->currentIndex()+1;
        break;
    case CELT_VBR_CODEC :
        if(ui.celtvbr_khz32Button->isChecked())
            newchannel.codec.celt_vbr.nSampleRate = 32000;
        if(ui.celtvbr_khz441Button->isChecked())
            newchannel.codec.celt_vbr.nSampleRate = 44100;

        newchannel.codec.celt_vbr.nBitRate = ui.celtvbr_bpsSpinBox->value();
        newchannel.codec.celt_vbr.nMSecPerPacket = ui.celtvbr_txdelaySpinBox->value();
        newchannel.codec.celt_vbr.nChannels = ui.celtvbr_channelsBox->currentIndex()+1;
        break;
    default :
        break;
    }

    newchannel.audiocfg.bEnableAGC = ui.agcBox->isChecked();
    newchannel.audiocfg.bEnableDenoise = ui.denoiseBox->isChecked();
    
    //set default values since they may otherwise be 0
    if(newchannel.audiocfg.bEnableAGC)
    {
        newchannel.audiocfg.nGainLevel = ui.gainlevelSlider->value()*1000;
        newchannel.audiocfg.nMaxIncDBSec = DEFAULT_AGC_INC_MAXDB;
        newchannel.audiocfg.nMaxDecDBSec = DEFAULT_AGC_DEC_MAXDB;
        newchannel.audiocfg.nMaxGainDB = DEFAULT_AGC_GAINMAXDB;
    }
    if(newchannel.audiocfg.bEnableDenoise)
        newchannel.audiocfg.nMaxNoiseSuppressDB = DEFAULT_DENOISE_SUPPRESS;
        

    return newchannel;
}

void ChannelDlg::slotAudioCodecChanged(int index)
{
    switch(ui.audiocodecBox->itemData(index).toInt())
    {
    case SPEEX_CODEC :
        ui.stackedWidget->setCurrentIndex(1);
        break;
    case CELT_0_5_2_OBSOLETE_CODEC :
        ui.stackedWidget->setCurrentIndex(2);
        break;
    case SPEEX_VBR_CODEC :
        ui.stackedWidget->setCurrentIndex(3);
        break;
    case CELT_CODEC :
        ui.stackedWidget->setCurrentIndex(4);
        break;
    case CELT_VBR_CODEC :
        ui.stackedWidget->setCurrentIndex(5);
        break;
    case NO_CODEC :
    default :
        ui.stackedWidget->setCurrentIndex(0);
        break;
    }
}

void ChannelDlg::slotUpdateSliderLabels()
{
    ui.spxQualityLabel->setText(QString("%1/10").arg(ui.spx_qualitySlider->value()));
    ui.spxvbr_QualityLabel->setText(QString("%1/10").arg(ui.spxvbr_qualitySlider->value()));
    ui.gainlevelLabel->setText(QString::number(ui.gainlevelSlider->value()*1000));   
}

void ChannelDlg::slotUpdateChannelPath(const QString& str)
{
    TTCHAR path[TT_STRLEN] = {0};
    if(TT_GetChannelPath(ttInst, m_channel.nParentID, path))
        ui.chanpathLabel->setText(_Q(path) + str);
    else
        ui.chanpathLabel->setText(QString());

    QPushButton* btn = ui.buttonBox->button(QDialogButtonBox::Ok);
    if(btn)
        btn->setEnabled(str.length()>0 || 
            TT_GetRootChannelID(ttInst) == m_channel.nChannelID);
}

void ChannelDlg::slotAudioChannelChanged(int aud_channels)
{
    if(aud_channels+1 == 2)
        QMessageBox::information(this, tr("Audio channels"),
        tr("Note that Echo Cancellation, Denoising and Audio Configuration currently isn't supported in stereo mode."));
}
