/*
 * Copyright (c) 2005-2014, BearWare.dk
 * 
 * Contact Information:
 *
 * Bjoern D. Rasmussen
 * Skanderborgvej 40 4-2
 * DK-8000 Aarhus C
 * Denmark
 * Email: contact@bearware.dk
 * Phone: +45 20 20 54 59
 * Web: http://www.bearware.dk
 *
 * This source code is part of the TeamTalk 4 SDK owned by
 * BearWare.dk. All copyright statements may not be removed 
 * or altered from any source distribution. If you use this
 * software in a product, an acknowledgment in the product 
 * documentation is required.
 *
 */

#include "mainwindow.h"
#include "serverlistdlg.h"
#include "serverpropertiesdlg.h"
#include "preferencesdlg.h"
#include "connectdlg.h"
#include "channeldlg.h"
#include "userinfodlg.h"
#include "bannedusersdlg.h"
#include "useraccountsdlg.h"
#include "keycompdlg.h"
#include "textmessagedlg.h"
#include "filetransferdlg.h"
#include "uservolumedlg.h"
#include "changestatusdlg.h"
#include "aboutdlg.h"
#include "audiostoragedlg.h"
#include "channelstree.h"
#include "desktopsharedlg.h"
#include "streamaudiofiledlg.h"
#include "streammediafiledlg.h"

#include <QMessageBox>
#include <QInputDialog>
#include <QFileDialog>
#include <QTimer>
#include <QSettings>
#include <QDesktopWidget>
#include <QFile>
#include <QTranslator>
#include <QUrl>
#include <QTabWidget>
#include <QHeaderView>
#include <QRegExp>
#include <QDesktopServices>
#include <QProcess>
#include <QNetworkRequest>
#include <QNetworkReply>

#ifdef Q_OS_LINUX //For hotkeys on X11
#include <QX11Info>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable:4800)
#endif

extern TTInstance* ttInst;

QSettings* ttSettings = NULL;
QTranslator* ttTranslator = NULL;

//strip ampersand from menutext
#define MENUTEXT(text) text.replace("&", "")

MainWindow::MainWindow()
: m_sysicon(NULL)
, m_sysmenu(NULL)
, m_current_cmdid(0)
, m_audiostorage_mode(AUDIOSTORAGE_NONE)
, m_idled_out(false)
, m_statusmode(STATUSMODE_AVAILABLE)
, m_clientstats()
, m_last_channel()
, m_srvprop()
, m_mychannel()
, m_http_manager(NULL)
, m_onlineusersdlg(NULL)
, m_useraccountsdlg(NULL)
, m_bannedusersdlg(NULL)
, m_serverstatsdlg(NULL)
, m_desktopsession_id(0)
, m_prev_desktopsession_id(0)
, m_desktopsession_total(0)
, m_desktopsession_remain(0)
, m_desktopsend_on_completion(false)
#if defined(Q_OS_WIN32)
, m_hShareWnd(NULL)
#elif defined(Q_OS_DARWIN)
, m_nCGShareWnd(kCGNullWindowID)
#elif defined(Q_OS_LINUX)
, m_display(NULL)
, m_nWindowShareWnd(0)
#endif
{
    //Ensure the correct version of the DLL is loaded
    if(QString(TEAMTALK4_VERSION) != _Q(TT_GetVersion()))
        QMessageBox::warning(0, ("DLL load error"),
                             QString("This %3 executable is built for DLL "
                                      "version %1 but the loaded DLL reports "
                                      "it's version %2. Loading an incorrent "
                                      "DLL for %3 may cause problems and crash "
                                      "the application. Please reinstall to "
                                      "solve this problem.")
                                     .arg(TEAMTALK4_VERSION).
                                     arg(_Q(TT_GetVersion())).
                                     arg(APPNAME_SHORT));

    QApplication::setOrganizationName(COMPANYNAME);
    QApplication::setApplicationName(APPNAME_SHORT);

    //figure out where to load settings from
    QString inifile = QString(APPINIFILE);
    QString inipath = QApplication::applicationDirPath();
    inipath += "/" + inifile;
    if(QFile::exists(inipath)) //first try same dir as executable
        ttSettings = new QSettings(inipath, QSettings::IniFormat, this);
    else
    {
        ttSettings = new QSettings(QSettings::IniFormat, 
                                   QSettings::UserScope,
                                   QApplication::organizationName(),
                                   QApplication::applicationName(), this);
    }

    ui.setupUi(this);

    setWindowIcon(QIcon(APPICON));
    updateWindowTitle();

    m_filesmodel = new FilesModel(this);
    ui.filesView->setModel(m_filesmodel);
    QItemSelectionModel* selmodel = ui.filesView->selectionModel();
    int gainMax = 4000; /* real max: SOUND_LEVEL_MAX but it's too much */
    ui.volumeSlider->setRange(SOUND_VOLUME_MIN, SOUND_VOLUME_MAX*(gainMax/SOUND_GAIN_DEFAULT));
    ui.micSlider->setRange(SOUND_GAIN_MIN, gainMax);
    m_pinglabel = new QLabel(ui.statusbar);
    m_pinglabel->setMinimumWidth(75);
    m_pinglabel->setAlignment(Qt::AlignHCenter);
    m_dtxlabel = new QLabel(ui.statusbar);
    QPixmap dtxImg(QString::fromUtf8(":/images/images/desktoptx.png"));
    m_dtxlabel->setPixmap(dtxImg.scaled(16, 16));
    m_dtxprogress = new QProgressBar(ui.statusbar);
    m_dtxprogress->setMaximumWidth(50);
    m_dtxprogress->setMaximumHeight(16);
    m_dtxprogress->setTextVisible(false);
    m_dtxlabel->hide();
    m_dtxprogress->hide();
    ui.statusbar->addPermanentWidget(m_dtxlabel);
    ui.statusbar->addPermanentWidget(m_dtxprogress);
    ui.statusbar->addPermanentWidget(m_pinglabel);

    ui.actionStreamAudioFileToChannel->setVisible(false);
    //ui.actionStreamMediaFileToChannel->setVisible(false);

    connect(ui.sendButton, SIGNAL(clicked()),
            SLOT(slotSendChannelMessage()));
    connect(ui.msgEdit, SIGNAL(returnPressed()),
            SLOT(slotSendChannelMessage()));
    connect(ui.desktopsendButton, SIGNAL(clicked()),
            SLOT(slotSendChannelMessage()));
    connect(ui.desktopmsgEdit, SIGNAL(returnPressed()),
            SLOT(slotSendChannelMessage()));
    connect(ui.micSlider, SIGNAL(valueChanged(int)), 
            SLOT(slotMicrophoneGainChanged(int)));
    connect(ui.volumeSlider, SIGNAL(valueChanged(int)),
            SLOT(slotMasterVolumeChanged(int)));
    connect(ui.voiceactSlider, SIGNAL(valueChanged(int)),
            SLOT(slotVoiceActivationLevelChanged(int)));

    /* ui.channelsWidget */
    connect(ui.channelsWidget, SIGNAL(itemSelectionChanged()),
            SLOT(slotTreeSelectionChanged()));
    connect(ui.channelsWidget, SIGNAL(customContextMenuRequested(const QPoint&)),
            SLOT(slotTreeContextMenu(const QPoint&)));
    connect(ui.channelsWidget, SIGNAL(userDoubleClicked(int)), 
            SLOT(slotUserDoubleClicked(int)));
    connect(ui.channelsWidget, SIGNAL(channelDoubleClicked(int)),
            SLOT(slotChannelDoubleClicked(int)));
    connect(ui.channelsWidget, SIGNAL(fileDropped(const QString&)),
            SLOT(slotLoadTTFile(const QString&)));
    connect(ui.channelsWidget,
            SIGNAL(classroomChanged(int, const QSet<int>&, const QSet<int>&, 
                                    const QSet<int>&)),
            SLOT(slotClassroomChanged(int, const QSet<int>&, const QSet<int>&,
                                      const QSet<int>&)));
    /* Video-tab (video-grid) */
    connect(this, SIGNAL(newVideoFrame(int,int)), ui.videogridWidget, 
            SLOT(slotNewVideoFrame(int,int)));
    connect(ui.videogridWidget, SIGNAL(userVideoEnded(int)),
            SLOT(slotRemoveUserVideo(int)));
    connect(ui.videogridWidget, SIGNAL(videoCountChanged(int)),
            SLOT(slotUpdateVideoCount(int)));
    connect(ui.videogridWidget, SIGNAL(videoCountChanged(int)),
            SLOT(slotUpdateVideoTabUI()));
    connect(ui.videogridWidget, SIGNAL(userVideoSelected(bool)),
            SLOT(slotUpdateVideoTabUI()));            
    /* Desktop-tab (desktop-grid) */
    connect(this, SIGNAL(newDesktopWindow(int,int)), ui.desktopgridWidget,
            SIGNAL(userDesktopWindowUpdate(int,int)));
    connect(this, SIGNAL(userDesktopCursor(int,int)), ui.desktopgridWidget,
            SIGNAL(userDesktopCursorUpdate(int,int)));
    connect(this, SIGNAL(userUpdate(int,int)), ui.desktopgridWidget,
            SIGNAL(userUpdated(int,int)));
    connect(this, SIGNAL(userLeft(int,int)), ui.desktopgridWidget,
            SLOT(slotRemoveUser(int,int)));
    connect(ui.desktopgridWidget, SIGNAL(userDesktopWindowEnded(int)),
            SLOT(slotRemoveUserDesktop(int)));
    connect(ui.desktopgridWidget, SIGNAL(desktopCountChanged(int)),
            SLOT(slotUpdateDesktopCount(int)));
    connect(ui.desktopgridWidget, SIGNAL(desktopCountChanged(int)),
            SLOT(slotUpdateDesktopTabUI()));
    connect(ui.desktopgridWidget, SIGNAL(userDesktopSelected(bool)),
            SLOT(slotUpdateDesktopTabUI()));            
    /* Files-tab */
    connect(ui.uploadButton, SIGNAL(clicked()), SLOT(slotChannelsUploadFile()));
    connect(ui.downloadButton, SIGNAL(clicked()), SLOT(slotChannelsDownloadFile()));
    connect(ui.deleteButton, SIGNAL(clicked()), SLOT(slotChannelsDeleteFile()));
    connect(selmodel, SIGNAL(selectionChanged(const QItemSelection&, 
                                              const QItemSelection&)),
            ui.filesView, SLOT(slotNewSelection(const QItemSelection&)));
    connect(ui.filesView, SIGNAL(filesSelected(bool)), ui.actionDeleteFile, 
            SLOT(setEnabled(bool)));
    connect(ui.filesView, SIGNAL(filesSelected(bool)), ui.deleteButton, 
            SLOT(setEnabled(bool)));
    connect(ui.filesView, SIGNAL(uploadFiles(const QStringList&)),
            SLOT(slotUploadFiles(const QStringList&)));
    /* Video-tab buttons */
    connect(ui.initVideoButton, SIGNAL(clicked(bool)), 
            ui.actionEnableVideoTransmission, SIGNAL(triggered(bool)));
    connect(ui.addVideoButton, SIGNAL(clicked()), SLOT(slotAddUserVideo()));
    connect(ui.removeVideoButton, SIGNAL(clicked()), SLOT(slotRemoveUserVideo()));
    connect(ui.detachVideoButton, SIGNAL(clicked()), SLOT(slotDetachUserVideo()));
    connect(ui.videogridWidget, SIGNAL(userVideoSelected(bool)),
            ui.detachVideoButton, SLOT(setEnabled(bool)));
    connect(this, SIGNAL(preferencesModified()),
            ui.videogridWidget, SIGNAL(preferencesModified()));
    /* Desktop-tab buttons */
    connect(ui.detachDesktopButton, SIGNAL(clicked()),
            SLOT(slotDetachUserDesktop()));
    connect(ui.addDesktopButton, SIGNAL(clicked()),
            SLOT(slotAddUserDesktop()));
    connect(ui.removeDesktopButton, SIGNAL(clicked()),
            SLOT(slotRemoveUserDesktop()));
    connect(ui.desktopaccessButton, SIGNAL(clicked(bool)),
            SLOT(slotAccessUserDesktop(bool)));
    connect(ui.desktopgridWidget, SIGNAL(userDesktopSelected(bool)),
            ui.detachDesktopButton, SLOT(setEnabled(bool)));
    connect(ui.desktopgridWidget, SIGNAL(userDesktopSelected(bool)),
            ui.desktopaccessButton, SLOT(setEnabled(bool)));

    /* Begin - File menu */
    connect(ui.actionNewClient, SIGNAL(triggered(bool)),
            SLOT(slotClientNewInstance(bool)));
    connect(ui.actionConnect, SIGNAL(triggered(bool)),
            SLOT(slotClientConnect(bool)));
    connect(ui.actionPreferences, SIGNAL(triggered(bool)),
            SLOT(slotClientPreferences(bool)));
    connect(ui.actionExit, SIGNAL(triggered(bool)),
            SLOT(slotClientExit(bool)));
    /* End - File menu */

    /* Begin - Me menu */
    connect(ui.actionChangeNickname, SIGNAL(triggered(bool)),
            SLOT(slotMeChangeNickname(bool)));
    connect(ui.actionChangeStatus, SIGNAL(triggered(bool)),
            SLOT(slotMeChangeStatus(bool)));
    connect(ui.actionEnablePushToTalk, SIGNAL(triggered(bool)),
            SLOT(slotMeEnablePushToTalk(bool)));
    connect(ui.actionEnableVoiceActivation, SIGNAL(triggered(bool)),
            SLOT(slotMeEnableVoiceActivation(bool)));
    connect(ui.actionEnableVideoTransmission, SIGNAL(triggered(bool)),
            SLOT(slotMeEnableVideoTransmission(bool)));
    connect(ui.actionEnableDesktopSharing, SIGNAL(triggered(bool)),
            SLOT(slotMeEnableDesktopSharing(bool)));
    /* End - Me menu */

    /* Begin - Users menu */
    connect(ui.actionViewUserInformation, SIGNAL(triggered(bool)),
            SLOT(slotUsersViewUserInformation(bool)));
    connect(ui.actionMessages, SIGNAL(triggered(bool)),
            SLOT(slotUsersMessages(bool)));
    connect(ui.actionMute, SIGNAL(triggered(bool)),
            SLOT(slotUsersMute(bool)));
    connect(ui.actionVolume, SIGNAL(triggered(bool)),
            SLOT(slotUsersVolume(bool)));
    connect(ui.actionOp, SIGNAL(triggered(bool)),
            SLOT(slotUsersOp(bool)));
    connect(ui.actionKickBan, SIGNAL(triggered(bool)),
            SLOT(slotUsersKickBan(bool)));
    connect(ui.actionKick, SIGNAL(triggered(bool)),
            SLOT(slotUsersKick(bool)));
    connect(ui.actionStreamAudioFileToUser, SIGNAL(triggered(bool)),
            SLOT(slotUsersStreamAudioFileToUser(bool)));
    connect(ui.actionMuteAll, SIGNAL(triggered(bool)),
            SLOT(slotUsersMuteAll(bool)));
    connect(ui.actionSpecifyAudioFolder, SIGNAL(triggered(bool)),
            SLOT(slotUsersStoreAudioToDisk(bool)));
    //Desktop access
    connect(ui.actionDesktopAccessAllow, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsDesktopInput(bool)));
    //subscriptions
    connect(ui.actionUserMessages, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsUserMsg(bool)));
    connect(ui.actionChannelMessages, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsChannelMsg(bool)));
    connect(ui.actionBroadcastMessages, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsBCastMsg(bool)));
    connect(ui.actionAudio, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsAudio(bool)));
    connect(ui.actionVideo, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsVideo(bool)));
    connect(ui.actionDesktop, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsDesktop(bool)));
    connect(ui.actionDesktopInput, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsDesktopInput(bool)));
    connect(ui.actionInterceptUserMessages, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsInterceptUserMsg(bool)));
    connect(ui.actionInterceptChannelMessages, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsInterceptChannelMsg(bool)));
    connect(ui.actionInterceptAudio, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsInterceptAudio(bool)));
    connect(ui.actionInterceptVideo, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsInterceptVideo(bool)));
    connect(ui.actionInterceptDesktop, SIGNAL(triggered(bool)),
            SLOT(slotUsersSubscriptionsInterceptDesktop(bool)));

    //advanced
    connect(ui.actionIncreaseVolume, SIGNAL(triggered()),
            SLOT(slotUsersAdvancedIncVolume()));
    connect(ui.actionLowerVolume, SIGNAL(triggered()),
            SLOT(slotUsersAdvancedDecVolume()));
    connect(ui.actionStoreForMove, SIGNAL(triggered()),
            SLOT(slotUsersAdvancedStoreForMove()));
    connect(ui.actionMoveUser, SIGNAL(triggered()),
            SLOT(slotUsersAdvancedMoveUsers()));
    connect(ui.actionAllowVoiceTransmission, SIGNAL(triggered(bool)),
            SLOT(slotUsersAdvancedVoiceAllowed(bool)));
    connect(ui.actionAllowVideoTransmission, SIGNAL(triggered(bool)),
            SLOT(slotUsersAdvancedVideoAllowed(bool)));
    connect(ui.actionAllowDesktopTransmission, SIGNAL(triggered(bool)),
            SLOT(slotUsersAdvancedDesktopAllowed(bool)));
    /* End - Users menu */

    /* Begin - Channels menu */
    connect(ui.actionCreateChannel, SIGNAL(triggered(bool)),
            SLOT(slotChannelsCreateChannel(bool)));
    connect(ui.actionUpdateChannel, SIGNAL(triggered(bool)),
            SLOT(slotChannelsUpdateChannel(bool)));
    connect(ui.actionDeleteChannel, SIGNAL(triggered(bool)),
            SLOT(slotChannelsDeleteChannel(bool)));
    connect(ui.actionJoinChannel, SIGNAL(triggered(bool)),
            SLOT(slotChannelsJoinChannel(bool)));
    connect(ui.actionViewChannelInfo, SIGNAL(triggered(bool)),
            SLOT(slotChannelsViewChannelInfo(bool)));
    connect(ui.actionStreamAudioFileToChannel, SIGNAL(triggered(bool)),
            SLOT(slotChannelsStreamAudioFile(bool)));

    connect(ui.actionStreamMediaFileToChannel, SIGNAL(triggered(bool)),
            SLOT(slotChannelsStreamMediaFile(bool)));
    connect(ui.actionUploadFile, SIGNAL(triggered(bool)),
            SLOT(slotChannelsUploadFile(bool)));
    connect(ui.actionDownloadFile, SIGNAL(triggered(bool)),
            SLOT(slotChannelsDownloadFile(bool)));
    connect(ui.actionDeleteFile, SIGNAL(triggered(bool)),
            SLOT(slotChannelsDeleteFile(bool)));
    /* End - Channels menu */

    /* Begin - Server menu */
    connect(ui.actionUserAccounts, SIGNAL(triggered(bool)),
            SLOT(slotServerUserAccounts(bool)));
    connect(ui.actionBannedUsers, SIGNAL(triggered(bool)),
            SLOT(slotServerBannedUsers(bool)));
    connect(ui.actionOnlineUsers, SIGNAL(triggered(bool)),
            SLOT(slotServerOnlineUsers(bool)));
    connect(ui.actionBroadcastMessage, SIGNAL(triggered(bool)),
            SLOT(slotServerBroadcastMessage(bool)));
    connect(ui.actionServerProperties, SIGNAL(triggered(bool)),
            SLOT(slotServerServerProperties(bool)));
    connect(ui.actionSaveConfiguration, SIGNAL(triggered(bool)),
            SLOT(slotServerSaveConfiguration(bool)));
    connect(ui.actionServerStatistics, SIGNAL(triggered(bool)),
            SLOT(slotServerServerStatistics(bool)));
    /* End - Server menu */

    /* Begin - Help menu */
    connect(ui.actionManual, SIGNAL(triggered(bool)),
            SLOT(slotHelpManual(bool)));
    connect(ui.actionVisitBearWare, SIGNAL(triggered(bool)),
            SLOT(slotHelpVisitBearWare(bool)));
    connect(ui.actionAbout, SIGNAL(triggered(bool)),
            SLOT(slotHelpAbout(bool)));
    /* End - Help menu */

    /* Begin - Extra toolbar buttons */
    connect(ui.actionEnableQuestionMode, SIGNAL(triggered(bool)),
            SLOT(slotEnableQuestionMode(bool)));
    connect(ui.actionEnableStoreAudioToDisk, SIGNAL(triggered(bool)),
            SLOT(slotEnableAudioRecording(bool)));
    /* End - Extra toolbar buttons */

    /* Begin - WM_TEAMTALK_* messages */
    connect(this, SIGNAL(serverUpdate()), ui.channelsWidget,
            SLOT(slotServerUpdate()));
    connect(this, SIGNAL(newChannel(int)), ui.channelsWidget, 
            SLOT(slotAddChannel(int)));
    connect(this, SIGNAL(updateChannel(int)),
            SLOT(slotChannelUpdate(int)));
    connect(this, SIGNAL(updateChannel(int)), ui.channelsWidget, 
            SLOT(slotUpdateChannel(int)));
    connect(this, SIGNAL(removeChannel(int)), ui.channelsWidget, 
            SLOT(slotRemoveChannel(int)));
    connect(this, SIGNAL(userJoined(int,int)), ui.channelsWidget, 
            SLOT(slotUserJoin(int,int)));
    /* 'this' must be connected first since it needs to extract the old
     * user */
    connect(this, SIGNAL(userUpdate(int,int)), this, 
            SLOT(slotUserUpdate(int,int)));
    connect(this, SIGNAL(userUpdate(int,int)), ui.channelsWidget, 
            SLOT(slotUserUpdate(int,int)));
    connect(this, SIGNAL(userLeft(int,int)), ui.channelsWidget, 
            SLOT(slotUserRemove(int,int)));
    connect(this, SIGNAL(userTalking(int,bool)), ui.channelsWidget, 
            SLOT(slotUserTalking(int,bool)));
    connect(this, SIGNAL(userLeft(int,int)), ui.videogridWidget, 
            SLOT(slotRemoveUser(int,int)));
    connect(this, SIGNAL(newVideoFrame(int,int)), ui.channelsWidget, 
            SLOT(slotUserVideoFrame(int,int)));    
    /* End - WM_TEAMTALK_* messages */

    m_timers.insert(startTimer(1000), TIMER_ONE_SECOND);
    m_timers.insert(startTimer(50), TIMER_UI_UPDATE);

#if !defined(Q_OS_WIN32)
    //Windows uses its HWND for message handling, other platforms must
    //pull using a timer
    m_timers.insert(startTimer(20), TIMER_PROCESS_TTEVENT);
#endif
}

MainWindow::~MainWindow()
{
#ifdef Q_OS_LINUX
    if(m_display)
        XCloseDisplay(m_display);
#endif
}

void MainWindow::loadSettings()
{
    QString lang = ttSettings->value(SETTINGS_DISPLAY_LANGUAGE, "").toString();
    if(!lang.isEmpty())
    {
        ttTranslator = new QTranslator(this);
        if(!ttTranslator->load(lang, TRANSLATE_FOLDER))
        {
            QMessageBox::information(this, "Translate", 
                QString("Failed to load language file %1").arg(lang));
            delete ttTranslator;
            ttTranslator = NULL;
        }
        else
        {
            QApplication::installTranslator(ttTranslator);
            this->ui.retranslateUi(this);
        }
    }

    //load settings
    bool ptt = ttSettings->value(SETTINGS_GENERAL_PUSHTOTALK).toBool();
    slotMeEnablePushToTalk(ptt);
    bool vox = ttSettings->value(SETTINGS_GENERAL_VOICEACTIVATED).toBool();
    slotMeEnableVoiceActivation(vox);

    //load shortcuts
    loadHotKeys();

    int value = ttSettings->value(SETTINGS_SOUND_MASTERVOLUME, SOUND_VOLUME_MAX).toInt();
    ui.volumeSlider->setValue(value);
    //slotMasterVolumeChanged(value);
    value = ttSettings->value(SETTINGS_SOUND_VOICEACTIVATIONLEVEL, 0).toInt();
    ui.voiceactSlider->setValue(value);
    //slotVoiceActivationLevelChanged(value);
    value = ttSettings->value(SETTINGS_SOUND_MICROPHONEGAIN, SOUND_GAIN_DEFAULT).toInt();
    ui.micSlider->setValue(value);
    //slotMicrophoneGainChanged(value);

    if(ttSettings->value(SETTINGS_DISPLAY_ALWAYSONTOP, false).toBool())
    {
        setWindowFlags(windowFlags() | Qt::WindowStaysOnTopHint);
#ifdef Q_OS_WIN32
        //Qt creates a new HWND to support "always on top" so no more messages
        //are posted to our HWND. This is actually unsafe since the previous HWND
        //might have unhandled messages.
        TT_SwapTeamTalkHWND(ttInst, reinterpret_cast<HWND>(winId()));
#endif
    }

    //show number of users
    ui.channelsWidget->setShowUserCount(ttSettings->value(SETTINGS_DISPLAY_USERSCOUNT, true).toBool());
    ui.channelsWidget->updateItemTextLength(ttSettings->value(SETTINGS_DISPLAY_MAX_STRING,
                                            SETTINGS_DISPLAY_MAX_STRING_DEFAULT).toInt());
    slotUpdateUI();

    //move window to last position
    QVariantList windowpos = ttSettings->value(SETTINGS_DISPLAY_WINDOWPOS).toList();
    if(windowpos.size() == 4)
    {
        int x = windowpos[0].toInt();
        int y = windowpos[1].toInt();
        int w = windowpos[2].toInt();
        int h = windowpos[3].toInt();
        int desktopW = QApplication::desktop()->width();
        int desktopH = QApplication::desktop()->height();
        if(x <= desktopW && y <= desktopH)
            setGeometry(x, y, w, h);
        ui.splitter->restoreState(ttSettings->value(SETTINGS_DISPLAY_SPLITTER).toByteArray());
        ui.desktopsplitter->restoreState(ttSettings->value(SETTINGS_DISPLAY_DESKTOPSPLITTER).toByteArray());
    }
    //set files header to last position
    ui.filesView->header()->restoreState(ttSettings->value(SETTINGS_DISPLAY_FILESHEADER).toByteArray());

    if(ttSettings->value(SETTINGS_DISPLAY_APPUPDATE, true).toBool())
    {
        QUrl url(URL_APPUPDATE);

        m_http_manager = new QNetworkAccessManager(this);
        connect(m_http_manager, SIGNAL(finished(QNetworkReply*)),
                SLOT(slotHttpUpdateReply(QNetworkReply*)));

        QNetworkRequest request(url);
        m_http_manager->get(request);
    }

    if(ttSettings->value(SETTINGS_DISPLAY_STARTMINIMIZED, false).toBool())
        QTimer::singleShot(0, this, SLOT(showMinimized()));

    //started from .tt file?
    bool connect_ok = parseArgs(QApplication::arguments());

    if(connect_ok)
        QTimer::singleShot(0, this, SLOT(slotConnectToLatest()));

    // TT_EnableAudioBlockEvent(ttInst, true);
}

bool MainWindow::parseArgs(const QStringList& args)
{
    for(int i=1;i<args.size();i++)
    {
        if(args[i].right(QString(TTFILE_EXT).size()).toLower() == TTFILE_EXT)
        {
            slotLoadTTFile(args[i]);
            return false;
        }
        else if(args[i].left(QString(TTLINK_PREFIX).size()).toLower() == TTLINK_PREFIX)
        {
            //tt://([^\??!/]*)/?\?
            QRegExp rx(QString("%1([^\\??!/]*)/?\\?").arg(TTLINK_PREFIX));
            if(rx.indexIn(args[i])>=0)
            {
                HostEntry entry;
                entry.ipaddr = rx.cap(1);
                entry.tcpport = DEFAULT_TCPPORT;
                entry.udpport = DEFAULT_UDPPORT;
                QString prop = args[i].mid(rx.matchedLength());
                //&?tcpport=(\d+)&?
                rx.setPattern("&?tcpport=(\\d+)&?");
                if(rx.indexIn(prop)>=0)
                    entry.tcpport = rx.cap(1).toInt();
                //&?udpport=(\d+)&?
                rx.setPattern("&?udpport=(\\d+)&?");
                if(rx.indexIn(prop)>=0)
                    entry.udpport = rx.cap(1).toInt();
                //&?srvpasswd=([^&]*)&?
                rx.setPattern("&?srvpasswd=([^&]*)&?");
                if(rx.indexIn(prop)>=0)
                    entry.srvpasswd = rx.cap(1);
                //&?username=([^&]*)&?
                rx.setPattern("&?username=([^&]*)&?");
                if(rx.indexIn(prop)>=0)
                    entry.username = rx.cap(1);
                //&?password=([^&]*)&?
                rx.setPattern("&?password=([^&]*)&?");
                if(rx.indexIn(prop)>=0)
                    entry.password = rx.cap(1);
                //&?channel=([^&]*)&?
                rx.setPattern("&?channel=([^&]*)&?");
                if(rx.indexIn(prop)>=0)
                    entry.channel = rx.cap(1);
                //&?chanpasswd=([^&]*)&?
                rx.setPattern("&?chanpasswd=([^&]*)&?");
                if(rx.indexIn(prop)>=0)
                    entry.chanpasswd = rx.cap(1);

                AddLatestHost(entry);
                Disconnect();
                Connect();
                return false;
            }
        }
        else if(args[i] == "-noconnect")
            return false;
#if defined(Q_OS_WIN32)
        else if(args[i] == "-fwadd")
        {
            //add firewall exception
            QString appPath = QApplication::applicationFilePath();
            appPath = QDir::toNativeSeparators(appPath);
            if(!TT_Firewall_AppExceptionExists(_W(appPath)))
            {
                QMessageBox::StandardButton answer = QMessageBox::question(this, 
                    APPTITLE, 
                    tr("Do you wish to add %1 to the Windows Firewall exception list?")
                    .arg(APPTITLE), 
                    QMessageBox::Yes | QMessageBox::No);

                if(answer == QMessageBox::Yes &&
                   !TT_Firewall_AddAppException(_W(QString(APPTITLE)), 
                                                _W(appPath)))
                {
                    QMessageBox::critical(this, tr("Firewall exception"),
                        tr("Failed to add %1 to Windows Firewall exceptions.")
                        .arg(appPath));
                }
            }
        }
        else if(args[i] == "-fwremove")
        {
            //remove firewall exception
            QString appPath = QApplication::applicationFilePath();
            appPath = QDir::toNativeSeparators(appPath);
            if(TT_Firewall_AppExceptionExists(_W(appPath)))
            {
                if(!TT_Firewall_RemoveAppException(_W(appPath)))
                {
                    QMessageBox::critical(this, tr("Firewall exception"),
                        tr("Failed to remove %1 from Windows Firewall exceptions.")
                        .arg(appPath));
                }
            }
        }
#endif
        else
        {
            QMessageBox::information(this, tr("Startup arguments"), 
                tr("Program argument \"%1\" is unrecognized.").arg(args[i]));
        }
    }

    return true;
}


void MainWindow::processTTMessage(const TTMessage& msg)
{
    bool update_ui = false;

    switch(msg.wmMsg)
    {
    case WM_TEAMTALK_CON_SUCCESS :
    {
        //disable reconnect timer
        killLocalTimer(TIMER_RECONNECT);

        //switch to connected icon in tray
        if(m_sysicon)
            m_sysicon->setIcon(QIcon(APPTRAYICON_CON));

        //reset stats
        ZERO_STRUCT(m_clientstats);

        HostEntry entry;
        GetLatestHost(0, entry);
        QString nick = ttSettings->value(QString(SETTINGS_GENERAL_NICKNAME)).toString();

        int cmdid = TT_DoLogin(ttInst, _W(nick), _W(entry.srvpasswd), 
                               _W(entry.username), _W(entry.password));
        if(cmdid>0)
            m_commands.insert(cmdid, CMD_COMPLETE_LOGIN);

        addStatusMsg(tr("Connected to %1 TCP port %2 UDP port %3")
                     .arg(entry.ipaddr).arg(entry.tcpport).arg(entry.udpport));

        //query server's max payload
        if(ttSettings->value(SETTINGS_CONNECTION_QUERYMAXPAYLOAD, false).toBool())
            TT_QueryMaxPayload(ttInst, 0);

        update_ui = true;
    }
    break;
    case WM_TEAMTALK_CON_FAILED :
    {
        Disconnect();
        HostEntry entry;
        GetLatestHost(0, entry);

        addStatusMsg(tr("Failed to connect to %1 TCP port %2 UDP port %3")
                     .arg(entry.ipaddr).arg(entry.tcpport).arg(entry.udpport));

        update_ui = true;
    }
    break;
    case WM_TEAMTALK_CON_LOST :
    {
        Disconnect();
        if(ttSettings->value(SETTINGS_CONNECTION_RECONNECT, true).toBool())
            m_timers[startTimer(5000)] = TIMER_RECONNECT;

        HostEntry entry;
        GetLatestHost(0, entry);

        addStatusMsg(tr("Connection lost to %1 TCP port %2 UDP port %3")
                     .arg(entry.ipaddr).arg(entry.tcpport).arg(entry.udpport));

        PlaySoundEvent(SOUNDEVENT_SERVERLOST);

        update_ui = true;
    }
    break;
    case WM_TEAMTALK_CON_P2P :
    {
        User user;
        if(TT_GetUser(ttInst, msg.wParam, &user))
        {
            if(msg.lParam)
                addStatusMsg(tr("Established P2P connection to %1")
                             .arg(_Q(user.szNickname)));
            else
            {
                addStatusMsg(tr("Failed to establish P2P connection to %1")
                             .arg(_Q(user.szNickname)));
                if((m_srvprop.uUserRights & USERRIGHT_FORWARD_AUDIO) == 0)
                    addStatusMsg(tr("%1 will be unavailable for audio transmission")
                                 .arg(_Q(user.szNickname)));
                if((m_srvprop.uUserRights & USERRIGHT_FORWARD_VIDEO) == 0)
                    addStatusMsg(tr("%1 will be unavailable for video transmission")
                                 .arg(_Q(user.szNickname)));
            }
        }
    }
    break;
    case WM_TEAMTALK_CON_MAX_PAYLOAD_UPDATED :
        qDebug() << "User #" << msg.wParam << "max payload is" << msg.lParam;
    break;
    case WM_TEAMTALK_CMD_PROCESSING :
        commandProcessing(msg.wParam, msg.lParam);
        break;
    case WM_TEAMTALK_CMD_ERROR :
    {
        //store current command type
        CommandComplete cmd_type = m_commands[m_current_cmdid];

        //if we're waiting for a result from this command then clear it
        //since the command failed (this must be done before showing 
        //and error message since the login will otherwise complete).
        m_commands.remove(m_current_cmdid);

        emit(cmdError(msg.wParam, msg.lParam));

        showTTErrorMessage(msg.wParam, cmd_type);

        update_ui = true;
    }
    break;
    case WM_TEAMTALK_CMD_SUCCESS :
        emit(cmdSuccess(msg.wParam));

        update_ui = true;
        break;
    case WM_TEAMTALK_CMD_MYSELF_LOGGEDIN :
        //ui.chatEdit->updateServer();
        addStatusMsg(tr("Logged in"));

        update_ui = true;
        break;
    case WM_TEAMTALK_CMD_MYSELF_LOGGEDOUT :
        addStatusMsg(tr("Logged out"));
        Disconnect();
        update_ui = true;

        break;
    case WM_TEAMTALK_CMD_MYSELF_JOINED :
        processMyselfJoined(msg.wParam);
        update_ui = true;
        break;
    case WM_TEAMTALK_CMD_MYSELF_LEFT :
        processMyselfLeft(msg.wParam);
        update_ui = true;
        break;
    case WM_TEAMTALK_CMD_MYSELF_KICKED :
    {
        User user;
        if(ui.channelsWidget->getUser(msg.wParam, user))
            addStatusMsg(tr("Kicked by %1").arg(_Q(user.szNickname)));
        else
            addStatusMsg(tr("Kicked by unknown user"));

        update_ui = true;
    }
    break;
    case WM_TEAMTALK_CMD_SERVER_UPDATE :
    {
        ui.chatEdit->updateServer(); 
        //write warning if server doesn't allowed forwarding
        if(TT_GetServerProperties(ttInst, &m_srvprop))
        {
            //enable peer to peer networking if server doesn't allow forwarding
            if((USERRIGHT_FORWARD_AUDIO & m_srvprop.uUserRights) == 0 &&
               (TT_GetFlags(ttInst) & CLIENT_P2P_AUDIO) == 0)
            {
                addStatusMsg(tr("Peer to peer networking is required"
                                " for audio transmission on this server. "
                                "Now enabled!"));
                TT_EnablePeerToPeer(ttInst, TRANSMIT_AUDIO, TRUE);
            }
            else if(ttSettings->value(SETTINGS_CONNECTION_P2PAUDIO, false).toBool() !=
                    (bool)(TT_GetFlags(ttInst) & CLIENT_P2P_AUDIO))
            {
                TT_EnablePeerToPeer(ttInst, TRANSMIT_AUDIO, ttSettings->value(SETTINGS_CONNECTION_P2PAUDIO, false).toBool());
                addStatusMsg(tr("Peer to peer networking for audio disabled!"));
            }

            if((USERRIGHT_FORWARD_VIDEO & m_srvprop.uUserRights) == 0 &&
               (TT_GetFlags(ttInst) & CLIENT_P2P_VIDEO) == 0)
            {
                addStatusMsg(tr("Peer to peer networking is required "
                                "for video transmission on this server. "
                                "Now enabled!"));
                TT_EnablePeerToPeer(ttInst, TRANSMIT_VIDEO, TRUE);
            }
            else if(ttSettings->value(SETTINGS_CONNECTION_P2PVIDEO, false).toBool() !=
                    (bool)(TT_GetFlags(ttInst) & CLIENT_P2P_VIDEO))
            {
                TT_EnablePeerToPeer(ttInst, TRANSMIT_VIDEO, ttSettings->value(SETTINGS_CONNECTION_P2PVIDEO, false).toBool());
                addStatusMsg(tr("Peer to peer networking for video disabled!"));
            }
        }
        emit(serverUpdate());

        update_ui = true;
    }
    break;
    case WM_TEAMTALK_CMD_CHANNEL_NEW :
        emit(newChannel(msg.wParam));
        break;
    case WM_TEAMTALK_CMD_CHANNEL_UPDATE :
    {
        if(m_mychannel.nChannelID == (int)msg.wParam)
        {
            TT_GetChannel(ttInst, msg.wParam, &m_mychannel);
            update_ui = true;
        }
        emit(updateChannel(msg.wParam));

    }
    break;
    case WM_TEAMTALK_CMD_CHANNEL_REMOVE :
        emit(removeChannel(msg.wParam));
        break;
    case WM_TEAMTALK_CMD_USER_LOGGEDIN :
    {
        emit(userLogin(msg.wParam));
        QString audiofolder = ttSettings->value(SETTINGS_GENERAL_AUDIOSTORAGE_FOLDER).toString();
        AudioFileFormat aff = (AudioFileFormat)ttSettings->value(SETTINGS_GENERAL_AUDIOSTORAGE_FILEFORMAT, AFF_WAVE_FORMAT).toInt();
        if(m_audiostorage_mode & AUDIOSTORAGE_SEPARATEFILES)
            TT_SetUserAudioFolder(ttInst, msg.wParam, _W(audiofolder), NULL, aff);

        updateUserSubscription(msg.wParam);
    }
    break;
    case WM_TEAMTALK_CMD_USER_LOGGEDOUT :
        emit(userLogout(msg.wParam));
        break;
    case WM_TEAMTALK_CMD_USER_JOINED :
        emit(userJoined(msg.wParam, msg.lParam));
        processUserJoined(msg.wParam, msg.lParam);
        break;
    case WM_TEAMTALK_CMD_USER_LEFT :
        emit(userLeft(msg.wParam, msg.lParam));
        processUserLeft(msg.wParam, msg.lParam);
        break;
    case WM_TEAMTALK_CMD_USER_UPDATE :
    {
        User prev_user;
        ZERO_STRUCT(prev_user);
        ui.channelsWidget->getUser(msg.wParam, prev_user);

        emit(userUpdate(msg.wParam, msg.lParam));

        //store nick name so we can display who left the server
        User user;
        if(TT_GetUser(ttInst, msg.wParam, &user))
            m_nicknames[msg.wParam] = _Q(user.szNickname);

        if(user.nStatusMode & STATUSMODE_STREAM_MEDIAFILE)
            TT_SetUserMediaBufferSize(ttInst, user.nUserID, STREAMMEDIA_AUDIOBUF_SIZE);
        else
            TT_SetUserMediaBufferSize(ttInst, user.nUserID, 0);

        if(msg.wParam != TT_GetMyUserID(ttInst) &&
           msg.lParam == m_mychannel.nChannelID)
        {
            if((prev_user.nStatusMode & STATUSMODE_QUESTION) == 0 &&
               (user.nStatusMode & STATUSMODE_QUESTION))
               PlaySoundEvent(SOUNDEVENT_QUESTIONMODE);
        }

        //update desktop access button
        if((prev_user.uPeerSubscriptions & SUBSCRIBE_DESKTOPINPUT) !=
           (user.uPeerSubscriptions & SUBSCRIBE_DESKTOPINPUT))
           slotUpdateDesktopTabUI();
    }
    break;
    case WM_TEAMTALK_CMD_USER_TEXTMSG :
        processTextMessage(msg.wParam, msg.lParam);
        break;
    case WM_TEAMTALK_CMD_FILE_NEW :
    case WM_TEAMTALK_CMD_FILE_REMOVE :
    {
        //only update files list if we're not currently logging in or 
        //joining a channel
        cmdreply_t::iterator ite = m_commands.find(m_current_cmdid);
        if(m_filesmodel->getChannelID() == (INT32)msg.lParam &&
           (ite == m_commands.end() || (*ite != CMD_COMPLETE_LOGIN && 
                                        *ite != CMD_COMPLETE_JOINCHANNEL)) )
        {
            updateChannelFiles(msg.lParam);
            PlaySoundEvent(SOUNDEVENT_FILESUPD);
        }

        update_ui = true;
    }
    break;
    case WM_TEAMTALK_FILETRANSFER :
        if(msg.lParam == FILETRANSFER_STARTED)
        {
            FileTransferDlg* dlg = new FileTransferDlg(msg.wParam, NULL);
            connect(this, SIGNAL(filetranferCompleted(int)), dlg, 
                    SLOT(slotTransferCompleted(int)));
            connect(this, SIGNAL(filetransferFailed(int)), dlg, 
                    SLOT(slotTransferFailed(int)));
            dlg->setAttribute(Qt::WA_DeleteOnClose);
            dlg->show();
        }
        else if(msg.lParam == FILETRANSFER_FINISHED)
            emit(filetranferCompleted(msg.wParam));
        else if(msg.lParam == FILETRANSFER_ERROR)
        {
            FileTransfer transfer;
            if(TT_GetFileTransferInfo(ttInst, msg.wParam, &transfer))
            {
                if(transfer.bInbound)
                    QMessageBox::critical(this, MENUTEXT(ui.actionDownloadFile->text()),
                                          tr("Failed to download file %1")
                                          .arg(_Q(transfer.szRemoteFileName)));
                else
                    QMessageBox::critical(this, MENUTEXT(ui.actionUploadFile->text()),
                                          tr("Failed to upload file %1")
                                          .arg(_Q(transfer.szLocalFilePath)));
                emit(filetransferFailed(msg.wParam));
            }
        }

        update_ui = true;
        break;
    case WM_TEAMTALK_INTERNAL_ERROR :
    {
        QString textmsg;
        switch(msg.wParam)
        {
            /***** Internal TeamTalk errors not related to commands ********/
        case INTERR_SNDINPUT_FAILURE :
            textmsg = tr("Failed to initialize sound input device"); break;
        case INTERR_SNDOUTPUT_FAILURE :
            textmsg = tr("Failed to initialize sound output device"); break;
        default :
            TTCHAR buff[TT_STRLEN] = {0};
            TT_GetErrorMessage(msg.wParam, buff);
            textmsg = _Q(buff);
            break;
        }
        QMessageBox::critical(this, tr("Internal Error"), textmsg);
        addStatusMsg(textmsg);

        update_ui = true;
    }
    break;
    case WM_TEAMTALK_USER_TALKING :
        emit(userTalking(msg.wParam, (bool)msg.lParam));
        if(msg.lParam)
        {
            m_talking.insert(msg.wParam);

            if(m_sysicon && m_sysicon->isVisible() &&
               m_talking.size() == 1)
                m_sysicon->setIcon(QIcon(APPTRAYICON_ACTIVE));

            //update voice gain if not already set
            updateUserGainLevel(msg.wParam);
        }
        else
        {
            m_talking.remove(msg.wParam);

            if(m_sysicon && m_sysicon->isVisible() &&
               m_talking.size() == 0)
                m_sysicon->setIcon(QIcon(APPTRAYICON_CON));
        }
        
        if(m_talking.empty() /*&& !TT_IsTransmitting(ttInst, TRUE, FALSE)*/)
            PlaySoundEvent(SOUNDEVENT_SILENCE);

        break;
    case WM_TEAMTALK_VOICE_ACTIVATION :
        emit(userTalking(TT_GetMyUserID(ttInst), 
                         TT_IsTransmitting(ttInst, TRANSMIT_AUDIO)));
        break;
    case WM_TEAMTALK_STREAM_AUDIOFILE_USER :
    {
        User user;
        if(!TT_GetUser(ttInst, msg.wParam, &user))
            break;
        switch(msg.lParam)
        {
        case AFS_ERROR :
            addStatusMsg(tr("Error streaming audio file to %1")
                         .arg(_Q(user.szNickname)));
            break;
        case AFS_STARTED :
            addStatusMsg(tr("Started streaming audio file to %1")
                         .arg(_Q(user.szNickname)));
            break;
        case AFS_FINISHED :
            addStatusMsg(tr("Finished streaming audio file to %1")
                         .arg(_Q(user.szNickname)));
            break;
        case AFS_ABORTED :
            addStatusMsg(tr("Aborted streaming audio file to %1")
                         .arg(_Q(user.szNickname)));
            break;
        }
    }
    break;
    case WM_TEAMTALK_STREAM_AUDIOFILE_CHANNEL :
    {
        TTCHAR chpath[TT_STRLEN] = {0};
        if(!TT_GetChannelPath(ttInst, msg.wParam, chpath))
            break;

        switch(msg.lParam)
        {
        case AFS_ERROR :
            addStatusMsg(tr("Error streaming audio file to %1")
                         .arg(_Q(chpath)));
            break;
        case AFS_STARTED :
            addStatusMsg(tr("Started streaming audio file to %1")
                         .arg(_Q(chpath)));
            break;
        case AFS_FINISHED :
            addStatusMsg(tr("Finished streaming audio file to %1")
                         .arg(_Q(chpath)));
            break;
        case AFS_ABORTED :
            addStatusMsg(tr("Aborted streaming audio file to %1")
                         .arg(_Q(chpath)));
            break;
        }

        if(ui.actionStreamAudioFileToChannel->isChecked() &&
           msg.lParam == AFS_FINISHED &&
           ttSettings->value(SETTINGS_GENERAL_STREAMAUDIO_LOOP, false).toBool())
        {
            QString fileName = ttSettings->value(SETTINGS_GENERAL_STREAMAUDIO_FILENAME).toString();
            if(!TT_StartStreamingAudioFileToChannel(ttInst, 0, _W(fileName)))
            {
                ui.actionStreamAudioFileToChannel->setChecked(false);
                addStatusMsg(tr("Error streaming audio file to %1")
                             .arg(_Q(chpath)));
            }
        }
        else
            ui.actionStreamAudioFileToChannel->setChecked(msg.lParam == AFS_STARTED);

        update_ui = true;
    }
    break;
    case WM_TEAMTALK_STREAM_MEDIAFILE_CHANNEL :
    {
        switch(msg.lParam)
        {
        case AFS_ERROR :
            addStatusMsg(tr("Error streaming media file to channel"));
            stopStreamMediaFile();
            break;
        case AFS_STARTED :
            addStatusMsg(tr("Started streaming media file to channel"));
            break;
        case AFS_FINISHED :
            addStatusMsg(tr("Finished streaming media file to channel"));
            stopStreamMediaFile();
            break;
        case AFS_ABORTED :
            addStatusMsg(tr("Aborted streaming media file to channel"));
            stopStreamMediaFile();
            break;
        }

        if(msg.lParam == AFS_FINISHED &&
           ttSettings->value(SETTINGS_GENERAL_STREAMMEDIA_LOOP, false).toBool())
        {
            startStreamMediaFile();
        }

        //update if still talking
        emit(userTalking(TT_GetMyUserID(ttInst), 
                         TT_IsTransmitting(ttInst, TRANSMIT_AUDIO)));
        update_ui = true;
    }
    break;
    case WM_TEAMTALK_USER_VIDEOFRAME :
        //local video is userID = 0 (msg.wParam)

        //if local video is disabled then don't process pending video
        //frames
        if(msg.wParam == 0 && 
           (TT_GetFlags(ttInst) & CLIENT_VIDEO_READY) == 0)
            break;
        
        //pass new video frame (if it's not being ignored)
        if(m_vid_exclude.find(msg.wParam) == m_vid_exclude.end())
        {
            if(m_uservideo.find(msg.wParam) == m_uservideo.end() &&
               !ui.videogridWidget->userExists(msg.wParam))
            {
                //it's a new video session

                if(ttSettings->value(SETTINGS_DISPLAY_VIDEOPOPUP, false).toBool())
                    slotNewUserVideoDlg(msg.wParam, QSize(),
                                        getVideoStreamID(msg.wParam));
                else
                    ui.videogridWidget->slotAddUser(msg.wParam, 
                                                    getVideoStreamID(msg.wParam));

                PlaySoundEvent(SOUNDEVENT_NEWVIDEO);

                User user;
                if(TT_GetUser(ttInst, msg.wParam, &user))
                    addStatusMsg(tr("New video session from %1")
                    .arg(_Q(user.szNickname)));

                update_ui = true;
            }
            emit(newVideoFrame(msg.wParam, msg.lParam));
        }
        break;
    case WM_TEAMTALK_USER_DESKTOPWINDOW :
        if(m_userdesktop.find(msg.wParam) == m_userdesktop.end() &&
           !ui.desktopgridWidget->userExists(msg.wParam))
        {
            //it's a new desktop session

            //sesion id == 0 means it has ended
            if(msg.lParam)
            {
                if(ttSettings->value(SETTINGS_DISPLAY_DESKTOPPOPUP, false).toBool())
                    slotDetachUserDesktop(msg.wParam, QSize());
                else
                    ui.desktopgridWidget->slotAddUser(msg.wParam);

                PlaySoundEvent(SOUNDEVENT_NEWDESKTOP);

                User user;
                if(TT_GetUser(ttInst, msg.wParam, &user))
                    addStatusMsg(tr("New desktop session from %1")
                    .arg(_Q(user.szNickname)));
            }

            update_ui = true;
        }
        emit(newDesktopWindow(msg.wParam, msg.lParam));
        break;
    case WM_TEAMTALK_DESKTOPWINDOW_TRANSFER :
        if(m_desktopsession_id != msg.wParam) //new session
            m_desktopsession_remain = 0;
        m_desktopsession_id = msg.wParam;

        if(m_desktopsession_remain == 0)
            m_desktopsession_total = msg.lParam;
        m_desktopsession_remain = msg.lParam;

        if(msg.lParam == 0 && m_desktopsend_on_completion)
        {
            if(sendDesktopWindow())
            {
                restartSendDesktopWindowTimer();
                m_desktopsend_on_completion = false;
            }
        }

        if(msg.wParam == 0)
            addStatusMsg(tr("Your desktop session was cancelled"));
        break;
    case WM_TEAMTALK_USER_DESKTOPCURSOR :
        emit(userDesktopCursor(msg.wParam, msg.lParam));
        break;
    case WM_TEAMTALK_USER_DESKTOPINPUT :
        processDesktopInput(msg.wParam);
        break;
    case WM_TEAMTALK_USER_AUDIOFILE :
    {
        User user;
        if(!TT_GetUser(ttInst, msg.wParam, &user))
            break;
        if(msg.lParam == AFS_ERROR)
            addStatusMsg(tr("Failed to write audio file for %1")
                         .arg(_Q(user.szNickname)));
    }
    break;
    case WM_TEAMTALK_USER_AUDIOBLOCK :
    {
        // AudioBlock block;
        // if(TT_AcquireUserAudioBlock(ttInst, msg.wParam, &block))
        // {
        //    int mean = 0;
        //    int size = block.nSamples * block.nChannels;
        //    short* ptr = (short*)block.lpRawAudio;
        //    for(int i=0;i<size;i++)
        //        mean += ptr[i];
        //    mean /= size;
        //    qDebug() << "AudioBlock from #" << msg.wParam << "index" 
        //        << block.uSampleIndex << "samples" << block.nSamples
        //        << "avg" << mean;
        //    TT_ReleaseUserAudioBlock(ttInst, msg.wParam);
        // }
    }
    break;
    case WM_TEAMTALK_HOTKEY :
        hotkeyToggle((HotKeyID)msg.wParam, (bool)msg.lParam);
        break;
    default :
        qDebug() << "Unknown message type" << msg.wmMsg;
    }

    //update menus, button, etc.
    if(update_ui)
        slotUpdateUI();
}


void MainWindow::commandProcessing(int cmdid, bool complete)
{
    //command reply starting -> store command reply ID
    if(!complete)
        m_current_cmdid = cmdid;

    //check if a dialog should be opened to show server reply
    cmdreply_t::iterator ite;
    if(complete && /* Command has completed */
       (ite = m_commands.find(cmdid)) != m_commands.end()) //Command we're interested in knowing the reply of
    {
        switch(*ite)
        {
        case CMD_COMPLETE_LOGIN :
            cmdLoggedIn(TT_GetMyUserID(ttInst));
            break;
        case CMD_COMPLETE_JOINCHANNEL :
            //unsubscribe
            cmdJoinedChannel(TT_GetMyChannelID(ttInst));
            break;
        case CMD_COMPLETE_LISTBANS :
        {
            if(!m_bannedusersdlg)
            {
                m_bannedusersdlg = new BannedUsersDlg();
                connect(m_bannedusersdlg, SIGNAL(finished(int)),
                        SLOT(slotClosedBannedUsersDlg(int)));
                m_bannedusersdlg->setAttribute(Qt::WA_DeleteOnClose);
                m_bannedusersdlg->show();
            }
            else
                m_bannedusersdlg->activateWindow();
        }
        break;
        case CMD_COMPLETE_LISTACCOUNTS :
        {
            if(!m_useraccountsdlg)
            {
                m_useraccountsdlg = new UserAccountsDlg();
                connect(this, SIGNAL(cmdSuccess(int)), m_useraccountsdlg, 
                        SLOT(slotCmdSuccess(int)));
                connect(this, SIGNAL(cmdError(int, int)), m_useraccountsdlg, 
                        SLOT(slotCmdError(int,int)));
                connect(m_useraccountsdlg, SIGNAL(finished(int)),
                        SLOT(slotClosedUserAccountsDlg(int)));
                m_useraccountsdlg->setAttribute(Qt::WA_DeleteOnClose);
                m_useraccountsdlg->show();
            }
            else
                m_useraccountsdlg->activateWindow();
            break;
        }
        }
    }

    //command reply completed -> clear command reply ID
    if(complete)
    {
        m_commands.remove(cmdid);
        m_current_cmdid = 0;
    }
}

void MainWindow::cmdLoggedIn(int myuserid)
{
    Q_UNUSED(myuserid);

    //login command completed

    QString statusmsg = ttSettings->value(SETTINGS_GENERAL_STATUSMESSAGE).toString();
    //change to female if set
    if(!ttSettings->value(SETTINGS_GENERAL_GENDER, true).toBool())
        m_statusmode |= STATUSMODE_FEMALE;
    //set status mode flags
    if(m_statusmode || statusmsg.size())
        TT_DoChangeStatus(ttInst, m_statusmode, _W(statusmsg));

    UserAccount account;
    ZERO_STRUCT(account);
    TT_GetMyUserAccount(ttInst, &account);

    //join channel (if specified)
    HostEntry host;
    GetLatestHost(0, host);

    Channel tmpchan;
    int channelid = TT_GetChannelIDFromPath(ttInst, _W(host.channel));
    int parentid = 0;
    int account_channelid = TT_GetChannelIDFromPath(ttInst, 
                                                    account.szInitChannel);

    //see if parent channel exists (otherwise we cannot create it)
    QStringList subchannels = host.channel.split('/');
    if(subchannels.size())
    {
        QStringList parent = subchannels;
        parent.erase(parent.end()-1);
        QString chanpath = parent.join("/");
        parentid = TT_GetChannelIDFromPath(ttInst, _W(chanpath));
    }

    if(m_last_channel.nChannelID && //join using last channel
       TT_GetChannel(ttInst, m_last_channel.nParentID, 
                     &tmpchan)>0)
    {
        int cmdid = TT_DoJoinChannel(ttInst, &m_last_channel);
        if(cmdid>0)
            m_commands.insert(cmdid, CMD_COMPLETE_JOINCHANNEL);
    }
    else if(_Q(account.szInitChannel).size() && account_channelid > 0)
    {
        int cmdid = TT_DoJoinChannelByID(ttInst, account_channelid, 
                                         _W(QString()));
        if(cmdid>0)
            m_commands.insert(cmdid, CMD_COMPLETE_JOINCHANNEL);
    }
    else if(host.channel.size() && channelid > 0) //join if channel exists
    {
        int cmdid = TT_DoJoinChannelByID(ttInst, channelid, _W(host.chanpasswd));
        if(cmdid>0)
            m_commands.insert(cmdid, CMD_COMPLETE_JOINCHANNEL);
    }
    else if(host.channel.size() && parentid>0) //make a new channel if parent exists
    {
        QString name;
        if(subchannels.size())
            name = subchannels.last();

        Channel chan;
        ZERO_STRUCT(chan);
        chan.nParentID = parentid;
        chan.nMaxUsers = m_srvprop.nMaxUsers;
#if DEFAULT_AUDIOCODEC != SPEEX_CODEC
#pragma error Configured as Speex
#endif
        chan.codec.nCodec = DEFAULT_AUDIOCODEC;
        chan.codec.speex.nQuality = DEFAULT_SPEEX_QUALITY;
        chan.codec.speex.nBandmode = DEFAULT_SPEEX_BANDMODE;
        chan.codec.speex.nMSecPerPacket = DEFAULT_SPEEX_DELAY;

        chan.audiocfg.bEnableAGC = DEFAULT_AGC_ENABLE;
        chan.audiocfg.nGainLevel = DEFAULT_AGC_GAINLEVEL;
        chan.audiocfg.nMaxIncDBSec = DEFAULT_AGC_INC_MAXDB;
        chan.audiocfg.nMaxDecDBSec = DEFAULT_AGC_DEC_MAXDB;
        chan.audiocfg.nMaxGainDB = DEFAULT_AGC_GAINMAXDB;
        chan.audiocfg.bEnableDenoise = DEFAULT_DENOISE_ENABLE;
        chan.audiocfg.nMaxNoiseSuppressDB = DEFAULT_DENOISE_SUPPRESS;

        COPY_TTSTR(chan.szName, name);
        COPY_TTSTR(chan.szPassword, host.chanpasswd);
        int cmdid = TT_DoJoinChannel(ttInst, &chan);
        if(cmdid>0)
            m_commands.insert(cmdid, CMD_COMPLETE_JOINCHANNEL);
    }
    else if(ttSettings->value(SETTINGS_CONNECTION_AUTOJOIN, true).toBool()) //just join root
    {
        if(host.channel.size())
            addStatusMsg(tr("Cannot join channel %1").arg(host.channel));

        //auto join root channel
        int cmdid = TT_DoJoinChannelByID(ttInst, 
                                         TT_GetRootChannelID(ttInst), 
                                         _W(QString("")));
        if(cmdid>0)
            m_commands.insert(cmdid, CMD_COMPLETE_JOINCHANNEL);
    }
}

void MainWindow::cmdJoinedChannel(int channelid)
{
    //update list of files in current channel
    updateChannelFiles(TT_GetMyChannelID(ttInst));
    //show channel information in chat window
    ui.chatEdit->joinedChannel(channelid);
}

void MainWindow::addStatusMsg(const QString& msg)
{
    if(ttSettings->value(SETTINGS_DISPLAY_LOGSTATUSBAR).toBool(), true)
    {
        ui.chatEdit->addLogMessage(msg);
        ui.desktopchatEdit->addLogMessage(msg);
    }
    m_statusmsg.enqueue(msg);

    //just erase status messages if there's too many
    while(m_statusmsg.size()>10)
        m_statusmsg.dequeue();

    if(!timerExists(TIMER_STATUSMSG))
        m_timers[startTimer(1000)] = TIMER_STATUSMSG;
}

void MainWindow::Connect()
{
    Q_ASSERT((TT_GetFlags(ttInst) & CLIENT_CONNECTION) == 0);

    //Restart sound system so we have the latest sound devices
    TT_RestartSoundSystem();

    int inputid = getSelectedSndInputDevice();
    bool init_indev = (TT_GetFlags(ttInst) & CLIENT_SNDINPUT_READY) == 0 &&
                      inputid != SOUNDDEVICEID_NODEVICE;

    int outputid = getSelectedSndOutputDevice();
    bool init_outdev = (TT_GetFlags(ttInst) & CLIENT_SNDOUTPUT_READY) == 0 &&
                       outputid != SOUNDDEVICEID_NODEVICE;

    bool snd_init_ok = true;

    if(ttSettings->value(SETTINGS_SOUND_DUPLEXMODE, DEFAULT_SOUND_DUPLEXMODE).toBool())
    {
        if(init_indev && init_outdev &&
           !TT_InitSoundDuplexDevices(ttInst, inputid, outputid))
        {
            addStatusMsg(tr("Failed to initialize sound duplex mode"));
            snd_init_ok = false;
        }
    }
    else
    {
        if(init_indev && !TT_InitSoundInputDevice(ttInst, inputid))
        {
            addStatusMsg(tr("Failed to initialize sound input device"));
            snd_init_ok = false;
        }
        if(init_outdev && !TT_InitSoundOutputDevice(ttInst, outputid))
        {
            addStatusMsg(tr("Failed to initialize sound output device"));
            snd_init_ok = false;
        }
    }

    //choose default sound devices if configuration failed
    if(!snd_init_ok)
    {
        TT_CloseSoundInputDevice(ttInst);
        TT_CloseSoundOutputDevice(ttInst);

        addStatusMsg(tr("Switching to default sound devices"));

        if(!TT_GetDefaultSoundDevices(ttInst, &inputid, &outputid))
        {
            addStatusMsg(tr("Unable to get default sound devices"));
            return;
        }

        if(!TT_InitSoundInputDevice(ttInst, inputid) ||
           !TT_InitSoundOutputDevice(ttInst, outputid))
        {
            TT_CloseSoundInputDevice(ttInst);
            TT_CloseSoundOutputDevice(ttInst);
            addStatusMsg(tr("Failed to initialize default sound devices"));
            return;
        }
    }

    //input settings
    TT_SetVoiceActivationLevel(ttInst, ui.voiceactSlider->value());
    TT_SetSoundInputGainLevel(ttInst, ui.micSlider->value());
    TT_EnableVoiceActivation(ttInst, ttSettings->value(SETTINGS_GENERAL_VOICEACTIVATED, false).toBool());
    TT_EnableDenoising(ttInst, ttSettings->value(SETTINGS_SOUND_DENOISING, DEFAULT_SOUND_DENOISING).toBool());
    TT_EnableEchoCancellation(ttInst, ttSettings->value(SETTINGS_SOUND_ECHOCANCEL, DEFAULT_SOUND_ECHOCANCEL).toBool());

    //output settings
    int volume = ui.volumeSlider->value();
    if(volume >= SOUND_VOLUME_MAX) //compensate for software gain
        volume = SOUND_VOLUME_MAX;
    TT_SetSoundOutputVolume(ttInst, volume); 
    //this call will have no use if sound duplex mode is disabled
    TT_Enable3DSoundPositioning(ttInst, ttSettings->value(SETTINGS_SOUND_3DPOSITIONING, true).toBool());

    //find host to connect to
    HostEntry host;
    GetLatestHost(0, host);
    int localtcpport = ttSettings->value(SETTINGS_CONNECTION_TCPPORT, 0).toInt();
    int localudpport = ttSettings->value(SETTINGS_CONNECTION_UDPPORT, 0).toInt();

    addStatusMsg(tr("Connecting to %1 TCP port %2 UDP port %3")
        .arg(host.ipaddr).arg(host.tcpport).arg(host.udpport));

    TT_EnablePeerToPeer(ttInst, TRANSMIT_AUDIO, ttSettings->value(SETTINGS_CONNECTION_P2PAUDIO, false).toBool());
    TT_EnablePeerToPeer(ttInst, TRANSMIT_VIDEO, ttSettings->value(SETTINGS_CONNECTION_P2PVIDEO, false).toBool());

    m_desktopaccess_entries.clear();
    GetDesktopAccessList(m_desktopaccess_entries, host.ipaddr, host.tcpport);

#ifdef ENABLE_ENCRYPTION
    if(host.encrypted)
    {
        if(!TT_Connect(ttInst, _W(host.ipaddr), host.tcpport,
                       host.udpport, localtcpport, localudpport))
            addStatusMsg(tr("Failed to connect to encrypted server"));
    }
    else if(!TT_ConnectNonEncrypted(ttInst, _W(host.ipaddr), host.tcpport, host.udpport, 
                                    _W(QString()),localtcpport, localudpport))
    {
        addStatusMsg(tr("Failed to connect to server"));
    }
#else
    if(!TT_Connect(ttInst, _W(host.ipaddr), host.tcpport, host.udpport, 
                   localtcpport, localudpport))
    {
        addStatusMsg(tr("Failed to connect to server"));
    }
#endif
}

void MainWindow::Disconnect()
{
    TT_Disconnect(ttInst);

    if(TT_GetFlags(ttInst) & CLIENT_SNDINOUTPUT_DUPLEX)
        TT_CloseSoundDuplexDevices(ttInst);
    else
    {
        TT_CloseSoundInputDevice(ttInst);
        TT_CloseSoundOutputDevice(ttInst);
    }

    ui.channelsWidget->reset();
    ui.videogridWidget->ResetGrid();
    ui.desktopgridWidget->ResetGrid();

    m_nicknames.clear();
    m_vid_exclude.clear();
    m_videostreams.clear();

    m_desktopaccess_entries.clear();

    updateWindowTitle();

    ZERO_STRUCT(m_srvprop);
    ZERO_STRUCT(m_mychannel);

    if(m_sysicon)
        m_sysicon->setIcon(QIcon(APPTRAYICON));
}

void MainWindow::showTTErrorMessage(int error_no, CommandComplete cmd_type)
{
    QString title = tr("Error");
    QString textmsg;
    switch(error_no)
    {
        // command errors
    case CMDERR_SYNTAX_ERROR :
    case CMDERR_UNKNOWN_COMMAND :
    case CMDERR_INCOMPATIBLE_PROTOCOLS :
    case CMDERR_MISSING_PARAMETER :
        textmsg = tr("This client is not compatible with the server, "
                    " so the action cannot be performed."); break;

        // command errors due to rights
    case CMDERR_INCORRECT_SERVER_PASSWORD :
        {
            HostEntry host;
            GetLatestHost(0, host);
            bool ok = false;
            host.srvpasswd = QInputDialog::getText(this, tr("Login error"), 
                tr("Incorrect server password. Try again:"), 
                QLineEdit::Password, host.srvpasswd, &ok);
            if(!ok)
                return;
            AddLatestHost(host);
            int cmdid = TT_DoLogin(ttInst, 
                _W(ttSettings->value(SETTINGS_GENERAL_NICKNAME).toString()),
                _W(host.srvpasswd), _W(host.username), _W(host.password));
            if(cmdid>0)
                m_commands.insert(cmdid, CMD_COMPLETE_LOGIN);            
            return;
        }
    case CMDERR_INVALID_ACCOUNT :
        {
            HostEntry host;
            GetLatestHost(0, host);
            bool ok = false;
            host.username = QInputDialog::getText(this, tr("Login error"), 
                tr("Invalid user account. Type username:"), 
                QLineEdit::Normal, host.username, &ok);
            if(!ok)
                return;
            host.password = QInputDialog::getText(this, tr("Login error"), 
                tr("Invalid user account. Type password:"), 
                QLineEdit::Password, host.password, &ok);
            if(!ok)
                return;
            AddLatestHost(host);
            int cmdid = TT_DoLogin(ttInst, 
                _W(ttSettings->value(SETTINGS_GENERAL_NICKNAME).toString()),
                _W(host.srvpasswd), _W(host.username), _W(host.password));
            if(cmdid>0)
                m_commands.insert(cmdid, CMD_COMPLETE_LOGIN);            
            return;
        }
    case CMDERR_INCORRECT_CHANNEL_PASSWORD :
        {
            bool ok = false;
            QString passwd = QInputDialog::getText(this, tr("Join channel error"), 
                tr("Incorrect channel password. Try again:"), 
                QLineEdit::Password, _Q(m_last_channel.szPassword), &ok);
            if(!ok)
                return;
            COPY_TTSTR(m_last_channel.szPassword, passwd);
            int cmdid = TT_DoJoinChannel(ttInst, &m_last_channel);
            if(cmdid>0)
                m_commands.insert(cmdid, CMD_COMPLETE_JOINCHANNEL);            
            return;
        }
    case CMDERR_SERVER_BANNED :
        title = tr("Login error");
        textmsg = tr("Banned from server"); break;
    case CMDERR_NOT_AUTHORIZED :
        textmsg = tr("Command not authorized");    break;
    case CMDERR_CANNOT_CREATE_CHANNELS :
        textmsg = tr("Channel creation not allowed");    break;
    case CMDERR_MAX_SERVER_USERS_EXCEEDED :
        title = tr("Login error");
        textmsg = tr("Maximum number of users on server exceeded"); break;
    case CMDERR_MAX_DISKUSAGE_EXCEEDED :
        textmsg = tr("Maximum disk usage exceeded"); break;
    case CMDERR_MAX_CHANNEL_USERS_EXCEEDED :
        textmsg = tr("Maximum number of users in channel exceeded"); break;
    case CMDERR_SUBSCRIPTIONS_DISABLED:
        textmsg = tr("Modifying subscriptions is not allowed"); break;
    case CMDERR_INCORRECT_OP_PASSWORD :
        textmsg = tr("Incorrect channel operator password"); break;

        // state errors
    case CMDERR_ALREADY_LOGGEDIN :
        textmsg = tr("Already logged in");break;
    case CMDERR_NOT_LOGGEDIN :
        textmsg = tr("Cannot perform action because client is currently not logged in");break;
    case CMDERR_ALREADY_IN_CHANNEL :
        title = MENUTEXT(ui.actionJoinChannel->text());        
        textmsg = tr("Cannot join the same channel twice");break;
    case CMDERR_CHANNEL_ALREADY_EXISTS :
        textmsg = tr("Channel already exists");break;
    case CMDERR_USER_NOT_FOUND :
        //unsubscribe is also used to close streams, so just ignore if
        //user doesn't exist.
        if(cmd_type != CMD_COMPLETE_SUBSCRIBE &&
           cmd_type != CMD_COMPLETE_UNSUBSCRIBE)
            textmsg = tr("User not found");
        break;
    case CMDERR_CHANNEL_NOT_FOUND :
        textmsg = tr("Channel not found");break;
    case CMDERR_BAN_NOT_FOUND :
        textmsg = tr("Banned user not found"); break;
    case CMDERR_FILETRANSFER_NOT_FOUND :
        textmsg = tr("File transfer not found"); break;
    case CMDERR_ACCOUNT_NOT_FOUND :
        textmsg = tr("User account not found"); break;
    case CMDERR_FILE_NOT_FOUND :
        textmsg = tr("File not found"); break;
    case CMDERR_FILE_ALREADY_EXISTS :
        textmsg = tr("File already exists"); break;
    case CMDERR_FILESHARING_DISABLED :
        textmsg = tr("File sharing is disabled"); break;
    case CMDERR_SERVER_HAS_USERS :
        textmsg = tr("Server has active users"); break;
    case CMDERR_CHANNEL_HAS_USERS :
        textmsg = tr("Channel has active users"); break;
    default :
        {
            TTCHAR msgBuf[TT_STRLEN] = {0};
            //this will get an error message in English
            TT_GetErrorMessage(error_no, msgBuf);
            textmsg = _Q(msgBuf);
            if(textmsg.isEmpty())
                textmsg = tr("Unknown error occured");
            break;
        }
    }

    if(textmsg.size() && error_no < 10000)
    {
        textmsg = tr("The server reported an error:") + "\r\n\r\n" + textmsg;
        QMessageBox::critical(this, title, textmsg);
    }
}

#ifdef Q_OS_LINUX
void MainWindow::keysActive(quint32 keycode, quint32 mods, bool active)
{
    keycomp_t comp;
    if(mods & ControlMask)
        comp.insert(ControlMask);
    if(mods & Mod1Mask)
        comp.insert(Mod1Mask);
    if(mods & ShiftMask)
        comp.insert(ShiftMask);
    comp.insert(keycode);

    reghotkeys_t::const_iterator ite = m_hotkeys.begin();
    while(ite != m_hotkeys.end())
    {
        const keycomp_t& curcomp = ite.value();
        
        if(keycomp_t(comp).intersect(curcomp) == curcomp)
            hotkeyToggle(ite.key(), active);
        ite++;
    }
}
#endif

void MainWindow::hotkeyToggle(HotKeyID id, bool active)
{
    switch(id)
    {
    case HOTKEY_PUSHTOTALK :
        TT_EnableTransmission(ttInst, TRANSMIT_AUDIO, active);
        emit(userTalking(TT_GetMyUserID(ttInst), 
                         TT_IsTransmitting(ttInst, TRANSMIT_AUDIO)));
        PlaySoundEvent(SOUNDEVENT_HOTKEY);
        break;
    case HOTKEY_VOICEACTIVATION :
        if(active)
            TT_EnableVoiceActivation(ttInst, 
                     !(TT_GetFlags(ttInst) & CLIENT_SNDINPUT_VOICEACTIVATED));
        break;
    case HOTKEY_INCVOLUME :
        if(active)
            ui.volumeSlider->setValue(ui.volumeSlider->value()+15);
        break;
    case HOTKEY_DECVOLUME :
        if(active)
            ui.volumeSlider->setValue(ui.volumeSlider->value()-15);
        break;
    case HOTKEY_MUTEALL :
        if(active)
            TT_SetSoundOutputMute(ttInst, 
                                  !(TT_GetFlags(ttInst) & CLIENT_SNDOUTPUT_MUTE));
        break;
    case HOTKEY_MICROPHONEGAIN_INC :
        if(active)
            ui.micSlider->setValue(TT_GetSoundInputGainLevel(ttInst)+200);
        break;
    case HOTKEY_MICROPHONEGAIN_DEC :
        if(active)
            ui.micSlider->setValue(TT_GetSoundInputGainLevel(ttInst)-200);
        break;
    case HOTKEY_VIDEOTX :
        if(active)
            TT_EnableTransmission(ttInst, TRANSMIT_VIDEO, 
                                  !TT_IsTransmitting(ttInst, TRANSMIT_VIDEO));
        break;
    }
}

void MainWindow::timerEvent(QTimerEvent *event)
{
    QMainWindow::timerEvent(event);

    timers_t::iterator ite = m_timers.find(event->timerId());
    Q_ASSERT(ite != m_timers.end());
    switch(*ite)
    {
    case TIMER_PROCESS_TTEVENT :
    {
#if !defined(Q_OS_WIN32)
        TTMessage msg;
        int wait_ms = 0;
        while(TT_GetMessage(ttInst, &msg, &wait_ms))
            processTTMessage(msg);
#endif
    }
    break;
    case TIMER_ONE_SECOND :
        if(TT_GetFlags(ttInst) & CLIENT_CONNECTED)
        {
            ClientStatistics stats;
            TT_GetStatistics(ttInst, &stats);
            float rx = (float)(stats.nUdpBytesRecv - m_clientstats.nUdpBytesRecv);
            float tx = (float)(stats.nUdpBytesSent - m_clientstats.nUdpBytesSent);
            int ping = stats.nUdpPingTimeMs;
            m_clientstats = stats;

            QString status;
            status.sprintf("RX: %.2fKB TX: %.2fKB", rx/1024.0f, tx/1024.0f);

            if(ping != -1)
                m_pinglabel->setText(QString("PING: %1").arg(ping));
            else
                m_pinglabel->setText("");

#if defined(Q_OS_WIN32)
            //if(m_pttKeys.size())
            //    status = "Push-To-Talk: " + m_pttKeys + "    " + status;
#endif

            //don't show in status if there's status messages
            if(!timerExists(TIMER_STATUSMSG))
                ui.statusbar->showMessage(status);
        }

        //check if desktop transfer has been closed
        if((TT_GetFlags(ttInst) & CLIENT_DESKTOP_ACTIVE) == 0)
        {
            m_desktopsession_id = 0;
            m_prev_desktopsession_id = 0;
            m_desktopsession_total = 0;
            m_desktopsession_remain = 0;
        }

        //only show desktop transfer it's continuesly active
        if(m_prev_desktopsession_id)
        {
            float progress = 100.f;
            if(m_prev_desktopsession_id == m_desktopsession_id &&
               m_desktopsession_total)
                progress = 100.0f * (m_desktopsession_total - m_desktopsession_remain) 
                           / (float)m_desktopsession_total;

            m_dtxprogress->setValue((int)progress);

            if(m_dtxprogress->isHidden())
            {
                m_dtxlabel->show();
                m_dtxprogress->show();
            }
            if(m_prev_desktopsession_id != m_desktopsession_id ||
               m_desktopsession_remain == 0)
                m_prev_desktopsession_id = 0;
        }
        else
        {
            if(m_dtxprogress->isVisible())
            {
                m_dtxlabel->hide();
                m_dtxprogress->hide();
            }
            else if(m_desktopsession_remain>0)
                m_prev_desktopsession_id = m_desktopsession_id;
        }

        if(TT_GetFlags(ttInst) & CLIENT_AUTHORIZED)
        {
            //change to away status if idle-time has been exceeded
            int idle_time = ttSettings->value(SETTINGS_GENERAL_AUTOAWAY).toInt();
            if(idle_time != 0)
            {
                QString statusmsg = ttSettings->value(SETTINGS_GENERAL_STATUSMESSAGE).toString();
                if(IsComputerIdle(idle_time) && (m_statusmode & STATUSMODE_AWAY) == 0)
                {
                    m_statusmode |= STATUSMODE_AWAY;
                    TT_DoChangeStatus(ttInst, m_statusmode, _W(statusmsg));
                    m_idled_out = true;
                }
                else if(m_idled_out && !IsComputerIdle(idle_time))
                {
                    m_statusmode &= ~STATUSMODE_AWAY;
                    TT_DoChangeStatus(ttInst, m_statusmode, _W(statusmsg));
                    m_idled_out = false;
                }
            }
        }
        break;
    case TIMER_UI_UPDATE :
    {
        if(TT_GetFlags(ttInst) & CLIENT_SNDINPUT_READY)
        {
            int voicelevel = TT_GetSoundInputLevel(ttInst);
            if(ui.voiceactBar->value() != voicelevel)
                ui.voiceactBar->setValue(voicelevel);
        }
    
        if((TT_GetFlags(ttInst) & CLIENT_DESKTOP_ACTIVE) &&
           ttSettings->value(SETTINGS_GENERAL_DESKTOPSHARE_CURSOR, false).toBool())
            sendDesktopCursor();
        break;
    }
    case TIMER_RECONNECT :
        Q_ASSERT( (TT_GetFlags(ttInst) & CLIENT_CONNECTED) == 0);
        Disconnect();
        Connect();
        break;
    case TIMER_STATUSMSG :
        if(m_statusmsg.size())
        {
            QString status = m_statusmsg.dequeue();
            ui.statusbar->showMessage(status);
            if(m_statusmsg.empty())
            {
                killTimer(event->timerId());
                m_timers.erase(ite);
                m_timers[startTimer(2000)] = TIMER_STATUSMSG;
            }
        }
        else
        {
            killTimer(event->timerId());
            m_timers.erase(ite);
        }
        break;
    case TIMER_SEND_DESKTOPWINDOW :
        if((TT_GetFlags(ttInst) & CLIENT_TX_DESKTOP) == 0)
        {
            //only update desktop if there's users in the channel
            //(save bandwidth)
            int user_cnt = 0;
            if(TT_GetChannelUsers(ttInst, TT_GetMyChannelID(ttInst),
                                  NULL, &user_cnt) && (user_cnt > 1 || 
               (m_mychannel.uChannelType & CHANNEL_ECHO_DESKTOP)))
                sendDesktopWindow();
        }
        else
            m_desktopsend_on_completion = true;
        break;
    default :
        Q_ASSERT(0);
    }
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    ttSettings->setValue(SETTINGS_SOUND_MASTERVOLUME, ui.volumeSlider->value());
    ttSettings->setValue(SETTINGS_SOUND_MICROPHONEGAIN, ui.micSlider->value());
    ttSettings->setValue(SETTINGS_SOUND_VOICEACTIVATIONLEVEL, ui.voiceactSlider->value());

    ttSettings->setValue(SETTINGS_GENERAL_PUSHTOTALK, ui.actionEnablePushToTalk->isChecked());
    ttSettings->setValue(SETTINGS_GENERAL_VOICEACTIVATED, ui.actionEnableVoiceActivation->isChecked());

    if(windowState() == Qt::WindowNoState)
    {
        QRect r = geometry();
        QVariantList windowpos;
        windowpos.push_back(r.x());
        windowpos.push_back(r.y());
        windowpos.push_back(r.width());
        windowpos.push_back(r.height());
        ttSettings->setValue(SETTINGS_DISPLAY_WINDOWPOS, windowpos);
        ttSettings->setValue(SETTINGS_DISPLAY_SPLITTER, ui.splitter->saveState());
        ttSettings->setValue(SETTINGS_DISPLAY_DESKTOPSPLITTER, ui.desktopsplitter->saveState());
    }

    ttSettings->setValue(SETTINGS_DISPLAY_FILESHEADER, ui.filesView->header()->saveState());

    return QMainWindow::closeEvent(event);
}

void MainWindow::changeEvent(QEvent* event )
{
    if(event->type() == QEvent::WindowStateChange)
    {
        if(ttSettings->value(SETTINGS_DISPLAY_TRAYMINIMIZE, false).toBool())
        {
            if(isMinimized())
            {
                if(!m_sysicon)
                {
                    m_sysicon = new QSystemTrayIcon(this);
                    connect(m_sysicon, SIGNAL(activated(QSystemTrayIcon::ActivationReason)),
                        SLOT(slotTrayIconChange(QSystemTrayIcon::ActivationReason)));
                    m_sysmenu = new QMenu(this);
                    QAction* restore = new QAction(tr("&Restore"), m_sysmenu);
                    connect(restore, SIGNAL(triggered()), SLOT(showNormal()));
                    QAction* exit = new QAction(tr("&Exit"), m_sysmenu);
                    connect(exit, SIGNAL(triggered(bool)), SLOT(slotClientExit(bool)));
                    m_sysmenu->addAction(restore);
                    m_sysmenu->addSeparator();
                    m_sysmenu->addAction(exit);
                    m_sysicon->setContextMenu(m_sysmenu);
                    m_sysicon->setToolTip(APPTITLE);
                }
                //choose icon to show in sys tray
                QString rc_icon;
                if(m_talking.size())
                    rc_icon = APPTRAYICON_ACTIVE;
                else if(TT_GetFlags(ttInst) & CLIENT_CONNECTED)
                    rc_icon = APPTRAYICON_CON;
                else
                    rc_icon = APPTRAYICON;
                m_sysicon->setIcon(QIcon(rc_icon));

                //ensure application doesn't close when QSystemTrayIcon is active
                //(Qt is weird! Seems it doesn't distinguish 'closed' and 'hidden')
                QApplication::setQuitOnLastWindowClosed(false);
                m_sysicon->show();

                QTimer::singleShot(0, this, SLOT(hide()));
            }
            else if(m_sysicon)
            {
                QApplication::setQuitOnLastWindowClosed(true);
                m_sysicon->hide();
            }
        }
    }
    QMainWindow::changeEvent(event);
}

#if defined(Q_OS_WIN32) && QT_VERSION >= 0x050000

bool MainWindow::nativeEvent(const QByteArray& eventType, void* message,
                             long* result)
{
    MSG* msg = reinterpret_cast<MSG*>(message);
    if(msg->message >= WM_TEAMTALK_DUMMY_FIRST &&
        msg->message <= WM_TEAMTALK_DUMMY_LAST)
    {
        TTMessage ttmsg;
        ttmsg.wmMsg = (ClientEvent)msg->message;
        ttmsg.wParam = msg->wParam;
        ttmsg.lParam = msg->lParam;
        processTTMessage(ttmsg);
        //can't return here since WM_TEAMTALK_HOTKEY_TEST might be posted
        //to a different HWND
    }
    return QMainWindow::nativeEvent(eventType, message, result);
}
#endif

void MainWindow::updateWindowTitle()
{
    if(m_mychannel.nChannelID > 0 &&
       m_mychannel.nChannelID != TT_GetRootChannelID(ttInst))
        setWindowTitle(QString("%1 - %2").arg(_Q(m_mychannel.szName)).arg(APPTITLE));
    else
        setWindowTitle(APPTITLE);
}

void MainWindow::subscribeCommon(bool checked, Subscriptions subs, int userid/* = 0*/)
{
    if((m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS) == SUBSCRIBE_NONE)
        return;

    if(userid == 0)
        userid = ui.channelsWidget->selectedUser();
    if(!userid)return;

    if(checked)
    {
        int cmdid = TT_DoSubscribe(ttInst, userid, subs);
        if(cmdid>0)
            m_commands[cmdid] = CMD_COMPLETE_SUBSCRIBE;
    }
    else
    {
        int cmdid = TT_DoUnsubscribe(ttInst, userid, subs);
        if(cmdid>0)
            m_commands[cmdid] = CMD_COMPLETE_UNSUBSCRIBE;
    }
}

TextMessageDlg* MainWindow::getTextMessageDlg(int userid)
{
    usermsg_t::iterator ite = m_usermsg.find(userid);
    if(ite != m_usermsg.end())
        return *ite;
    else
    {
        TextMessageDlg* dlg;
        usermessages_t::iterator ii = m_usermessages.find(userid);
        if(ii != m_usermessages.end())
        {
            dlg = new TextMessageDlg(userid, ii.value(), 0);
            m_usermessages.erase(ii);
        }
        else
            dlg = new TextMessageDlg(userid, 0);

        dlg->setAttribute(Qt::WA_DeleteOnClose);
        m_usermsg.insert(userid, dlg);
        connect(dlg, SIGNAL(closedTextMessage(int)), SLOT(slotTextMessageClosed(int)));
        connect(this, SIGNAL(userUpdate(int,int)), dlg, SLOT(slotUpdateUser(int,int)));
        connect(this, SIGNAL(newTextMessage(int,int)), dlg, SLOT(slotNewMessage(int,int)));
        connect(this, SIGNAL(userLogout(int)), dlg, SLOT(slotUserLogout(int)));
        return dlg;
    }
}

void MainWindow::processTextMessage(int userid, int msgid)
{
    TextMessage textmsg;
    //don't clear message since a user text-message will be retrived
    //in its own dialog
    if(!TT_GetTextMessage(ttInst, msgid, FALSE, &textmsg))
        return;

    switch(textmsg.nMsgType)
    {
    case MSGTYPE_CHANNEL :
        PlaySoundEvent(SOUNDEVENT_CHANNELMSG);
    case MSGTYPE_BROADCAST :
        ui.chatEdit->addTextMessage(textmsg);
        ui.desktopchatEdit->addTextMessage(textmsg);
        TT_GetTextMessage(ttInst, msgid, TRUE, NULL);
        break;
    case MSGTYPE_USER :
        if(ttSettings->value(SETTINGS_DISPLAY_MESSAGEPOPUP, true).toBool())
        {
            TextMessageDlg* dlg = getTextMessageDlg(userid);
            if(dlg)
                dlg->show();
        }
        ui.channelsWidget->setUserMessaged(userid, true);
        emit(newTextMessage(userid, msgid));
        PlaySoundEvent(SOUNDEVENT_USERMSG);

        //store text message if a TextMessageDlg didn't extract it
        if(TT_GetTextMessage(ttInst, msgid, TRUE, NULL))
        {
            usermessages_t::iterator ii = m_usermessages.find(userid);
            if(ii != m_usermessages.end())
                ii.value().push_back(textmsg);
            else
            {
                textmessages_t msgs;
                msgs.push_back(textmsg);
                m_usermessages.insert(userid, msgs);
            }
        }
        break;
    case MSGTYPE_CUSTOM :
    {
        emit(newTextMessage(userid, msgid));

        QStringList cmd = getCustomCommand(textmsg);
        if(cmd.size() < 2)
            break;
        
        if(cmd[0] == TT_INTCMD_DESKTOP_ACCESS)
        {
            User user;
            if(!ui.channelsWidget->getUser(userid, user))
                break;

            ui.channelsWidget->setUserDesktopAccess(userid, cmd[1] == "1");
            if(cmd[1] == "1")
            {
                addStatusMsg(QString(tr("%1 is requesting desktop access").arg(_Q(user.szNickname))));
                PlaySoundEvent(SOUNDEVENT_DESKTOPACCESS);
                if(hasDesktopAccess(m_desktopaccess_entries, user))
                {
                    subscribeCommon(true, SUBSCRIBE_DESKTOPINPUT, user.nUserID);
                    addStatusMsg(QString(tr("%1 granted desktop access").arg(_Q(user.szNickname))));
                }
            }
            else
            {
                addStatusMsg(QString(tr("%1 retracted desktop access").arg(_Q(user.szNickname))));
                subscribeCommon(false, SUBSCRIBE_DESKTOPINPUT, user.nUserID);
            }
        }
    }
    break;
    }
}

void MainWindow::processMyselfJoined(int channelid)
{
    TT_GetChannel(ttInst, channelid, &m_mychannel);

    TTCHAR buff[TT_STRLEN] = {0};
    TT_GetChannelPath(ttInst, channelid, buff);
    addStatusMsg(tr("Joined channel %1").arg(_Q(buff)));

    //store new muxed audio file if we're changing channel
    if(ui.actionEnableStoreAudioToDisk->isChecked() &&
        (m_audiostorage_mode & AUDIOSTORAGE_SINGLEFILE))
    {
        updateAudioStorage(false, AUDIOSTORAGE_SINGLEFILE);
        updateAudioStorage(true, AUDIOSTORAGE_SINGLEFILE);
    }

    updateWindowTitle();
}

void MainWindow::processMyselfLeft(int channelid)
{
    Q_UNUSED(channelid);
    ZERO_STRUCT(m_mychannel);

    m_talking.clear();
    ui.videogridWidget->ResetGrid();
    ui.desktopgridWidget->ResetGrid();

    updateWindowTitle();
}

void MainWindow::processUserJoined(int userid, int channelid)
{
    User user;
    if(TT_GetUser(ttInst, userid, &user))
        m_nicknames[userid] = _Q(user.szNickname);

    //also set here in case VIEW_ALL_USERS is false
    QString audiofolder = ttSettings->value(SETTINGS_GENERAL_AUDIOSTORAGE_FOLDER).toString();
    AudioFileFormat aff = (AudioFileFormat)ttSettings->value(SETTINGS_GENERAL_AUDIOSTORAGE_FILEFORMAT, AFF_WAVE_FORMAT).toInt();
    if(m_audiostorage_mode & AUDIOSTORAGE_SEPARATEFILES)
        TT_SetUserAudioFolder(ttInst, userid, _W(audiofolder), NULL, aff);

    if(user.nStatusMode & STATUSMODE_STREAM_MEDIAFILE)
        TT_SetUserMediaBufferSize(ttInst, user.nUserID, STREAMMEDIA_AUDIOBUF_SIZE);
    else
        TT_SetUserMediaBufferSize(ttInst, user.nUserID, 0);

    //only play sound when we're not currently performing an operation
    //like e.g. joining a new channel
    if(m_mychannel.nChannelID == channelid && m_current_cmdid == 0)
    {
        PlaySoundEvent(SOUNDEVENT_NEWUSER);
        addStatusMsg(tr("%1 joined channel").arg(_Q(user.szNickname)));
    }
}

void MainWindow::processUserLeft(int userid, int channelid)
{
    if(m_mychannel.nChannelID == channelid && m_current_cmdid == 0)
    {
        PlaySoundEvent(SOUNDEVENT_REMOVEUSER);
        //we need to find the user in the nickname cache since the user
        //might not exist anymore
        nicknames_t::iterator ite = m_nicknames.find(userid);
        if(ite != m_nicknames.end())
            addStatusMsg(tr("%1 left channel").arg(*ite));
    }

    //we cannot get user from channels-widget since user has left channel
    User user;
    if(m_mychannel.nChannelID == channelid &&
       TT_GetUser(ttInst, userid, &user) &&
       (user.uLocalSubscriptions & SUBSCRIBE_DESKTOPINPUT))
       subscribeCommon(false, SUBSCRIBE_DESKTOPINPUT, userid);
}

int MainWindow::getVideoStreamID(int userid) const
{
    videostreams_t::const_iterator streamIte = m_videostreams.find(userid);
    if(streamIte != m_videostreams.end())
        return streamIte.value();
    return 0;
}

void MainWindow::killLocalTimer(TimerEvent e)
{
    timers_t::iterator ite = m_timers.begin();
    while(ite != m_timers.end())
    {
        if(*ite == e)
        {
            killTimer(ite.key());
            m_timers.erase(ite);
            break;
        }
        ite++;
    }
}

bool MainWindow::timerExists(TimerEvent e)
{
    timers_t::iterator ite = m_timers.begin();
    while(ite != m_timers.end())
    {
        if(*ite == e)
            return true;
        ite++;
    }
    return false;
}

void MainWindow::updateChannelFiles(int channelid)
{
    m_filesmodel->slotChannelUpdated(channelid);
    TTCHAR chanpath[TT_STRLEN] = {0};
    TT_GetChannelPath(ttInst, channelid, chanpath);
    ui.channelLabel->setText(tr("Files in channel: %1").arg(_Q(chanpath)));

    if(m_filesmodel->rowCount() == 0)
        ui.tabWidget->setTabText(TAB_FILES, tr("Files"));
    else
        ui.tabWidget->setTabText(TAB_FILES, tr("Files (%1)").arg(m_filesmodel->rowCount()));
}

void MainWindow::updateUserGainLevel(int userid)
{
    int volume = ui.volumeSlider->value();
    int gain = gainLevel(ui.volumeSlider->value());

    QVector<int> users;
    if(userid == -1)
        users = ui.channelsWidget->getUsers();
    else
        users.push_back(userid);

    for(int i=0;i<users.size();i++)
    {
        if(volume <= SOUND_VOLUME_MAX)
            //disable soft gain
            TT_SetUserGainLevel(ttInst, users[i], SOUND_GAIN_DEFAULT);
        else
            TT_SetUserGainLevel(ttInst, users[i], gain);
    }
}

void MainWindow::updateUserSubscription(int userid)
{
    if((m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS) == USERRIGHT_NONE)
        return;

    Subscriptions unsub = SUBSCRIBE_NONE;
    if(!ttSettings->value(SETTINGS_CONNECTION_SUBSCRIBE_USERMSG, true).toBool())
        unsub |= SUBSCRIBE_USER_MSG;
    if(!ttSettings->value(SETTINGS_CONNECTION_SUBSCRIBE_CHANNELMSG, true).toBool())
        unsub |= SUBSCRIBE_CHANNEL_MSG;
    if(!ttSettings->value(SETTINGS_CONNECTION_SUBSCRIBE_BROADCASTMSG, true).toBool())
        unsub |= SUBSCRIBE_BROADCAST_MSG;
    if(!ttSettings->value(SETTINGS_CONNECTION_SUBSCRIBE_AUDIO, true).toBool())
        unsub |= SUBSCRIBE_AUDIO;
    if(!ttSettings->value(SETTINGS_CONNECTION_SUBSCRIBE_VIDEO, true).toBool())
        unsub |= SUBSCRIBE_VIDEO;
    if(!ttSettings->value(SETTINGS_CONNECTION_SUBSCRIBE_DESKTOP, true).toBool())
        unsub |= SUBSCRIBE_DESKTOP;

    if(unsub)
        subscribeCommon(false, unsub, userid);
}

void MainWindow::updateAudioStorage(bool enable, AudioStorageMode mode)
{
    AudioFileFormat aff = (AudioFileFormat)ttSettings->value(SETTINGS_GENERAL_AUDIOSTORAGE_FILEFORMAT, 
                                                AFF_WAVE_FORMAT).toInt();
    QString audiofolder = ttSettings->value(SETTINGS_GENERAL_AUDIOSTORAGE_FOLDER).toString();

    if(mode == AUDIOSTORAGE_SINGLEFILE)
    {
        if(enable)
        {
            QString filepath = audiofolder;
            filepath += QDir::toNativeSeparators("/") + GenerateAudioStorageFilename(aff);

            if(!TT_StartRecordingMuxedAudioFile(ttInst, &m_mychannel.codec, 
                                                _W(filepath), aff))
            {
                QMessageBox::critical(this, tr("Error"), tr("Failed to start recording"));
                return;
            }
            else
                addStatusMsg(tr("Recording to file: %1").arg(filepath));
        }
        else
            TT_StopRecordingMuxedAudioFile(ttInst);
    }
    
    if(mode == AUDIOSTORAGE_SEPARATEFILES)
    {
        int userCount = 0;
        QVector<int> userids;
        TT_GetServerUsers(ttInst, NULL, &userCount);
        if(userCount)
        {
            userids.resize(userCount);
            TT_GetServerUsers(ttInst, &userids[0], &userCount);
        }
        for(int i=0;i<userCount;i++)
        {
            if(enable)
                TT_SetUserAudioFolder(ttInst, userids[i], _W(audiofolder), NULL, aff);
            else
                TT_SetUserAudioFolder(ttInst, userids[i], _W(QString()), NULL, aff);
        }
    }
}

bool MainWindow::sendDesktopWindow()
{
    switch(ttSettings->value(SETTINGS_GENERAL_DESKTOPSHARE_MODE, 
                             DESKTOPSHARE_NONE).toInt())
    {
    case DESKTOPSHARE_DESKTOP :
#if defined(Q_OS_DARWIN)
        //TODO: MacOS, entire desktop
        m_nCGShareWnd = kCGNullWindowID;
#endif
        break;
    case DESKTOPSHARE_ACTIVE_WINDOW :
#if defined(Q_OS_WIN32)
        m_hShareWnd = TT_Windows_GetDesktopActiveHWND();
#elif defined(Q_OS_LINUX)
    {
        if(!m_display)
            m_display = XOpenDisplay(0);

        if(!m_display)
            return false;

        int revert_to_return = 0;
        Window winid = 0;
        XGetInputFocus(m_display, &winid, &revert_to_return);
        m_nWindowShareWnd = winid;
    }
#elif defined(Q_OS_DARWIN)
    //TODO: MacOS, active window
        m_nCGShareWnd = kCGNullWindowID;
#endif
    break;
    case DESKTOPSHARE_SPECIFIC_WINDOW :
        break;
    default :
        return false;
    }

    BitmapFormat bmp_mode = 
        (BitmapFormat)ttSettings->value(SETTINGS_GENERAL_DESKTOPSHARE_BMPMODE, 
                                        BMP_RGB16_555).toInt();

    int ret = 0;
#if defined(Q_OS_WIN32)
    if(m_hShareWnd)
        ret = TT_SendDesktopWindowFromHWND(ttInst, m_hShareWnd, bmp_mode, DESKTOPPROTOCOL_ZLIB_1);
/*
    const int W = 100, H = 100;
    QImage img(W, H, QImage::Format_RGB32);
    for(int i=0;i<W;i++)
        img.setPixel(i, 0, 0xF000);
    for(int i=0;i<W;i++)
        img.setPixel(i, H-1, 0xF000);
    for(int i=0;i<H;i++)
        img.setPixel(0, i, 0x00F0);
    for(int i=0;i<H;i++)
        img.setPixel(W-1, i, 0x00F0);

    DesktopWindow wnd;
    ZERO_STRUCT(wnd);
    wnd.bmpFormat = BMP_RGB32;
    wnd.nProtocol = DESKTOPPROTOCOL_ZLIB_1;
    wnd.nWidth = W;
    wnd.nHeight = H;
    ret = TT_SendDesktopWindow(ttInst, img.bits(), img.byteCount(), &wnd, BMP_RGB32);
*/
#elif defined(Q_OS_DARWIN)
    if(m_nCGShareWnd)
        ret = TT_SendDesktopFromWindowID(ttInst, m_nCGShareWnd, bmp_mode, DESKTOPPROTOCOL_ZLIB_1);

#elif defined(Q_OS_LINUX)

    if(!m_display)
        return false;

    XWindowAttributes attr;
    Status s = XGetWindowAttributes(m_display, m_nWindowShareWnd, &attr);
    if(s)
    {
        XImage* img = XGetImage(m_display, m_nWindowShareWnd, 0, 0, attr.width,
                                attr.height, -1, ZPixmap);
        if(img)
        {
            DesktopWindow wnd;
            ZERO_STRUCT(wnd);
            
            wnd.nProtocol = DESKTOPPROTOCOL_ZLIB_1;
            wnd.nWidth = img->width;
            wnd.nHeight = img->height;
            wnd.nBytesPerLine = img->bytes_per_line;
            switch(img->bits_per_pixel)
            {
            case 16 :
                wnd.bmpFormat = BMP_RGB16_555;
                break;
            case 24 :
                wnd.bmpFormat = BMP_RGB24;
                break;
            case 32 :
                wnd.bmpFormat = BMP_RGB32;
                break;
            default :
                return false;
            }

            int bitmap_size = img->bytes_per_line * img->height;
            ret = TT_SendDesktopWindow(ttInst, img->data, bitmap_size, &wnd, bmp_mode);

            XDestroyImage(img);
        }
    }
#endif
    return ret >= 0;
}

void MainWindow::restartSendDesktopWindowTimer()
{
    killLocalTimer(TIMER_SEND_DESKTOPWINDOW);
    int upd_interval = ttSettings->value(SETTINGS_GENERAL_DESKTOPSHARE_INTERVAL).toInt();
    if(upd_interval>0)
        m_timers.insert(startTimer(upd_interval), TIMER_SEND_DESKTOPWINDOW);
}

QRect MainWindow::getSharedWindowRect()
{
#if defined(Q_OS_LINUX)
    //TODO: X11, get active window (this only handles entire desktop)

    if(!m_display)
        return QRect();

    XWindowAttributes attr;
    Status s = XGetWindowAttributes(m_display, m_nWindowShareWnd, &attr);
    if(s)
        return QRect(attr.x, attr.y, attr.x + attr.width, attr.y + attr.height);

#elif defined(Q_OS_DARWIN)
    ShareWindow wnd;
    ZERO_STRUCT(wnd);

    if(TT_MacOS_GetWindowFromWindowID(m_nCGShareWnd, &wnd))
        return QRect(wnd.nWindowX, wnd.nWindowY, wnd.nWindowX + wnd.nWidth,
                     wnd.nWindowY + wnd.nHeight);
#elif defined(Q_OS_WIN32)
    RECT r;
    if(::GetWindowRect(m_hShareWnd, &r))
        return QRect(r.left, r.top, r.right-r.left, r.bottom - r.top);
#endif
    return QRect();
}

void MainWindow::sendDesktopCursor()
{   
    QPoint curPos = QCursor::pos();

#if defined(Q_OS_LINUX)
    if(!m_display)
        return;
    Window root, winid;
    int x, y;
    unsigned int w, h, bwr, depth;
    XWindowAttributes attr;
    Status s = XGetWindowAttributes(m_display, m_nWindowShareWnd, &attr);

    // winid = m_nWindowShareWnd;
    // s = XGetGeometry(m_display, winid, &root, &x, &y, &w, &h, &bwr, &depth);

    // XQueryPointer(m_display, &root, 
    if(s)
    {
        int x = curPos.x() - attr.x;
        int y = curPos.y() - attr.y;
        x = qMax(-1, x);
        y = qMax(-1, y);
        curPos = QPoint(x, y);
    }
    else
    {
        curPos = QPoint(-1, -1);
    }
#elif defined(Q_OS_DARWIN)
    ShareWindow wnd;
    if(!TT_MacOS_GetWindowFromWindowID(m_nCGShareWnd, &wnd))
        return;

    int x = curPos.x() - wnd.nWindowX;
    int y = curPos.y() - wnd.nWindowY;
    if(x<0 || x>wnd.nWidth || y < 0 || y > wnd.nHeight)
        x = y = -1;
    curPos = QPoint(x, y);
#elif defined(Q_OS_WIN32)
    ShareWindow wnd;
    if(!TT_Windows_GetWindow(m_hShareWnd, &wnd))
        return;

    int x = curPos.x() - wnd.nWndX;
    int y = curPos.y() - wnd.nWndY;
    if(x<0 || x>wnd.nWidth || y < 0 || y > wnd.nHeight)
        x = y = -1;
    curPos = QPoint(x, y);
#endif

    if(m_lastCursorPos != curPos)
    {
        m_lastCursorPos = curPos;
        TT_SendDesktopCursorPosition(ttInst, 0, 
                                     m_lastCursorPos.x(), 
                                     m_lastCursorPos.y());
        
    }
}

void MainWindow::processDesktopInput(int userid)
{
    //extract as many DesktopInputs as possible
    INT32 n_count = 0;
    DesktopInput tmp_inputs[TT_DESKTOPINPUT_MAX];
    QVector<DesktopInput> inputs;
    while(TT_GetUserDesktopInput(ttInst, userid, tmp_inputs, &n_count))
    {
        for(int i=0;i<n_count;i++)
            inputs.push_back(tmp_inputs[i]);
    }

    if(inputs.isEmpty())return;

    //ignore input if we're not sharing a window
    if((TT_GetFlags(ttInst) & CLIENT_DESKTOP_ACTIVE) == 0)
        return;
/*
    //ignore input if incorrect window has focus
    DesktopShareMode mode = (DesktopShareMode)ttSettings->value(SETTINGS_GENERAL_DESKTOPSHARE_MODE, 
                             DESKTOPSHARE_NONE).toInt();
    switch(mode)
    {
    case DESKTOPSHARE_SPECIFIC_WINDOW :
#if defined(Q_OS_WIN32)
        if(m_hShareWnd != TT_Windows_GetDesktopActiveHWND())
        {
            qDebug() << "Ignoring desktop input. Incorrect window has focus.";
            return;
        }
#elif defined(Q_OS_LINUX)
     {
        if(!m_display)
            return;

        int revert_to_return = 0;
        Window winid = 0;
        XGetInputFocus(m_display, &winid, &revert_to_return);
        if(m_nWindowShareWnd != winid)
        {
            qDebug() << "Ignoring desktop input. Incorrect window has focus.";
            return;
        }
     }
#else
        qDebug() << "Ignoring desktop input. Share specific window not supported on Mac OS X";
        return;
#endif
     break;
    case DESKTOPSHARE_ACTIVE_WINDOW :
    case DESKTOPSHARE_DESKTOP :
    default :
        break;
    }
*/
    TTKeyTranslate key_trans = TTKEY_NO_TRANSLATE;
#if ENABLE_KEY_TRANSLATION
#if defined(Q_OS_WIN32)
    key_trans = TTKEY_TTKEYCODE_TO_WINKEYCODE;
#elif defined(Q_OS_DARWIN)
    key_trans = TTKEY_TTKEYCODE_TO_MACKEYCODE;
#elif defined(Q_OS_LINUX)
//TODO: X11, key translate for X11
#endif
#endif /* ENABLE_KEY_TRANSLATION */

    QVector<DesktopInput> executeInputs;

    QRect offset = getSharedWindowRect();
    for(int i=0;i<inputs.size();i++)
    {
        //calculate absolute offset
        if(inputs[i].uMousePosX != TT_DESKTOPINPUT_MOUSEPOS_IGNORE &&
           inputs[i].uMousePosY != TT_DESKTOPINPUT_MOUSEPOS_IGNORE)
        {
            //don't allow mouse input outside shared window
            if(inputs[i].uMousePosX > offset.width() ||
               inputs[i].uMousePosY > offset.height())
                continue;

            inputs[i].uMousePosX += offset.left();
            inputs[i].uMousePosY += offset.top();

        }
        //QDebug(QtDebugMsg) << "DesktopInput " << inputs[i].uMousePosX << "," << inputs[i].uMousePosY
        //                   << inputs[i].uKeyCode << inputs[i].uKeyState;

        //QDebug dbg2(QtDebugMsg);
        //dbg2 << "Mouse input: ";
        //dbg2 << input.mouseInputs[i].nPosX << "," << input.mouseInputs[i].nPosY;
        //dbg2 << "Buttons: "; 
        //for(int j=0;j<TT_DESKTOPINPUT_KEYS_MAX;j++)
        //    dbg2 << input.mouseInputs[i].buttonsDown[j] << ",";

        DesktopInput trans_input;
        ZERO_STRUCT(trans_input);

        if(key_trans == TTKEY_NO_TRANSLATE)
            executeInputs.push_back(inputs[i]);
        else if(TT_DesktopInput_KeyTranslate(key_trans, &inputs[i], &trans_input, 1))
            executeInputs.push_back(trans_input);
        else
            qDebug() << "Failed to translate received desktop input. KeyCode: 0x" 
                     << QString::number(inputs[i].uKeyCode, 16);
    }

    if(executeInputs.size())
    {
#if defined(Q_OS_LINUX)
        for(int i=0;i<executeInputs.size();i++)
            executeDesktopInput(executeInputs[i]);
#else
        TT_DesktopInput_Execute(&executeInputs[0], executeInputs.size());
#endif
        //send desktop update immediately and restart desktop tx timer.
        if(sendDesktopWindow())
            restartSendDesktopWindowTimer();
        else
            m_desktopsend_on_completion = true;
    }
}

void MainWindow::startStreamMediaFile()
{
    QString fileName = ttSettings->value(SETTINGS_GENERAL_STREAMMEDIA_FILENAME).toString();
    
    TransmitTypes tx_mode = TRANSMIT_NONE;
    if(ttSettings->value(SETTINGS_GENERAL_STREAMMEDIA_AUDIO).toBool())
        tx_mode |= TRANSMIT_AUDIO;
    if(ttSettings->value(SETTINGS_GENERAL_STREAMMEDIA_VIDEO).toBool())
        tx_mode |= TRANSMIT_VIDEO;

    VideoCodec vidcodec, *pVideoCodec = NULL;

    vidcodec.nCodec = THEORA_CODEC;
    vidcodec.theora.nQuality = ttSettings->value(SETTINGS_GENERAL_THEORA_QUALITY, DEFAULT_THEORA_QUALITY).toInt();
    vidcodec.theora.nTargetBitRate = ttSettings->value(SETTINGS_GENERAL_THEORA_BITRATE, DEFAULT_THEORA_BITRATE).toInt();

    if(tx_mode & TRANSMIT_VIDEO)
        pVideoCodec = &vidcodec;

    if(fileName.size() && 
       !TT_StartStreamingMediaFileToChannel(ttInst, _W(fileName), pVideoCodec, tx_mode))
    {
        QMessageBox::information(this,
                                 MENUTEXT(ui.actionStreamMediaFileToChannel->text()),
                                 QString(tr("Failed to stream media file %1").arg(fileName)));
        stopStreamMediaFile();
    }
    else
    {
        m_statusmode |= STATUSMODE_STREAM_MEDIAFILE;
        ////since streaming video takes over webcam stream we show as 
        ////transmitting video
        //if(tx_mode & TRANSMIT_VIDEO)
        //    m_statusmode |= STATUSMODE_VIDEOTX;

        QString statusmsg = ttSettings->value(SETTINGS_GENERAL_STATUSMESSAGE).toString();
        TT_DoChangeStatus(ttInst, m_statusmode, _W(statusmsg));
    }
}

void MainWindow::stopStreamMediaFile()
{
    TT_StopStreamingMediaFileToChannel(ttInst);

    m_statusmode &= ~STATUSMODE_STREAM_MEDIAFILE;
    ////clear video if not transmitting
    //if(TT_IsTransmitting(ttInst, TRANSMIT_VIDEO) == TRANSMIT_NONE)
    //    m_statusmode &= ~STATUSMODE_VIDEOTX;

    QString statusmsg = ttSettings->value(SETTINGS_GENERAL_STATUSMESSAGE).toString();
    TT_DoChangeStatus(ttInst, m_statusmode, _W(statusmsg));

    slotUpdateUI();
}

void MainWindow::loadHotKeys()
{
    hotkey_t hotkey;
    if(LoadHotKeySettings(HOTKEY_VOICEACTIVATION, hotkey))
        enableHotKey(HOTKEY_VOICEACTIVATION, hotkey);
    hotkey.clear();
    if(LoadHotKeySettings(HOTKEY_INCVOLUME, hotkey))
        enableHotKey(HOTKEY_INCVOLUME, hotkey);
    hotkey.clear();
    if(LoadHotKeySettings(HOTKEY_DECVOLUME, hotkey))
        enableHotKey(HOTKEY_DECVOLUME, hotkey);
    hotkey.clear();
    if(LoadHotKeySettings(HOTKEY_MUTEALL, hotkey))
        enableHotKey(HOTKEY_MUTEALL, hotkey);
    hotkey.clear();
    if(LoadHotKeySettings(HOTKEY_MICROPHONEGAIN_INC, hotkey))
        enableHotKey(HOTKEY_MICROPHONEGAIN_INC, hotkey);
    hotkey.clear();
    if(LoadHotKeySettings(HOTKEY_MICROPHONEGAIN_DEC, hotkey))
        enableHotKey(HOTKEY_MICROPHONEGAIN_DEC, hotkey);
    hotkey.clear();
    if(LoadHotKeySettings(HOTKEY_VIDEOTX, hotkey))
        enableHotKey(HOTKEY_VIDEOTX, hotkey);
}

void MainWindow::enableHotKey(HotKeyID id, const hotkey_t& hk)
{
    Q_ASSERT(hk.size());

    //disable first so we don't double register
    disableHotKey(id);

#ifdef Q_OS_WIN32
    TT_HotKey_Register(ttInst, id, &hk[0], hk.size());

#elif defined(Q_OS_LINUX)

    Display* display = QX11Info::display();
    Window x11window = QX11Info::appRootWindow();

    keycomp_t keycomp;
    quint32 mods = 0, keycode = 0;
    for(int i=0;i<hk.size();i++)
    {
        switch(hk[i])
        {
        case Qt::CTRL :
            mods |= ControlMask;
            keycomp.insert(ControlMask);
            break;
        case Qt::ALT :
            mods |= Mod1Mask;
            keycomp.insert(Mod1Mask);
            break;
        case Qt::SHIFT :
            mods |= ShiftMask;
            keycomp.insert(ShiftMask);
            break;
        default:
            keycode = XKeysymToKeycode(display, XStringToKeysym(QKeySequence(hk[i]).toString().toLatin1().data()));  
            keycomp.insert(keycode);
            break;
        }
    }

    m_hotkeys.insert(id, keycomp);
    Bool owner = True;
    int pointer = GrabModeAsync;
    int keyboard = GrabModeAsync;

    // no way to check for success
    XGrabKey(display, keycode, mods, x11window, owner, pointer, keyboard);
    // allow numlock
    XGrabKey(display, keycode, mods | Mod2Mask, x11window, owner, pointer, keyboard);

#elif defined(Q_OS_DARWIN)

    if(hk.empty() || hk.size() != MAC_HOTKEY_SIZE)
        return;

    quint32 mods = 0, keycode = 0;
    if(hk[0] != MAC_NO_KEY)
    {
        if(hk[0] & cmdKey)
            mods |= cmdKey;
        if(hk[0] & optionKey)
            mods |= optionKey;
        if(hk[0] & shiftKey)
            mods |= shiftKey;
        if(hk[0] & controlKey)
            mods |= controlKey;
    }

    keycode = hk[1];

    EventHotKeyID keyID;
    keyID.signature = 'cute';
    keyID.id = id;

    EventHotKeyRef ref = 0;
    if(RegisterEventHotKey(keycode, mods, keyID, GetApplicationEventTarget(), 0, &ref) == 0)
        m_hotkeys[id] = ref;
    else
    {
        QMessageBox::warning(this, tr("Enable HotKey"), 
                             tr("Failed to register hotkey. Please try another key combination."));
    }

#endif
}

void MainWindow::disableHotKey(HotKeyID id)
{
#ifdef Q_OS_WIN32

    TT_HotKey_Unregister(ttInst, id);

#elif defined(Q_OS_LINUX)

    Display* display = QX11Info::display();
    Window window = QX11Info::appRootWindow();

    reghotkeys_t::iterator hk_ite = m_hotkeys.find(id);
    if(hk_ite != m_hotkeys.end())
    {
        quint32 mods = 0;
        quint32 keycode = 0;
        const keycomp_t& comp = hk_ite.value();
        keycomp_t::const_iterator ite = comp.begin();
        while(ite != comp.end())
        {
            switch(*ite)
            {
            case ControlMask :
                mods |= ControlMask;
                break;
            case Mod1Mask :
                mods |= Mod1Mask;
                break;
            case ShiftMask :
                mods |= ShiftMask;
                break;
            default:
                keycode = *ite;
                break;
            }
            ite++;
        }
        XUngrabKey(display, keycode, mods, window);
        XUngrabKey(display, keycode, mods | Mod2Mask, window);
    }

    m_hotkeys.remove(id);

#elif defined(Q_OS_DARWIN)

    reghotkeys_t::iterator i = m_hotkeys.find(id);
    if(i != m_hotkeys.end())
    {
        UnregisterEventHotKey(i.value());
        m_hotkeys.erase(i);
    }

#endif    
}

#if defined(Q_OS_LINUX)
void MainWindow::executeDesktopInput(const DesktopInput& input)
{
    //TODO: X11, mouse held down for selection (e.g. text in edit field)
    //TODO: X11, key strokes

    if(input.uMousePosX != TT_DESKTOPINPUT_MOUSEPOS_IGNORE &&
       input.uMousePosY != TT_DESKTOPINPUT_MOUSEPOS_IGNORE)
    {
         XWarpPointer(m_display, None, (Window)m_nWindowShareWnd, 0, 0, 0, 0,
                      input.uMousePosX, input.uMousePosY);

        XEvent event;
        ZERO_STRUCT(event);

        long event_mask = 0;
//        event_mask |= ButtonMotionMask;
//        event_mask |= ButtonPressMask;
//        event_mask |= ButtonReleaseMask;
//        event_mask |= PointerMotionMask;
        
        switch(input.uKeyState)
        {
        case DESKTOPKEYSTATE_DOWN :
            event.type = ButtonPress;
            event.xbutton.same_screen = True;
            event.xbutton.send_event = True;
            event.xbutton.x = input.uMousePosX;
            event.xbutton.y = input.uMousePosY;
            break;
        case DESKTOPKEYSTATE_UP :
            event.type = ButtonRelease;
            event.xbutton.same_screen = True;
            event.xbutton.send_event = True;
            event.xbutton.x = input.uMousePosX;
            event.xbutton.y = input.uMousePosY;
            break;
        default :
            return;
        }

        switch(input.uKeyCode)
        {
        case TT_DESKTOPINPUT_KEYCODE_LMOUSEBTN :
            event.xbutton.button = Button1;
            break;
        case TT_DESKTOPINPUT_KEYCODE_RMOUSEBTN :
            event.xbutton.button = Button3;
            break;
        case TT_DESKTOPINPUT_KEYCODE_MMOUSEBTN :
            event.xbutton.button = Button2;
            break;
        }

        if(input.uKeyState == DESKTOPKEYSTATE_DOWN || input.uKeyState == DESKTOPKEYSTATE_UP)
        {
            XQueryPointer(m_display, (Window)m_nWindowShareWnd, &event.xbutton.root,
                          &event.xbutton.window, &event.xbutton.x_root,
                          &event.xbutton.y_root, &event.xbutton.x, &event.xbutton.y,
                          &event.xbutton.state);

            event.xbutton.subwindow = event.xbutton.window;

            while(event.xbutton.subwindow)
            {
                event.xbutton.window = event.xbutton.subwindow;

                XQueryPointer(m_display, event.xbutton.window, &event.xbutton.root,
                              &event.xbutton.subwindow, &event.xbutton.x_root,
                              &event.xbutton.y_root, &event.xbutton.x,
                              &event.xbutton.y, &event.xbutton.state);
            }
        }

        int ret = XSendEvent(m_display, PointerWindow, True, 0, &event);
        //qDebug() << "Sent event" << ret << "btn" << event.xbutton.button;
        XFlush(m_display);
    }
}
#endif

void MainWindow::slotClientNewInstance(bool /*checked=false*/)
{
    QString path = QApplication::applicationFilePath();
#if defined(_DEBUG)
    QProcess::startDetached(path, (QStringList() << "-noconnect"));
#else
    QProcess::startDetached(path, (QStringList() << "-noconnect"),
                            QApplication::applicationDirPath());
#endif
}

void MainWindow::slotClientConnect(bool /*checked =false */)
{
    killLocalTimer(TIMER_RECONNECT);

    //reset last channel, since we're starting a new connection
    ZERO_STRUCT(m_last_channel);

    if(TT_GetFlags(ttInst) & CLIENT_CONNECTION)
        Disconnect();
    else
    {
        ServerListDlg dlg(this);
        if(dlg.exec())
            Connect();
    }

    //update checked state
    slotUpdateUI();
}

void MainWindow::slotClientPreferences(bool /*checked =false */)
{
    PreferencesDlg dlg(this);

    //we need to be able to process local frames (userid 0),
    //so ensure these video frames are not being displayed elsewhere
    bool ignore_set = m_vid_exclude.find(0) != m_vid_exclude.end();
    disconnect(this, SIGNAL(newVideoFrame(int,int)),
               ui.videogridWidget, SLOT(slotNewVideoFrame(int,int)));
    m_vid_exclude.remove(0);

    uservideo_t::iterator local_ite = m_uservideo.find(0);
    if(local_ite != m_uservideo.end())
        disconnect(this, SIGNAL(newVideoFrame(int,int)),
                   (*local_ite)->ui.uservideoWidget, 
                   SLOT(slotNewVideoFrame(int,int)));

    connect(this, SIGNAL(newVideoFrame(int,int)), &dlg, 
            SLOT(slotNewVideoFrame(int,int)));

    //see if we need to retranslate
    QString lang = ttSettings->value(SETTINGS_DISPLAY_LANGUAGE).toString();

    //show dialog
    bool b = dlg.exec();

    disconnect(this, SIGNAL(newVideoFrame(int,int)), &dlg, 
               SLOT(slotNewVideoFrame(int,int)));

    if(ignore_set)
        m_vid_exclude.insert(0);
    if(local_ite != m_uservideo.end())
        connect(this, SIGNAL(newVideoFrame(int,int)),
                (*local_ite)->ui.uservideoWidget, 
                SLOT(slotNewVideoFrame(int,int)));

    connect(this, SIGNAL(newVideoFrame(int,int)),
            ui.videogridWidget, SLOT(slotNewVideoFrame(int,int)));

    if(!b)return;

    User myself;
    if((TT_GetFlags(ttInst) & CLIENT_AUTHORIZED) &&
        TT_GetUser(ttInst, TT_GetMyUserID(ttInst), &myself))
    {
        if(_Q(myself.szNickname) != ttSettings->value(SETTINGS_GENERAL_NICKNAME))
            TT_DoChangeNickname(ttInst, _W(ttSettings->value(SETTINGS_GENERAL_NICKNAME).toString()));

        QString statusmsg = ttSettings->value(SETTINGS_GENERAL_STATUSMESSAGE).toString();
        //change to female if set
        if(!ttSettings->value(SETTINGS_GENERAL_GENDER, true).toBool())
            m_statusmode |= STATUSMODE_FEMALE;
        else
            m_statusmode &= ~STATUSMODE_FEMALE;

        //set status mode flags
        if(m_statusmode != myself.nStatusMode || statusmsg != _Q(myself.szStatusMsg))
            TT_DoChangeStatus(ttInst, m_statusmode, _W(statusmsg));
    }
    slotMeEnableVoiceActivation(ttSettings->value(SETTINGS_GENERAL_VOICEACTIVATED).toBool());
    slotMeEnablePushToTalk(ttSettings->value(SETTINGS_GENERAL_PUSHTOTALK).toBool());

    //always on top property
    Qt::WindowFlags wndflags = windowFlags();
    if(ttSettings->value(SETTINGS_DISPLAY_ALWAYSONTOP, false).toBool())
        wndflags |= Qt::WindowStaysOnTopHint;
    else
        wndflags &= ~Qt::WindowStaysOnTopHint;

    if(wndflags != windowFlags())
    {
        setWindowFlags(wndflags);
#ifdef Q_OS_WIN32
        //Qt creates a new HWND to support "always on top" so no more messages
        //are posted to our HWND. This is actually unsafe since the previous HWND
        //might have unhandled messages.
        TT_SwapTeamTalkHWND(ttInst, reinterpret_cast<HWND>(winId()));
#endif
        show(); //for some reason this has to be called, otherwise the window disappears
    }

    //show user count property
    ui.channelsWidget->setShowUserCount(ttSettings->value(SETTINGS_DISPLAY_USERSCOUNT, true).toBool());
    ui.channelsWidget->updateItemTextLength(ttSettings->value(SETTINGS_DISPLAY_MAX_STRING,
                                            SETTINGS_DISPLAY_MAX_STRING_DEFAULT).toInt());

    if(lang != ttSettings->value(SETTINGS_DISPLAY_LANGUAGE).toString())
        ui.retranslateUi(this);

#if defined(Q_OS_LINUX) || defined(Q_OS_DARWIN)
    //hotkeys are not registered in PreferencesDlg as on Windows
    loadHotKeys();
#endif

    emit(preferencesModified());

    HostEntry host;
    GetLatestHost(0, host);
    m_desktopaccess_entries.clear();
    GetDesktopAccessList(m_desktopaccess_entries, host.ipaddr, host.tcpport);

    slotUpdateUI();
}

void MainWindow::slotClientExit(bool /*checked =false */)
{
    //close using timer, otherwise gets a Qt assertion from the 
    //'setQuitOnLastWindowClosed' call.
    QApplication::setQuitOnLastWindowClosed(true);
    QTimer::singleShot(0, this, SLOT(close()));
}

void MainWindow::slotMeChangeNickname(bool /*checked =false */)
{
    bool ok = false;
    QString s = QInputDialog::getText(this, 
                            MENUTEXT(ui.actionChangeNickname->text()), 
        tr("Specify new nickname"), QLineEdit::Normal, 
        ttSettings->value(SETTINGS_GENERAL_NICKNAME).toString(), &ok);
    if(ok)
    {
        ttSettings->setValue(SETTINGS_GENERAL_NICKNAME, s);
        TT_DoChangeNickname(ttInst, _W(s));
    }
}

void MainWindow::slotMeChangeStatus(bool /*checked =false */)
{
    ChangeStatusDlg dlg(this);
    if(dlg.exec())
        m_statusmode = dlg.m_user.nStatusMode;
}

void MainWindow::slotMeEnablePushToTalk(bool checked)
{
    if(checked)
    {
        hotkey_t hotkey;
        if(!LoadHotKeySettings(HOTKEY_PUSHTOTALK, hotkey))
        {
            KeyCompDlg dlg(this);
            if(!dlg.exec())
                return;
            SaveHotKeySettings(HOTKEY_PUSHTOTALK, dlg.m_hotkey);
            hotkey = dlg.m_hotkey;
        }

        if(hotkey.size())
            enableHotKey(HOTKEY_PUSHTOTALK, hotkey);

#if defined(Q_OS_WIN32)
        m_pttKeys = GetHotKeyText(hotkey);
#endif
    }
    else
    {
        disableHotKey(HOTKEY_PUSHTOTALK);
#if defined(Q_OS_WIN32)
        m_pttKeys.clear();
#endif
    }

    ttSettings->setValue(SETTINGS_GENERAL_PUSHTOTALK, checked);

    slotUpdateUI();
}

void MainWindow::slotMeEnableVoiceActivation(bool checked)
{
    TT_EnableVoiceActivation(ttInst, checked);
    ui.voiceactSlider->setVisible(checked);
    ttSettings->setValue(SETTINGS_GENERAL_VOICEACTIVATED, checked);
    if(TT_GetFlags(ttInst) & CLIENT_CONNECTED)
        ui.channelsWidget->slotUserTalking(TT_GetMyUserID(ttInst), 
                                           TT_IsTransmitting(ttInst, TRANSMIT_AUDIO));
    slotUpdateUI();
}

void MainWindow::slotMeEnableVideoTransmission(bool /*checked*/)
{
    ClientFlags flags = TT_GetFlags(ttInst);
    if((flags & CLIENT_VIDEO_READY) == 0)
    {
        if(!InitVideoFromSettings())
            QMessageBox::warning(this, 
            MENUTEXT(ui.actionEnableVideoTransmission->text()), 
            tr("Video device hasn't been configured property. Check settings in 'Preferences'"));
        else 
        {
            m_statusmode |= STATUSMODE_VIDEOTX;
            if(flags & CLIENT_AUTHORIZED)
                TT_DoChangeStatus(ttInst, m_statusmode, 
                _W(ttSettings->value(SETTINGS_GENERAL_STATUSMESSAGE).toString()));

            TT_EnableTransmission(ttInst, TRANSMIT_VIDEO, true);
        }
    }
    else
    {
        TT_CloseVideoCaptureDevice(ttInst);
        m_statusmode &= ~STATUSMODE_VIDEOTX;
        if(flags & CLIENT_AUTHORIZED)
            TT_DoChangeStatus(ttInst, m_statusmode, 
            _W(ttSettings->value(SETTINGS_GENERAL_STATUSMESSAGE).toString()));

        //remove local from video grid
        if(ui.videogridWidget->userExists(0))
            ui.videogridWidget->slotRemoveUser(0 /* local video*/, 0);

        TT_EnableTransmission(ttInst, TRANSMIT_VIDEO, false);
    }


    slotUpdateUI();
    slotUpdateVideoTabUI();
}

void MainWindow::slotMeEnableDesktopSharing(bool checked/*=false*/)
{
    if(checked)
    {
#if defined(Q_OS_LINUX)
        if(!m_display)
            m_display = XOpenDisplay(0);

        if(!m_display)
            QMessageBox::critical(this, MENUTEXT(ui.actionEnableDesktopSharing->text()),
                                  tr("Failed to open X11 display."));

        DesktopShareDlg dlg(m_display, this);
#else
        DesktopShareDlg dlg(this);
#endif
        if(dlg.exec())
        {
#if defined(Q_OS_WIN32)
            m_hShareWnd = dlg.m_hShareWnd;
#elif defined(Q_OS_DARWIN)
            m_nCGShareWnd = dlg.m_nCGShareWnd;
#elif defined(Q_OS_LINUX)
            m_nWindowShareWnd = dlg.m_nWindowShareWnd;
#endif
            if(!sendDesktopWindow())
            {
                ui.actionEnableDesktopSharing->setChecked(false);
                QMessageBox::information(this, MENUTEXT(ui.actionEnableDesktopSharing->text()),
                    tr("Failed to start desktop sharing"));
                return;
            }

            restartSendDesktopWindowTimer();

            m_statusmode |= STATUSMODE_DESKTOP;
            QString statusmsg = ttSettings->value(SETTINGS_GENERAL_STATUSMESSAGE).toString();
            if(TT_GetFlags(ttInst) & CLIENT_AUTHORIZED)
                TT_DoChangeStatus(ttInst, m_statusmode, _W(statusmsg));
        }
        else
            ui.actionEnableDesktopSharing->setChecked(false);
    }
    else
    {
        killLocalTimer(TIMER_SEND_DESKTOPWINDOW);
        TT_CloseDesktopWindow(ttInst);

#if defined(Q_OS_LINUX)
        if(m_display)
            XCloseDisplay(m_display);
        m_display = NULL;
#endif
            m_statusmode &= ~STATUSMODE_DESKTOP;
            QString statusmsg = ttSettings->value(SETTINGS_GENERAL_STATUSMESSAGE).toString();
            if(TT_GetFlags(ttInst) & CLIENT_AUTHORIZED)
                TT_DoChangeStatus(ttInst, m_statusmode, _W(statusmsg));
    }
}

void MainWindow::slotUsersViewUserInformation(bool /*checked =false */)
{
    slotUsersViewUserInformation(ui.channelsWidget->selectedUser());
}

void MainWindow::slotUsersMessages(bool /*checked =false */)
{
    int userid = ui.channelsWidget->selectedUser();
    slotUsersMessages(userid);
}

void MainWindow::slotUsersMute(bool checked /*=false */)
{
    slotUsersMute(ui.channelsWidget->selectedUser(), checked);
}

void MainWindow::slotUsersVolume(bool /*checked =false */)
{
    int userid = ui.channelsWidget->selectedUser();
    slotUsersVolume(userid);
}

void MainWindow::slotUsersMuteAll(bool checked /*=false */)
{
    TT_SetSoundOutputMute(ttInst, checked);
}

void MainWindow::slotUsersOp(bool /*checked =false */)
{
    int userid = ui.channelsWidget->selectedUser();
    int chanid = ui.channelsWidget->selectedChannel(true);
    slotUsersOp(userid, chanid);
}

void MainWindow::slotUsersKick(bool /*checked =false */)
{
    int userid = ui.channelsWidget->selectedUser();
    int chanid = ui.channelsWidget->selectedChannel(true);
    slotUsersKick(userid, chanid);
}

void MainWindow::slotUsersKickBan(bool /*checked =false */)
{
    int userid = ui.channelsWidget->selectedUser();
    int chanid = ui.channelsWidget->selectedChannel(true);
    slotUsersKickBan(userid, chanid);
}

void MainWindow::slotUsersStreamAudioFileToUser(bool /*checked=false*/)
{
    int userid = ui.channelsWidget->selectedUser();
    slotUsersStreamAudioFileToUser(userid);
}

void MainWindow::slotUsersSubscriptionsUserMsg(bool checked /*=false */)
{
    subscribeCommon(checked, SUBSCRIBE_USER_MSG);
}

void MainWindow::slotUsersSubscriptionsChannelMsg(bool checked /*=false */)
{
    subscribeCommon(checked, SUBSCRIBE_CHANNEL_MSG);
}

void MainWindow::slotUsersSubscriptionsBCastMsg(bool checked /*=false */)
{
    subscribeCommon(checked, SUBSCRIBE_BROADCAST_MSG);
}

void MainWindow::slotUsersSubscriptionsAudio(bool checked /*=false */)
{
    subscribeCommon(checked, SUBSCRIBE_AUDIO);
}

void MainWindow::slotUsersSubscriptionsVideo(bool checked /*=false */)
{
    subscribeCommon(checked, SUBSCRIBE_VIDEO);
    if(checked)
    {
        //remove from ignore if user wants video again
        int userid = ui.channelsWidget->selectedUser();
        if(userid>0)
            m_vid_exclude.remove(userid);
    }
}

void MainWindow::slotUsersSubscriptionsDesktop(bool checked /*=false */)
{
    subscribeCommon(checked, SUBSCRIBE_DESKTOP);
}

void MainWindow::slotUsersSubscriptionsDesktopInput(bool checked /*=false */)
{
    subscribeCommon(checked, SUBSCRIBE_DESKTOPINPUT);
    int userid = ui.channelsWidget->selectedUser();
    User user;
    if(ui.channelsWidget->getUser(userid, user))
        addStatusMsg(QString(tr("%1 granted desktop access")
                        .arg(_Q(user.szNickname))));
}

void MainWindow::slotUsersSubscriptionsInterceptUserMsg(bool checked/*=false*/)
{
    subscribeCommon(checked, SUBSCRIBE_INTERCEPT_USER_MSG);
}

void MainWindow::slotUsersSubscriptionsInterceptChannelMsg(bool checked/*=false*/)
{
    subscribeCommon(checked, SUBSCRIBE_INTERCEPT_CHANNEL_MSG);
}

void MainWindow::slotUsersSubscriptionsInterceptAudio(bool checked/*=false*/)
{
    subscribeCommon(checked, SUBSCRIBE_INTERCEPT_AUDIO);
}

void MainWindow::slotUsersSubscriptionsInterceptVideo(bool checked/*=false*/)
{
    subscribeCommon(checked, SUBSCRIBE_INTERCEPT_VIDEO);
    if(checked)
    {
        //remove from ignore if user wants video again
        int userid = ui.channelsWidget->selectedUser();
        if(userid>0)
            m_vid_exclude.remove(userid);
    }
}

void MainWindow::slotUsersSubscriptionsInterceptDesktop(bool checked/*=false*/)
{
    subscribeCommon(checked, SUBSCRIBE_INTERCEPT_DESKTOP);
}

void MainWindow::slotUsersAdvancedIncVolume()
{
    userids_t users = ui.channelsWidget->selectedUsers();
    for(int i=0;i<users.size();i++)
    {
        int vol = TT_GetUserVolume(ttInst, users[i]) + 10;
        vol = vol > SOUND_VOLUME_MAX?SOUND_VOLUME_MAX:vol;
        TT_SetUserVolume(ttInst, users[i], vol);
    }
    slotUpdateUI();
}

void MainWindow::slotUsersAdvancedDecVolume()
{
    userids_t users = ui.channelsWidget->selectedUsers();
    for(int i=0;i<users.size();i++)
    {
        int vol = TT_GetUserVolume(ttInst, users[i]) - 10;
        vol = vol < SOUND_VOLUME_MIN?SOUND_VOLUME_MIN:vol;
        TT_SetUserVolume(ttInst, users[i], vol);
    }
    slotUpdateUI();
}

void MainWindow::slotUsersAdvancedStoreForMove()
{
    m_moveusers.clear();
    m_moveusers = ui.channelsWidget->selectedUsers();
    slotUpdateUI();
}

void MainWindow::slotUsersAdvancedMoveUsers()
{
    int chanid = ui.channelsWidget->selectedChannel(true);
    if(chanid>0)
    {
        for(int i=0;i<m_moveusers.size();i++)
            TT_DoMoveUser(ttInst, m_moveusers[i], chanid);
    }
    slotUpdateUI();
}

void MainWindow::slotUsersAdvancedVoiceAllowed(bool checked/*=false*/)
{
    int userid = ui.channelsWidget->selectedUser();
    int channelid = ui.channelsWidget->selectedChannel(true);
    if(userid>0 && channelid>0)
    {
        QSet<int> voiceUsers, videoUsers, desktopUsers;
        ui.channelsWidget->getClassRoomUsers(channelid, voiceUsers, 
                                             videoUsers, desktopUsers);
        if(checked)
            voiceUsers.insert(userid);
        else
            voiceUsers.remove(userid);
        slotClassroomChanged(channelid, voiceUsers, videoUsers, desktopUsers);
    }
}

void MainWindow::slotUsersAdvancedVideoAllowed(bool checked/*=false*/)
{
    int userid = ui.channelsWidget->selectedUser();
    int channelid = ui.channelsWidget->selectedChannel(true);
    if(userid>0 && channelid>0)
    {
        QSet<int> voiceUsers, videoUsers, desktopUsers;
        ui.channelsWidget->getClassRoomUsers(channelid, voiceUsers, 
                                             videoUsers, desktopUsers);
        if(checked)
            videoUsers.insert(userid);
        else
            videoUsers.remove(userid);
        slotClassroomChanged(channelid, voiceUsers, videoUsers, desktopUsers);
    }
}

void MainWindow::slotUsersAdvancedDesktopAllowed(bool checked/*=false*/)
{
    int userid = ui.channelsWidget->selectedUser();
    int channelid = ui.channelsWidget->selectedChannel(true);
    if(userid>0 && channelid>0)
    {
        QSet<int> voiceUsers, videoUsers, desktopUsers;
        ui.channelsWidget->getClassRoomUsers(channelid, voiceUsers, 
                                             videoUsers, desktopUsers);
        if(checked)
            desktopUsers.insert(userid);
        else
            desktopUsers.remove(userid);
        slotClassroomChanged(channelid, voiceUsers, videoUsers, desktopUsers);
    }
}

void MainWindow::slotUsersStoreAudioToDisk(bool/* checked*/)
{
    if(AudioStorageDlg(this).exec())
    {
        m_audiostorage_mode = ttSettings->value(SETTINGS_GENERAL_AUDIOSTORAGE_MODE, 
                                                AUDIOSTORAGE_NONE).toUInt();
        ui.actionEnableStoreAudioToDisk->setChecked(true);
        slotEnableAudioRecording(true);
    }
    else
        ui.actionEnableStoreAudioToDisk->setChecked(false);

    slotUpdateUI();
}

void MainWindow::slotChannelsCreateChannel(bool /*checked =false */)
{
    Channel chan;
    ZERO_STRUCT(chan);

    if(ui.channelsWidget->selectedChannel() && (TT_GetMyUserType(ttInst) & USERTYPE_ADMIN))
        chan.nParentID = ui.channelsWidget->selectedChannel();
    else
    {
        chan.nParentID = TT_GetMyChannelID(ttInst);
        if(!chan.nParentID)
            chan.nParentID = TT_GetRootChannelID(ttInst);
    }

    ChannelDlg dlg(ChannelDlg::CHANNEL_CREATE, chan, this);
    if(!dlg.exec())return;
    chan = dlg.GetChannel();

    //only admins can create channels. Users can only create a new channel
    //if they at the same time join it.
    if(TT_GetMyUserType(ttInst) & USERTYPE_ADMIN)
    {
        if(TT_DoMakeChannel(ttInst, &chan)<0)
            QMessageBox::critical(this, MENUTEXT(ui.actionCreateChannel->text()), 
            tr("Failed to issue command to create channel"));
    }
    else
    {
        int cmdid = TT_DoJoinChannel(ttInst, &chan);
        if(cmdid>0)
            m_commands.insert(cmdid, CMD_COMPLETE_JOINCHANNEL);
        else
            QMessageBox::critical(this, MENUTEXT(ui.actionCreateChannel->text()), 
                                  tr("Failed to issue command to create channel"));
    }
}

void MainWindow::slotChannelsUpdateChannel(bool /*checked =false */)
{
    Channel chan;
    if(ui.channelsWidget->getSelectedChannel(chan))
    {
        ChannelDlg dlg(ChannelDlg::CHANNEL_UPDATE, chan, this);
        if(!dlg.exec())return;
        chan = dlg.GetChannel();
        if(TT_DoUpdateChannel(ttInst, &chan)<0)
            QMessageBox::critical(this, MENUTEXT(ui.actionUpdateChannel->text()),
            tr("Failed to issue command to update channel"));
    }
}

void MainWindow::slotChannelsDeleteChannel(bool /*checked =false */)
{
    int chanid = ui.channelsWidget->selectedChannel();
    if(chanid<=0)
        return;

    TTCHAR buff[TT_STRLEN] = {0};
    TT_GetChannelPath(ttInst, chanid, buff);
    if(QMessageBox::information(this, MENUTEXT(ui.actionDeleteChannel->text()),
        tr("Are you sure you want to delete channel \"%1\"?").arg(_Q(buff)), 
        QMessageBox::Yes | QMessageBox::No) == QMessageBox::No)
        return;

    if(TT_DoRemoveChannel(ttInst, chanid)<0)
        QMessageBox::critical(this, MENUTEXT(ui.actionDeleteChannel->text()), 
        tr("Failed to issue command to delete channel"));
}

void MainWindow::slotChannelsJoinChannel(bool /*checked=false*/)
{
    Channel chan;
    if(!ui.channelsWidget->getSelectedChannel(chan))
        return;

    QString password;
    if(chan.bPassword)
    {
        bool ok = false;
        password = QInputDialog::getText(this, MENUTEXT(ui.actionJoinChannel->text()), 
            tr("Specify password"), QLineEdit::Password, "", &ok);
        if(!ok)
            return;
    }
    int cmdid = TT_DoJoinChannelByID(ttInst, chan.nChannelID, _W(password));
    if(cmdid>0)
    {
        m_commands.insert(cmdid, CMD_COMPLETE_JOINCHANNEL);
        m_last_channel = chan;
        COPY_TTSTR(m_last_channel.szPassword, password);
    }
    else
        QMessageBox::critical(this, MENUTEXT(ui.actionJoinChannel->text()),
                              tr("Failed to issue command to join channel"));
}

void MainWindow::slotChannelsViewChannelInfo(bool /*checked=false*/)
{
    Channel chan;
    if(ui.channelsWidget->getSelectedChannel(chan))
    {
        ChannelDlg dlg(ChannelDlg::CHANNEL_READONLY, chan, this);
        dlg.exec();
    }
}

void MainWindow::slotChannelsStreamAudioFile(bool checked)
{
    if(TT_GetMyChannelID(ttInst) <= 0)
        return;

    if(!checked)
    {
        TT_StopStreamingAudioFileToUser(ttInst, 0);
        return;
    }

    StreamAudioFileDlg dlg(this);
    if(!dlg.exec())
    {
        ui.actionStreamAudioFileToChannel->setChecked(false);
        return;
    }

    QString fileName = ttSettings->value(SETTINGS_GENERAL_STREAMAUDIO_FILENAME).toString();

    if(fileName.size() && 
       !TT_StartStreamingAudioFileToChannel(ttInst, 0, _W(fileName)))
    {
        QMessageBox::information(this,
                                 MENUTEXT(ui.actionStreamAudioFileToChannel->text()),
                                 QString(tr("Failed to stream file %1").arg(fileName)));
        ui.actionStreamAudioFileToChannel->setChecked(false);
    }
}

void MainWindow::slotChannelsStreamMediaFile(bool checked/*=false*/)
{
    if(!checked)
    {
        stopStreamMediaFile();
        return;
    }

    StreamMediaFileDlg dlg(this);
    if(!dlg.exec())
    {
        ui.actionStreamMediaFileToChannel->setChecked(false);
        return;
    }

    startStreamMediaFile();
}

void MainWindow::slotChannelsUploadFile(bool /*checked =false */)
{
    int channelid = m_filesmodel->getChannelID();
    if(channelid>0)
    {
        QString start_dir = ttSettings->value(SETTINGS_GENERAL_LAST_DIRECTORY, QDir::homePath()).toString();
        QString filename = QFileDialog::getOpenFileName(this, tr("Open File"),
                                                       start_dir/*, tr("All Files (*.*)")*/);
        if(filename.isEmpty())
            return;
        ttSettings->setValue(SETTINGS_GENERAL_LAST_DIRECTORY, QFileInfo(filename).absolutePath());
        QStringList files;
        files.push_back(filename);
        slotUploadFiles(files);
    }
}

void MainWindow::slotChannelsDownloadFile(bool /*checked =false */)
{
    int fileid = (int)ui.filesView->currentIndex().internalId();
    int channelid = m_filesmodel->getChannelID();
    FileInfo fileinfo;
    if(fileid>0 && channelid>0 && 
       TT_GetChannelFileInfo(ttInst, channelid, fileid, &fileinfo))
    {
        QDir lastDir(ttSettings->value(SETTINGS_GENERAL_LAST_DIRECTORY, QDir::homePath()).toString());
        QString filename = QFileDialog::getSaveFileName(this, tr("Save File"),
                           lastDir.filePath(_Q(fileinfo.szFileName))
                           /*, tr("All Files (*.*)")*/);
        if(filename.isEmpty())
            return;
        filename = QDir::toNativeSeparators(filename);
        ttSettings->setValue(SETTINGS_GENERAL_LAST_DIRECTORY, QFileInfo(filename).absolutePath());
        if(!TT_DoRecvFile(ttInst, channelid, fileid, _W(filename)))
            QMessageBox::critical(this, MENUTEXT(ui.actionDownloadFile->text()),
                                  tr("Failed to download file %1").arg(filename));
    }
}

void MainWindow::slotChannelsDeleteFile(bool /*checked =false */)
{
    int channelid = m_filesmodel->getChannelID();
    if(!channelid)
        return;
    QStringList filenames;
    QList<int> files = ui.filesView->selectedFiles(&filenames);
    bool delete_ok = false;
    if(filenames.size() == 1)
        delete_ok = QMessageBox::information(this,
                                             MENUTEXT(ui.actionDeleteFile->text()),
                                             tr("Are you sure you want to delete \"%1\"?").arg(filenames[0]),
                                             QMessageBox::Yes | QMessageBox::No) == QMessageBox::Yes;
    else if(filenames.size()>1)
        delete_ok = QMessageBox::information(this,
                                             MENUTEXT(ui.actionDeleteFile->text()),
                                             tr("Are you sure you want to delete %1 file(s)?").arg(filenames.size()),
                                             QMessageBox::Yes | QMessageBox::No) == QMessageBox::Yes;

    for(int i=0;i<files.size() && delete_ok;i++)
        TT_DoDeleteFile(ttInst, channelid, files[i]);
}
   
void MainWindow::slotServerUserAccounts(bool /*checked =false */)
{
    //don't display dialog box until we get the result
    int cmdid = TT_DoListUserAccounts(ttInst, 0, 1000000);
    if(cmdid>0)
        m_commands.insert(cmdid, CMD_COMPLETE_LISTACCOUNTS);
}

void MainWindow::slotServerBannedUsers(bool /*checked =false */)
{
    //don't display dialog box until we get the result
    int cmdid = TT_DoListBans(ttInst, 0, 1000000);
    if(cmdid>0)
        m_commands.insert(cmdid, CMD_COMPLETE_LISTBANS);
}

void MainWindow::slotServerOnlineUsers(bool /*checked=false*/)
{
    if(m_onlineusersdlg)
        m_onlineusersdlg->activateWindow();
    else
    {
        m_onlineusersdlg = new OnlineUsersDlg();
        connect(m_onlineusersdlg, SIGNAL(finished(int)),
                SLOT(slotClosedOnlineUsersDlg(int)));
        m_onlineusersdlg->setAttribute(Qt::WA_DeleteOnClose);
    }

    connect(this, SIGNAL(userLogin(int)), m_onlineusersdlg, 
            SLOT(slotUserLoggedIn(int)));
    connect(this, SIGNAL(userLogout(int)), m_onlineusersdlg, 
            SLOT(slotUserLoggedOut(int)));
    connect(this, SIGNAL(userJoined(int,int)), m_onlineusersdlg, 
            SLOT(slotUserJoin(int,int)));
    connect(this, SIGNAL(userUpdate(int,int)), m_onlineusersdlg, 
            SLOT(slotUserUpdate(int,int)));
    connect(this, SIGNAL(userLeft(int,int)), m_onlineusersdlg, 
            SLOT(slotUserRemove(int,int)));

    connect(m_onlineusersdlg, SIGNAL(viewUserInformation(int)), 
            SLOT(slotUsersViewUserInformation(int)));
    connect(m_onlineusersdlg, SIGNAL(sendUserMessage(int)), 
            SLOT(slotUsersMessages(int)));
    connect(m_onlineusersdlg, SIGNAL(muteUser(int,bool)), 
            SLOT(slotUsersMute(int, bool)));
    connect(m_onlineusersdlg, SIGNAL(changeUserVolume(int)), 
            SLOT(slotUsersVolume(int)));
    connect(m_onlineusersdlg, SIGNAL(opUser(int,int)), 
            SLOT(slotUsersOp(int,int)));
    connect(m_onlineusersdlg, SIGNAL(kickUser(int,int)), 
            SLOT(slotUsersKick(int,int)));
    connect(m_onlineusersdlg, SIGNAL(kickbanUser(int,int)), 
            SLOT(slotUsersKickBan(int,int)));
    connect(m_onlineusersdlg, SIGNAL(streamfileToUser(int)), 
            SLOT(slotUsersStreamAudioFileToUser(int)));
    m_onlineusersdlg->show();
}

void MainWindow::slotServerBroadcastMessage(bool /*checked=false*/)
{
    bool ok = false;
    QString bcast = QInputDialog::getText(this, 
        MENUTEXT(ui.actionBroadcastMessage->text()), 
        tr("Message to broadcast:"), QLineEdit::Normal, "", &ok);
    if(!ok)
        return;
    TextMessage msg;
    msg.nMsgType = MSGTYPE_BROADCAST;
    msg.nFromUserID = TT_GetMyUserID(ttInst);
    COPY_TTSTR(msg.szMessage, bcast);
    TT_DoTextMessage(ttInst, &msg);
}

void MainWindow::slotServerServerProperties(bool /*checked =false */)
{
    ServerPropertiesDlg dlg(this);
    if(!dlg.exec())return;
}

void MainWindow::slotServerSaveConfiguration(bool /*checked =false */)
{
    TT_DoSaveConfig(ttInst);
}

void MainWindow::slotServerServerStatistics(bool /*checked=false*/)
{
    if(m_serverstatsdlg)
        m_serverstatsdlg->activateWindow();
    else
    {
        m_serverstatsdlg = new ServerStatisticsDlg();
        connect(this, SIGNAL(cmdSuccess(int)), m_serverstatsdlg, 
                SLOT(slotCmdSuccess(int)));
        connect(m_serverstatsdlg, SIGNAL(finished(int)),
                SLOT(slotClosedServerStatsDlg(int)));
        m_serverstatsdlg->setAttribute(Qt::WA_DeleteOnClose);
        m_serverstatsdlg->show();
    }
}

void MainWindow::slotHelpManual(bool /*checked =false */)
{
    QString file = "file:///" + APPMANUAL;
    QDesktopServices::openUrl(file);
}
    
void MainWindow::slotHelpVisitBearWare(bool /*checked=false*/)
{
   QDesktopServices::openUrl(QUrl(APPWEBSITE));
}

void MainWindow::slotHelpAbout(bool /*checked =false */)
{
    AboutDlg(this).exec();
}

void MainWindow::slotConnectToLatest()
{
    //auto connect to latest host
    HostEntry host;
    if(ttSettings->value(SETTINGS_CONNECTION_AUTOCONNECT, false).toBool() &&
        GetLatestHost(0, host))
        Connect();
}

void MainWindow::slotUsersViewUserInformation(int userid)
{
    UserInfoDlg dlg(userid, this);
    dlg.exec();
}

void MainWindow::slotUsersMessages(int userid)
{
    TextMessageDlg* dlg = getTextMessageDlg(userid);
    if(dlg)
    {
        dlg->show();
        dlg->activateWindow();
    }
}

void MainWindow::slotUsersMute(int userid, bool mute)
{
    TT_SetUserMute(ttInst, userid, mute);
}

void MainWindow::slotUsersVolume(int userid)
{
    UserVolumeDlg dlg(userid, this);
    dlg.exec();
}

void MainWindow::slotUsersOp(int userid, int chanid)
{
    BOOL op = TT_IsChannelOperator(ttInst, userid, chanid);

    BOOL me_opadmin = TT_IsChannelOperator(ttInst, TT_GetMyUserID(ttInst), chanid);
    me_opadmin |= (BOOL)(TT_GetMyUserType(ttInst) & USERTYPE_ADMIN);
    if(me_opadmin) //don't need password in this case
        TT_DoChannelOp(ttInst, userid, chanid, !op);
    else
    {
        bool ok = false;
        QString oppasswd = QInputDialog::getText(this, MENUTEXT(ui.actionOp->text()), 
            tr("Specify password"), QLineEdit::Password, "", &ok);
        if(ok)
            TT_DoChannelOpEx(ttInst, userid, chanid, _W(oppasswd), !op);
    }
}

void MainWindow::slotUsersKick(int userid, int chanid)
{
    //if we're admin we kick the user off the server (and not just out of
    //the channel).
    if(TT_GetMyUserType(ttInst) & USERTYPE_ADMIN)
        TT_DoKickUser(ttInst, userid, 0);
    else
        TT_DoKickUser(ttInst, userid, chanid);
}

void MainWindow::slotUsersKickBan(int userid, int /*chanid*/)
{
    //ban first since the user will otherwise have disappeared
    TT_DoBanUser(ttInst, userid);
    TT_DoKickUser(ttInst, userid, 0);
}

void MainWindow::slotUsersStreamAudioFileToUser(int userid)
{
    //clear audio file if currently active.
    TT_StopStreamingAudioFileToUser(ttInst, userid);

    QString fileName = QFileDialog::getOpenFileName(this, tr("Open Wave File"),
                                                 "",
                                                 tr("Wave files (*.wav)"));
    if(fileName.size() && !TT_StartStreamingAudioFileToUser(ttInst, userid, _W(fileName)))
        QMessageBox::information(this, MENUTEXT(ui.actionStreamAudioFileToUser->text()),
        QString(tr("Failed to stream file %1").arg(fileName)));
}

void MainWindow::slotTreeSelectionChanged()
{
    slotUpdateUI();
    int channelid = ui.channelsWidget->selectedChannel(true);
    if(m_filesmodel->getChannelID() != channelid &&
       (TT_GetMyUserType(ttInst) & USERTYPE_ADMIN))
    {
        //if admin changed selection change files view to new channel.
        //if not admin then keep joined channel as file view.
        updateChannelFiles(channelid);
    }
}

void MainWindow::slotTreeContextMenu(const QPoint &/* pos*/)
{
    int userid = ui.channelsWidget->selectedUser();
    if(userid>0)
        ui.menuUsers->exec(QCursor::pos());

    int chanid = ui.channelsWidget->selectedChannel();
    if(chanid>0)
        ui.menuChannels->exec(QCursor::pos());
}

void MainWindow::slotUpdateUI()
{
    int userid = ui.channelsWidget->selectedUser();
    int chanid = ui.channelsWidget->selectedChannel();
    int user_chanid = chanid;
    if(!chanid)
        user_chanid = ui.channelsWidget->selectedChannel(true);
    int mychannel = TT_GetMyChannelID(ttInst);
    int filescount = ui.filesView->selectedFiles().size();
    ClientFlags statemask = TT_GetFlags(ttInst);
    bool auth = (statemask & CLIENT_AUTHORIZED);
    bool me_admin = (TT_GetMyUserType(ttInst) & USERTYPE_ADMIN);
    bool me_op = TT_IsChannelOperator(ttInst, TT_GetMyUserID(ttInst), user_chanid);

    ui.actionConnect->setChecked( (statemask & CLIENT_CONNECTING) || (statemask & CLIENT_CONNECTED));
    ui.actionChangeNickname->setEnabled(auth);
    ui.actionChangeStatus->setEnabled(auth);
#ifdef Q_OS_WIN32
    ui.actionEnablePushToTalk->setChecked(TT_HotKey_IsActive(ttInst, HOTKEY_PUSHTOTALK) >= 0);
#elif defined(Q_OS_LINUX) || defined(Q_OS_DARWIN)
    ui.actionEnablePushToTalk->setChecked(m_hotkeys.find(HOTKEY_PUSHTOTALK) != m_hotkeys.end());
#endif
    ui.actionEnableVoiceActivation->setChecked(statemask & CLIENT_SNDINPUT_VOICEACTIVATED);
    //don't allow web cam to stream when video streaming is active
    ui.actionEnableVideoTransmission->setEnabled((statemask & CLIENT_STREAM_VIDEO) == 0);
    ui.actionEnableVideoTransmission->setChecked((CLIENT_VIDEO_READY & statemask) && 
                                                 (CLIENT_TX_VIDEO & statemask));
    ui.actionEnableDesktopSharing->setEnabled(
                            VersionSameOrLater(_Q(m_srvprop.szServerVersion),
                                               TT_SERVER_DESKTOP_MIN));
    ui.actionEnableDesktopSharing->setChecked(statemask & CLIENT_DESKTOP_ACTIVE);

    User user;
    ZERO_STRUCT(user);
    if(TT_GetUser(ttInst, userid, &user))
    {
        ui.actionMute->setChecked(user.uUserState & USERSTATE_MUTE);
        ui.actionStreamAudioFileToUser->setEnabled(mychannel != user.nChannelID && me_admin);
        ui.actionDesktopAccessAllow->setChecked(user.uLocalSubscriptions & SUBSCRIBE_DESKTOPINPUT);

        ui.actionUserMessages->setChecked(user.uLocalSubscriptions & SUBSCRIBE_USER_MSG);
        ui.actionChannelMessages->setChecked(user.uLocalSubscriptions & SUBSCRIBE_CHANNEL_MSG);
        ui.actionBroadcastMessages->setChecked(user.uLocalSubscriptions & SUBSCRIBE_BROADCAST_MSG);
        ui.actionAudio->setChecked(user.uLocalSubscriptions & SUBSCRIBE_AUDIO);
        ui.actionVideo->setChecked(user.uLocalSubscriptions & SUBSCRIBE_VIDEO);
        ui.actionDesktop->setChecked(user.uLocalSubscriptions & SUBSCRIBE_DESKTOP);
        ui.actionDesktopInput->setChecked(user.uLocalSubscriptions & SUBSCRIBE_DESKTOPINPUT);

        ui.actionInterceptUserMessages->setChecked(user.uLocalSubscriptions & SUBSCRIBE_INTERCEPT_USER_MSG);
        ui.actionInterceptChannelMessages->setChecked(user.uLocalSubscriptions & SUBSCRIBE_INTERCEPT_CHANNEL_MSG);
        ui.actionInterceptAudio->setChecked(user.uLocalSubscriptions & SUBSCRIBE_INTERCEPT_AUDIO);
        ui.actionInterceptVideo->setChecked(user.uLocalSubscriptions & SUBSCRIBE_INTERCEPT_VIDEO);
        ui.actionInterceptDesktop->setChecked(user.uLocalSubscriptions & SUBSCRIBE_INTERCEPT_DESKTOP);
    }
    else
    {
        ui.actionStreamAudioFileToUser->setEnabled(false);
    }

    ui.actionViewUserInformation->setEnabled(userid>0);
    ui.actionMessages->setEnabled(userid>0);
    ui.actionMute->setEnabled(userid>0);
    ui.actionVolume->setEnabled(userid>0);
    ui.actionOp->setEnabled(userid>0);     
    ui.actionKick->setEnabled(userid>0);
    ui.actionKickBan->setEnabled(userid>0);
    ui.actionDesktopAccessAllow->setEnabled(userid>0);

    ui.actionUserMessages->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    ui.actionChannelMessages->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    ui.actionBroadcastMessages->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    ui.actionAudio->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    ui.actionVideo->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    ui.actionDesktop->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    ui.actionDesktopInput->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    //intercept only works for admins
    ui.actionInterceptUserMessages->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    ui.actionInterceptChannelMessages->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    ui.actionInterceptAudio->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    ui.actionInterceptVideo->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));
    ui.actionInterceptDesktop->setEnabled(userid>0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS));

    ui.actionIncreaseVolume->setEnabled(userid>0 && TT_GetUserVolume(ttInst, userid) < SOUND_VOLUME_MAX);
    ui.actionLowerVolume->setEnabled(userid>0 && TT_GetUserVolume(ttInst, userid) > SOUND_VOLUME_MIN);
    ui.actionStoreForMove->setEnabled(userid>0 && me_admin);
    ui.actionMoveUser->setEnabled(m_moveusers.size() && me_admin);

    //ui.actionMuteAll->setEnabled(statemask & CLIENT_SOUND_READY);
    ui.actionMuteAll->setChecked(statemask & CLIENT_SNDOUTPUT_MUTE);
    ui.actionSpecifyAudioFolder->setEnabled(true);

    //Channel-menu items
    Channel chan;
    ZERO_STRUCT(chan);
    if(TT_GetChannel(ttInst, user_chanid, &chan))
    {
    }

    ui.actionJoinChannel->setEnabled(chanid>0);
    ui.actionViewChannelInfo->setEnabled(chanid>0);
    ui.actionCreateChannel->setEnabled(chanid>0 || mychannel>0);
    ui.actionUpdateChannel->setEnabled(chanid>0);
    ui.actionDeleteChannel->setEnabled(chanid>0);
    ui.actionStreamAudioFileToChannel->setEnabled(mychannel > 0);
    ui.actionStreamMediaFileToChannel->setChecked(statemask & 
                                                  (CLIENT_STREAM_AUDIO | CLIENT_STREAM_VIDEO));
    ui.actionStreamMediaFileToChannel->setEnabled(mychannel > 0);
    ui.actionUploadFile->setEnabled(mychannel>0);
    ui.actionDownloadFile->setEnabled(mychannel>0);
    ui.actionDeleteFile->setEnabled(filescount>0);

    //Users-menu items dependent on Channel
    ui.actionAllowVoiceTransmission->setChecked(userCanVoiceTx(userid, chan));
    ui.actionAllowVoiceTransmission->setEnabled(userid>0 && (me_op || me_admin));
    ui.actionAllowVideoTransmission->setChecked(userCanVideoTx(userid, chan));
    ui.actionAllowVideoTransmission->setEnabled(userid>0 && (me_op || me_admin));
    ui.actionAllowDesktopTransmission->setChecked(userCanDesktopTx(userid, chan));
    ui.actionAllowDesktopTransmission->setEnabled(userid>0 && (me_op || me_admin));

    //Server-menu items
    ui.actionUserAccounts->setEnabled(auth && me_admin);
    ui.actionBannedUsers->setEnabled(auth && me_admin);
    ui.actionOnlineUsers->setEnabled(auth);
    ui.actionBroadcastMessage->setEnabled(auth && me_admin);
    ui.actionServerProperties->setEnabled(auth);
    ui.actionSaveConfiguration->setEnabled(auth && me_admin);
    ui.actionServerStatistics->setEnabled(auth && me_admin);

    ui.uploadButton->setEnabled(mychannel>0);
    ui.downloadButton->setEnabled(mychannel>0);
}

void MainWindow::slotUpdateVideoTabUI()
{
    ClientFlags statemask = TT_GetFlags(ttInst);
    int userid = ui.videogridWidget->selectedUser();

    ui.detachVideoButton->setEnabled(userid>0);
    ui.removeVideoButton->setEnabled(ui.videogridWidget->getActiveUsersCount());
    ui.initVideoButton->setChecked(statemask & CLIENT_VIDEO_READY);
}

void MainWindow::slotUpdateDesktopTabUI()
{
    int userid = ui.desktopgridWidget->selectedUser();
    User user;
    ZERO_STRUCT(user);
    ui.channelsWidget->getUser(userid, user);
    
    ui.detachDesktopButton->setEnabled(userid>0);
    ui.removeDesktopButton->setEnabled(ui.desktopgridWidget->getActiveUsersCount());
    ui.desktopaccessButton->setEnabled(userid>0);
    ui.desktopaccessButton->setChecked(user.uPeerSubscriptions & SUBSCRIBE_DESKTOPINPUT);
}

void MainWindow::slotUploadFiles(const QStringList& files)
{
    int channelid = m_filesmodel->getChannelID();
    Q_ASSERT(channelid>0);
    for(int i=0;i<files.size();i++)
    {
        QString filename = QDir::toNativeSeparators(files[i]);
        if(!TT_DoSendFile(ttInst, channelid, _W(filename)))
            QMessageBox::critical(this, MENUTEXT(ui.actionUploadFile->text()),
            tr("Failed to upload file %1").arg(filename));
    }
}

void MainWindow::slotSendChannelMessage()
{
    int mychanid = TT_GetMyChannelID(ttInst);
    if(mychanid<=0)
        return;

    QString txtmsg;
    switch(ui.tabWidget->currentIndex())
    {
    case TAB_CHAT :
        txtmsg = ui.msgEdit->text();
        ui.msgEdit->clear();
        break;
    case TAB_DESKTOP :
        txtmsg = ui.desktopmsgEdit->text();
        ui.desktopmsgEdit->clear();
    default :
        break;
    }

    if(txtmsg.isEmpty())
        return;

    TextMessage msg;
    msg.nFromUserID = TT_GetMyUserID(ttInst);
    msg.nChannelID = mychanid;
    msg.nMsgType = MSGTYPE_CHANNEL;
    COPY_TTSTR(msg.szMessage, txtmsg);
    TT_DoTextMessage(ttInst, &msg);
}

void MainWindow::slotUserDoubleClicked(int)
{
    if(QApplication::keyboardModifiers() & Qt::ControlModifier)
    {
        User user;
        if(ui.channelsWidget->getSelectedUser(user))
        {
            //ensure we don't open two dialogs
            if(m_uservideo.find(user.nUserID) != m_uservideo.end())
                m_uservideo.find(user.nUserID).value()->raise();
            else
            {
                //resubscribe in case disabled
                Subscriptions subs = SUBSCRIBE_NONE;
                if((user.uLocalSubscriptions & SUBSCRIBE_VIDEO) == 0)
                    subs |= SUBSCRIBE_VIDEO;
                if(user.nChannelID != m_mychannel.nChannelID)
                    subs |= SUBSCRIBE_INTERCEPT_VIDEO;

                subscribeCommon(true, subs, user.nUserID);

                m_vid_exclude.remove(user.nUserID);

                slotDetachUserVideo(user.nUserID);
            }
        }
    }
    else
        slotUsersMessages(false);
}

void MainWindow::slotChannelDoubleClicked(int)
{
    slotChannelsJoinChannel(false);
}

void MainWindow::slotTextMessageClosed(int userid)
{
    usermsg_t::iterator ite = m_usermsg.find(userid);
    Q_ASSERT(ite != m_usermsg.end());
    if(ite != m_usermsg.end())
    {
        //(*ite)->deleteLater();
        m_usermsg.erase(ite);
        ui.channelsWidget->setUserMessaged(userid, false);
    }
}

void MainWindow::slotClassroomChanged(int channelid, 
                                      const QSet<int>& voiceUsers, 
                                      const QSet<int>& videoUsers,
                                      const QSet<int>& desktopUsers)
{
    Channel chan;
    if(!TT_GetChannel(ttInst, channelid, &chan))
        return;

    if(voiceUsers.size()>TT_VOICEUSERS_MAX)
    {
        QMessageBox::information(this, 
            MENUTEXT(ui.actionAllowVoiceTransmission->text()),
            tr("The maximum number of users who can transmit is %1")
            .arg(TT_VOICEUSERS_MAX));
        return;
    }
    if(videoUsers.size()>TT_VIDEOUSERS_MAX)
    {
        QMessageBox::information(this, 
            MENUTEXT(ui.actionAllowVideoTransmission->text()),
            tr("The maximum number of users who can transmit is %1")
            .arg(TT_VIDEOUSERS_MAX));
        return;
    }
    if(desktopUsers.size()>TT_DESKTOPUSERS_MAX)
    {
        QMessageBox::information(this, 
            MENUTEXT(ui.actionAllowDesktopTransmission->text()),
            tr("The maximum number of users who can transmit is %1")
            .arg(TT_DESKTOPUSERS_MAX));
        return;
    }


    QSet<int>::const_iterator ite;
    ite = voiceUsers.begin();
    for(int i=0;i<TT_VOICEUSERS_MAX;i++)
    {
        if(ite != voiceUsers.end())
        {
            chan.voiceUsers[i] = *ite;
            ite++;
        }
        else
            chan.voiceUsers[i] = 0;
    }

    ite = videoUsers.begin();
    for(int i=0;i<TT_VIDEOUSERS_MAX;i++)
    {
        if(ite != videoUsers.end())
        {
            chan.videoUsers[i] = *ite;
            ite++;
        }
        else
            chan.videoUsers[i] = 0;
    }

    ite = desktopUsers.begin();
    for(int i=0;i<TT_DESKTOPUSERS_MAX;i++)
    {
        if(ite != desktopUsers.end())
        {
            chan.desktopUsers[i] = *ite;
            ite++;
        }
        else
            chan.desktopUsers[i] = 0;
    }

    TT_DoUpdateChannel(ttInst, &chan);
}

void MainWindow::slotChannelUpdate(int chanid)
{
    Channel chan, oldchan;
    if(!TT_GetChannel(ttInst, chanid, &chan) || 
       !ui.channelsWidget->getChannel(chanid, oldchan) ||
       (chan.uChannelType & CHANNEL_CLASSROOM) == 0)
        return;
    
    //specific to classroom channel
    QString msg;
    bool before = false, after = false;
    before = userCanVoiceTx(TT_GetMyUserID(ttInst), oldchan);
    after = userCanVoiceTx(TT_GetMyUserID(ttInst), chan);
    if(before != after)
    {
        if(after)
            msg = tr("You can now transmit audio!");
        else
            msg = tr("You can no longer transmit audio!");
        addStatusMsg(msg);
    }
    before = userCanVideoTx(TT_GetMyUserID(ttInst), oldchan);
    after = userCanVideoTx(TT_GetMyUserID(ttInst), chan);
    if(before != after)
    {
        if(after)
            msg = tr("You can now transmit video!");
        else
            msg = tr("You can no longer transmit video!");
        addStatusMsg(msg);
    }
    before = userCanDesktopTx(TT_GetMyUserID(ttInst), oldchan);
    after = userCanDesktopTx(TT_GetMyUserID(ttInst), chan);
    if(before != after)
    {
        if(after)
            msg = tr("You can now transmit desktop windows!");
        else
            msg = tr("You can no longer transmit desktop windows!");
        addStatusMsg(msg);
    }
}

void MainWindow::slotInitVideo()
{
    if((TT_GetFlags(ttInst) & CLIENT_VIDEO_READY) == 0)
    {
        if(!InitVideoFromSettings())
            QMessageBox::warning(this, tr("Start Webcam"), 
            tr("Video device hasn't been configured property. Check settings in 'Preferences'"));
    }
    else
    {
        TT_CloseVideoCaptureDevice(ttInst);
    }
    slotUpdateUI();
    slotUpdateVideoTabUI();
}

void MainWindow::slotAddUserVideo()
{
    int chanid = TT_GetMyChannelID(ttInst);
    QVector<int> users = ui.channelsWidget->getUsersInChannel(chanid);
    QMenu menu(this);

    //local video frames have userid 0
    if((TT_GetFlags(ttInst) & CLIENT_VIDEO_READY) &&
       m_uservideo.find(0) == m_uservideo.end() &&
       !ui.videogridWidget->userExists(0))
    {
        QAction* myself = menu.addAction(tr("Myself"));
        myself->setData(0);
        if(users.size())
            menu.addSeparator();
        users.push_back(0);
    }

    //we don't want "myself" in the list
    int index = -1;
    if((index = users.indexOf(TT_GetMyUserID(ttInst))) != -1)
        users.remove(index);

    for(int i=0;i<users.size();i++)
    {
        if(m_uservideo.find(users[i]) != m_uservideo.end() ||
           ui.videogridWidget->userExists(users[i]))
           continue;

        User user;
        if(TT_GetUser(ttInst, users[i], &user) &&
           (user.uLocalSubscriptions & SUBSCRIBE_VIDEO) == 0)
           menu.addAction(_Q(user.szNickname))->setData(users[i]);
    }

    if(menu.isEmpty())return;
    
    QAction* action = menu.exec(QCursor::pos());
    if(!action)
        return;

    int userid = action->data().toInt();
    if(userid != 0)
    {
        //this is basically the same as slotUsersSubscriptionsVideo(..)
        subscribeCommon(true, SUBSCRIBE_VIDEO, userid);
    }

    ui.videogridWidget->slotAddUser(userid, getVideoStreamID(userid));

    m_vid_exclude.remove(userid);
}

void MainWindow::slotRemoveUserVideo()
{
    int userid = ui.videogridWidget->selectedUser();
    if(userid<0)
    {
        QVector<int> users = ui.videogridWidget->activeUsers();
        QMenu menu(this);
        if(ui.videogridWidget->userExists(0))
        {
            QAction* myself = menu.addAction(tr("Myself"));
            myself->setData(0);
            if(users.size())
                menu.addSeparator();
        }
        User user;
        for(int i=0;i<users.size();i++)
        {
            if(TT_GetUser(ttInst, users[i], &user))
                menu.addAction(_Q(user.szNickname))->setData(users[i]);
        }

        if(menu.isEmpty())
            return;

        QAction* action = menu.exec(QCursor::pos());
        if(!action)
            return;
        userid = action->data().toInt();
    }

    if(userid>0)
        subscribeCommon(false, SUBSCRIBE_VIDEO, userid);
    m_vid_exclude.insert(userid);

    slotRemoveUserVideo(userid);
}

void MainWindow::slotRemoveUserVideo(int userid)
{
    ui.videogridWidget->slotRemoveUser(userid, 0);
}

void MainWindow::slotDetachUserVideo(int userid)
{
    if(userid == 0)
        userid = ui.videogridWidget->selectedUser();

    if(userid<0)
        return;

    int streamid = ui.videogridWidget->getUserStreamID(userid);
    QSize size = ui.videogridWidget->getUserImageSize(userid);
    //if size if (0,0) then set size to invalid
    if(QSize(0,0) == size)
        size = QSize();

    ui.videogridWidget->slotRemoveUser(userid, 0);
    Q_ASSERT(m_uservideo.find(userid) == m_uservideo.end());

    slotNewUserVideoDlg(userid, size, streamid);
}

void MainWindow::slotNewUserVideoDlg(int userid, const QSize& size,
                                     int streamid)
{
    UserVideoDlg* dlg;
    if(size.isValid())
        dlg = new UserVideoDlg(userid, streamid, size, NULL);
    else
    {
        CaptureFormat fmt;
        if(TT_GetUserVideoFrame(ttInst, userid, NULL, 0, &fmt))
        {
            QSize tmp_size = QSize(fmt.nWidth, fmt.nHeight);
            dlg = new UserVideoDlg(userid, streamid, tmp_size, NULL);
        }
        else
            dlg = new UserVideoDlg(userid, streamid, NULL);
    }
    connect(this, SIGNAL(newVideoFrame(int,int)),
            dlg->ui.uservideoWidget, SLOT(slotNewVideoFrame(int,int)));
    connect(this, SIGNAL(newVideoFrame(int,int)), dlg,
            SLOT(slotNewVideoFrame(int,int)));
    connect(this, SIGNAL(userUpdate(int,int)), dlg, 
            SLOT(slotUserUpdate(int,int)));
    connect(dlg, SIGNAL(userVideoDlgClosing(int,int)), 
            SLOT(slotUserVideoDlgClosing(int,int)));
    connect(this, SIGNAL(preferencesModified()), dlg->ui.uservideoWidget,
            SLOT(slotUpdateVideoTextBox()));
    dlg->setAttribute(Qt::WA_DeleteOnClose);
    m_uservideo[userid] = dlg;
    if(streamid>0)
        m_videostreams[userid] = streamid;
    dlg->show();
}

void MainWindow::slotUserVideoDlgClosing(int userid, int streamid)
{
    uservideo_t::iterator ite = m_uservideo.find(userid);
    Q_ASSERT(ite != m_uservideo.end());
    if(ite == m_uservideo.end())
        return;

    m_uservideo.erase(ite);

    if(streamid>0)
        m_videostreams[userid] = streamid;

    User user;
    if(!TT_GetUser(ttInst, userid, &user))
        return;
    //check to see if video session is still active
    if((user.uUserState & USERSTATE_VIDEO) == USERSTATE_NONE)
        return;
    //add back to grid (if enabled)
    if(ttSettings->value(SETTINGS_DISPLAY_VIDEORETURNTOGRID,
                         SETTINGS_DISPLAY_VIDEORETURNTOGRID_DEFAULT).toBool())
        ui.videogridWidget->slotAddUser(userid, user.nChannelID);
    else if(m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS)
    {
        Subscriptions subs = SUBSCRIBE_VIDEO;
        if(user.uLocalSubscriptions & SUBSCRIBE_INTERCEPT_VIDEO)
            subs |= SUBSCRIBE_INTERCEPT_VIDEO;
        subscribeCommon(false, subs, userid);
        m_vid_exclude.insert(userid);
    }
}

void MainWindow::slotAddUserDesktop()
{
    int chanid = TT_GetMyChannelID(ttInst);
    QVector<int> users = ui.channelsWidget->getUsersInChannel(chanid);
    QMenu menu(this);

    //we don't want "myself" in the list
    int index = -1;
    if((index = users.indexOf(TT_GetMyUserID(ttInst))) != -1)
        users.remove(index);

    for(int i=0;i<users.size();i++)
    {
        if(m_userdesktop.find(users[i]) != m_userdesktop.end() ||
           ui.desktopgridWidget->userExists(users[i]))
           continue;

        User user;
        if(TT_GetUser(ttInst, users[i], &user) &&
           (user.uLocalSubscriptions & SUBSCRIBE_DESKTOP) == 0)
           menu.addAction(_Q(user.szNickname))->setData(users[i]);
    }

    if(menu.isEmpty())
        return;
    
    QAction* action = menu.exec(QCursor::pos());
    if(!action)
        return;

    int userid = action->data().toInt();
    if(userid != 0 && (m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS))
    {
        //this is basically the same as slotUsersSubscriptionsDesktop(..)
        subscribeCommon(true, SUBSCRIBE_DESKTOP, userid);
    }
}

void MainWindow::slotRemoveUserDesktop()
{
    int userid = ui.desktopgridWidget->selectedUser();
    if(userid<0)
    {
        QVector<int> users = ui.desktopgridWidget->activeUsers();
        QMenu menu(this);
        User user;
        for(int i=0;i<users.size();i++)
        {
            if(TT_GetUser(ttInst, users[i], &user))
                menu.addAction(_Q(user.szNickname))->setData(users[i]);
        }

        if(menu.isEmpty())
            return;

        QAction* action = menu.exec(QCursor::pos());
        if(!action)
            return;
        userid = action->data().toInt();
    }
    ui.desktopgridWidget->slotRemoveUser(userid, 0);
    if(userid>0)
        subscribeCommon(false, SUBSCRIBE_DESKTOP, userid);
}

void MainWindow::slotRemoveUserDesktop(int userid)
{
    //remove from grid
    ui.desktopgridWidget->slotRemoveUser(userid, 0);

    //close dialog
    userdesktop_t::iterator ite = m_userdesktop.find(userid);
    if(ite == m_userdesktop.end())
        return;
    (*ite)->close();
}

void MainWindow::slotAccessUserDesktop(bool enable)
{
    int userid = ui.desktopgridWidget->selectedUser();
    if(!userid)
        return;

    TextMessage msg;
    ZERO_STRUCT(msg);
    msg.nFromUserID = TT_GetMyUserID(ttInst);
    msg.nMsgType = MSGTYPE_CUSTOM;
    msg.nToUserID = userid;
    QString cmd = makeCustomCommand(TT_INTCMD_DESKTOP_ACCESS, 
                                    QString::number(enable));
    COPY_TTSTR(msg.szMessage, cmd);
    TT_DoTextMessage(ttInst, &msg);

    slotUpdateDesktopTabUI();
}

void MainWindow::slotDetachUserDesktop()
{
    int userid = ui.desktopgridWidget->selectedUser();
    if(userid<0)
        return;

    QSize imgsize = ui.desktopgridWidget->getUserImageSize(userid);
    if(!imgsize.isValid())
        return;

    ui.desktopgridWidget->slotRemoveUser(userid, 0);
    Q_ASSERT(m_userdesktop.find(userid) == m_userdesktop.end());
    slotDetachUserDesktop(userid, imgsize);
}

void MainWindow::slotDetachUserDesktop(int userid, const QSize& size)
{
    UserDesktopDlg* dlg;
    if(size.isValid())
        dlg = new UserDesktopDlg(userid, size, NULL);
    else
    {
        DesktopWindow wnd;
        if(!TT_GetUserDesktopWindow(ttInst, userid, NULL, NULL, &wnd))
            return;

        dlg = new UserDesktopDlg(userid, QSize(wnd.nWidth, wnd.nHeight), NULL);
    }

    connect(this, SIGNAL(newDesktopWindow(int,int)),
            dlg, SIGNAL(userDesktopWindowUpdate(int,int)));
    connect(this, SIGNAL(userUpdate(int,int)), dlg, 
            SLOT(slotUserUpdate(int,int)));
    connect(this, SIGNAL(userUpdate(int,int)), dlg, 
            SIGNAL(userUpdated(int,int)));
    connect(this, SIGNAL(userDesktopCursor(int,int)), dlg,
            SIGNAL(userDesktopCursorUpdate(int,int)));
    connect(dlg, SIGNAL(userDesktopDlgClosing(int)), 
            SLOT(slotUserDesktopDlgClosing(int)));
    connect(dlg, SIGNAL(userDesktopWindowEnded(int)),
            SLOT(slotRemoveUserDesktop(int)));
    dlg->setAttribute(Qt::WA_DeleteOnClose);
    m_userdesktop[userid] = dlg;
    dlg->show();
}

void MainWindow::slotUserDesktopDlgClosing(int userid)
{
    userdesktop_t::iterator ite = m_userdesktop.find(userid);
    Q_ASSERT(ite != m_userdesktop.end());
    if(ite == m_userdesktop.end())
        return;

    m_userdesktop.erase(ite);

    User user;
    if(!TT_GetUser(ttInst, userid, &user))
        return;
    //check to see if desktop session is still active
    if((user.uUserState & USERSTATE_DESKTOP) == USERSTATE_NONE)
        return;
    //add back to grid   
    ui.desktopgridWidget->slotAddUser(userid);
}

void MainWindow::slotUserUpdate(int userid, int /*chanid*/)
{
    User user, oldUser;
    if(TT_GetUser(ttInst, userid, &user) &&
       ui.channelsWidget->getUser(userid, oldUser))
    {
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_USER_MSG) !=
            (user.uPeerSubscriptions & SUBSCRIBE_USER_MSG))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionUserMessages->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_USER_MSG?
                     tr("On"):tr("Off")));
        }
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_CHANNEL_MSG) !=
            (user.uPeerSubscriptions & SUBSCRIBE_CHANNEL_MSG))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionChannelMessages->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_CHANNEL_MSG?
                     tr("On"):tr("Off")));
        }
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_BROADCAST_MSG) !=
            (user.uPeerSubscriptions & SUBSCRIBE_BROADCAST_MSG))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionBroadcastMessages->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_BROADCAST_MSG?
                     tr("On"):tr("Off")));
        }
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_AUDIO) !=
            (user.uPeerSubscriptions & SUBSCRIBE_AUDIO))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionAudio->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_AUDIO?
                     tr("On"):tr("Off")));
        }
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_VIDEO) !=
            (user.uPeerSubscriptions & SUBSCRIBE_VIDEO))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionVideo->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_VIDEO?
                     tr("On"):tr("Off")));
        }
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_DESKTOP) !=
            (user.uPeerSubscriptions & SUBSCRIBE_DESKTOP))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionDesktop->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_DESKTOP?
                     tr("On"):tr("Off")));
        }
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_DESKTOPINPUT) !=
            (user.uPeerSubscriptions & SUBSCRIBE_DESKTOPINPUT))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionDesktopInput->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_DESKTOPINPUT?
                     tr("On"):tr("Off")));
        }
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_USER_MSG) !=
            (user.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_USER_MSG))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionInterceptUserMessages->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_USER_MSG?
                     tr("On"):tr("Off")));
        }
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_CHANNEL_MSG) !=
            (user.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_CHANNEL_MSG))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionInterceptChannelMessages->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_CHANNEL_MSG?
                     tr("On"):tr("Off")));
        }
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_AUDIO) !=
            (user.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_AUDIO))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionInterceptAudio->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_AUDIO?
                     tr("On"):tr("Off")));
        }
        if((oldUser.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_VIDEO) !=
            (user.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_VIDEO))
        {
            addStatusMsg(tr("%1 changed subscription \"%2\" to: %3")
                .arg(_Q(user.szNickname))
                .arg(MENUTEXT(ui.actionInterceptVideo->text()))
                .arg(user.uPeerSubscriptions & SUBSCRIBE_INTERCEPT_VIDEO?
                     tr("On"):tr("Off")));
        }
    }
}

void MainWindow::slotEnableQuestionMode(bool checked)
{
    if(checked)
        m_statusmode |= STATUSMODE_QUESTION;
    else
        m_statusmode &= ~STATUSMODE_QUESTION;

    QString statusmsg = ttSettings->value(SETTINGS_GENERAL_STATUSMESSAGE).toString();
    if(TT_GetFlags(ttInst) & CLIENT_AUTHORIZED)
        TT_DoChangeStatus(ttInst, m_statusmode, _W(statusmsg));
}

void MainWindow::slotUpdateVideoCount(int count)
{
    if(count == 0)
        ui.tabWidget->setTabText(TAB_VIDEO, tr("Video"));
    else
        ui.tabWidget->setTabText(TAB_VIDEO, tr("Video (%1)").arg(count));
}

void MainWindow::slotUpdateDesktopCount(int count)
{
    if(count == 0)
        ui.tabWidget->setTabText(TAB_DESKTOP, tr("Desktops"));
    else
        ui.tabWidget->setTabText(TAB_DESKTOP, tr("Desktops (%1)").arg(count));
}

void MainWindow::slotMasterVolumeChanged(int value)
{
    if(value <= SOUND_VOLUME_MAX) //compensate for soft volume
        TT_SetSoundOutputVolume(ttInst, value);
    updateUserGainLevel(-1);
}

void MainWindow::slotMicrophoneGainChanged(int value)
{
    TT_SetSoundInputGainLevel(ttInst, value);
}

void MainWindow::slotVoiceActivationLevelChanged(int value)
{
    TT_SetVoiceActivationLevel(ttInst, value);
}

void MainWindow::slotEnableAudioRecording(bool checked)
{
    quint32 old_mode = m_audiostorage_mode;

    if(checked &&
       ttSettings->value(SETTINGS_GENERAL_AUDIOSTORAGE_MODE, 
                         AUDIOSTORAGE_NONE).toUInt() == AUDIOSTORAGE_NONE)
        slotUsersStoreAudioToDisk(checked);
    else
    {
        quint32 new_mode = ttSettings->value(SETTINGS_GENERAL_AUDIOSTORAGE_MODE, 
                                             AUDIOSTORAGE_NONE).toUInt();
        if(old_mode & AUDIOSTORAGE_SINGLEFILE)
        {
            updateAudioStorage(false, AUDIOSTORAGE_SINGLEFILE);
            m_audiostorage_mode &= ~AUDIOSTORAGE_SINGLEFILE;
        }
        if(old_mode & AUDIOSTORAGE_SEPARATEFILES)
        {
            updateAudioStorage(false, AUDIOSTORAGE_SEPARATEFILES);
            m_audiostorage_mode &= ~AUDIOSTORAGE_SEPARATEFILES;
        }

        if(checked)
        {
            if(new_mode & AUDIOSTORAGE_SINGLEFILE)
            {
                updateAudioStorage(true, AUDIOSTORAGE_SINGLEFILE);
                m_audiostorage_mode |= AUDIOSTORAGE_SINGLEFILE;
            }
            if(new_mode & AUDIOSTORAGE_SEPARATEFILES)
            {
                updateAudioStorage(true, AUDIOSTORAGE_SEPARATEFILES);
                m_audiostorage_mode |= AUDIOSTORAGE_SEPARATEFILES;
            }
        }
        else
            m_audiostorage_mode = AUDIOSTORAGE_NONE;
    }
}

void MainWindow::slotTrayIconChange(QSystemTrayIcon::ActivationReason reason)
{
    if(reason == QSystemTrayIcon::Trigger)
    {
        showNormal();
        activateWindow(); //if we don't call this the window will end up in 
                          //the background for some reason (on win32)
    }
}

void MainWindow::slotLoadTTFile(const QString& filepath)
{
    QFile ttfile(QDir::fromNativeSeparators(filepath));
    if(ttfile.open(QFile::ReadOnly))
    {
        QByteArray data = ttfile.readAll();
        QDomDocument doc(TTFILE_ROOT);
        if(doc.setContent(data))
        {
            QDomElement rootElement(doc.documentElement());
            QDomElement element = rootElement.firstChildElement("host");
            HostEntry entry;
            if(GetServerEntry(element, entry))
            {
                AddLatestHost(entry);
                Disconnect();
                Connect();
                return;
            }
        }
    }
    QMessageBox::information(this, tr("Load File"), 
        tr("Failed to load file %1").arg(filepath));
}

void MainWindow::slotHttpUpdateReply(QNetworkReply* reply)
{
    Q_ASSERT(m_http_manager);
    QByteArray data = reply->readAll();

    QDomDocument doc("foo");
    if(doc.setContent(data))
    {
        QString version = NewVersionAvailable(doc);
        if(version.size())
            addStatusMsg(tr("New version available: %1").arg(version));
    }

    m_http_manager->deleteLater();
}

void MainWindow::slotClosedOnlineUsersDlg(int)
{
    m_onlineusersdlg = NULL;
}

void MainWindow::slotClosedServerStatsDlg(int)
{
    m_serverstatsdlg = NULL;
}

void MainWindow::slotClosedUserAccountsDlg(int)
{
    m_useraccountsdlg = NULL;
}

void MainWindow::slotClosedBannedUsersDlg(int)
{
    m_bannedusersdlg = NULL;
}

