/*
 * Copyright (c) 2005-2014, BearWare.dk
 * 
 * Contact Information:
 *
 * Bjoern D. Rasmussen
 * Skanderborgvej 40 4-2
 * DK-8000 Aarhus C
 * Denmark
 * Email: contact@bearware.dk
 * Phone: +45 20 20 54 59
 * Web: http://www.bearware.dk
 *
 * This source code is part of the TeamTalk 4 SDK owned by
 * BearWare.dk. All copyright statements may not be removed 
 * or altered from any source distribution. If you use this
 * software in a product, an acknowledgment in the product 
 * documentation is required.
 *
 */

#include "onlineusersdlg.h"
#include "mainwindow.h"

#include <QHeaderView>

extern TTInstance* ttInst;

OnlineUsersDlg::OnlineUsersDlg(QWidget* parent/* = 0 */)
: QDialog(parent, Qt::Dialog | Qt::WindowMinMaxButtonsHint | Qt::WindowSystemMenuHint)
{
    ui.setupUi(this);
    setWindowIcon(QIcon(APPICON));

    ui.treeView->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(ui.treeView, SIGNAL(customContextMenuRequested(const QPoint&)),
            SLOT(slotTreeContextMenu(const QPoint&)));

    m_model = new OnlineUsersModel(this);
    m_proxyModel = new QSortFilterProxyModel(this);
    m_proxyModel->setSourceModel(m_model);
    ui.treeView->setModel(m_proxyModel);

    ui.treeView->header()->resizeSection(COLUMN_USERID, 40);
    //ui.treeView->header()->resizeSection(COLUMN_STATUSMSG, 70);
    ui.treeView->header()->resizeSection(COLUMN_USERNAME, 65);

    m_model->resetUsers();
    updateTitle();
}

void OnlineUsersDlg::updateTitle()
{
    QString title = tr("Users Currently on Server");
    title += QString(" (%1)").arg(m_model->rowCount());
    ui.groupBox->setTitle(title);
}

void OnlineUsersDlg::slotUserLoggedIn(int userid)
{
    m_model->addUser(userid);
    updateTitle();
}

void OnlineUsersDlg::slotUserLoggedOut(int userid)
{
    m_model->removeUser(userid);
    updateTitle();
}

void OnlineUsersDlg::slotUserJoin(int userid, int /*channelid*/)
{
    m_model->updateUser(userid);
    QModelIndex index = m_model->userRow(userid);
    if(index.isValid())
        ui.treeView->update(index);
}

void OnlineUsersDlg::slotUserUpdate(int userid, int /*channelid*/)
{
    m_model->updateUser(userid);
    QModelIndex index = m_model->userRow(userid);
    if(index.isValid())
        ui.treeView->update(index);
}

void OnlineUsersDlg::slotUserRemove(int userid, int /*channelid*/)
{
    m_model->updateUser(userid);
    QModelIndex index = m_model->userRow(userid);
    if(index.isValid())
        ui.treeView->update(index);
}

void OnlineUsersDlg::slotTreeContextMenu(const QPoint& /*point*/)
{
    QMenu menu(this);
    QAction* info = menu.addAction(tr("&View User Information"));
    QAction* messages = menu.addAction(tr("M&essages"));
    //QAction* mute = menu.addAction(tr("&Mute"));
    //QAction* volume = menu.addAction(tr("&Volume"));
    QAction* op = menu.addAction(tr("&Op"));
    QAction* kick = menu.addAction(tr("&Kick"));
    QAction* ban = menu.addAction(tr("Kick and &Ban"));    
    QAction* stream = menu.addAction(tr("Stream Audio &File to User"));
    QAction* action = menu.exec(QCursor::pos());

    QItemSelectionModel* selModel = ui.treeView->selectionModel();
    QModelIndexList indexes = selModel->selectedRows();
    QVector<int> userids, chanids;
    for(int i=0;i<indexes.size();i++)
    {
        //QModelIndex index = ui.treeView->indexAt(point);
        QModelIndex index = m_proxyModel->mapToSource(indexes[i]);
        if(!index.isValid())
            return;

        int userid = index.internalId();
        User user;
        if(!TT_GetUser(ttInst, userid, &user))
            continue;
        userids.push_back(user.nUserID);
        chanids.push_back(user.nChannelID);
        qDebug() << "Processing " << user.nUserID;
    }
    
    for(int i=0;i<userids.size();i++)
    {
        if(action == info)
            emit(viewUserInformation(userids[i]));
        else if(action == messages)
            emit(sendUserMessage(userids[i]));
        //else if(action == mute)
        //    emit(muteUser(userid, !(user.uUserState & USERSTATE_MUTE)));
        //else if(action == volume)
        //    emit(changeUserVolume(userid));
        else if(action == op)
            emit(opUser(userids[i], chanids[i]));
        else if(action == kick)
            emit(kickUser(userids[i], chanids[i]));
        else if(action == ban)
            emit(kickbanUser(userids[i], chanids[i]));
        else if(action == stream)
            emit(streamfileToUser(userids[i]));
    }
}
