/*
 * Copyright (c) 2005-2014, BearWare.dk
 * 
 * Contact Information:
 *
 * Bjoern D. Rasmussen
 * Skanderborgvej 40 4-2
 * DK-8000 Aarhus C
 * Denmark
 * Email: contact@bearware.dk
 * Phone: +45 20 20 54 59
 * Web: http://www.bearware.dk
 *
 * This source code is part of the TeamTalk 4 SDK owned by
 * BearWare.dk. All copyright statements may not be removed 
 * or altered from any source distribution. If you use this
 * software in a product, an acknowledgment in the product 
 * documentation is required.
 *
 */

#include "serverpropertiesdlg.h"
#include "mainwindow.h"

#include <QMessageBox>

extern TTInstance* ttInst;

ServerPropertiesDlg::ServerPropertiesDlg(QWidget * parent/* = 0*/)
    : QDialog(parent, Qt::Dialog | Qt::WindowTitleHint)
{
    ui.setupUi(this);
    setWindowIcon(QIcon(APPICON));

    bool admin = (TT_GetMyUserType(ttInst) & USERTYPE_ADMIN);

    if(!admin)
        ui.buttonBox->setStandardButtons(QDialogButtonBox::Close);

    connect(ui.motdChkBox, SIGNAL(clicked(bool)), SLOT(slotShowMOTDVars(bool)));
    connect(this, SIGNAL(accepted()), SLOT(slotAccepted()));
    ZERO_STRUCT(m_srvprop);

    TT_GetServerProperties(ttInst, &m_srvprop);

    ui.servernameEdit->setText(_Q(m_srvprop.szServerName));
    ui.serverpasswdEdit->setText(_Q(m_srvprop.szServerPasswd));
    ui.maxusersSpinBox->setValue(m_srvprop.nMaxUsers);
    ui.motdTextEdit->setPlainText(_Q(m_srvprop.szMOTD));
    ui.tcpportSpinBox->setValue(m_srvprop.nTcpPort);
    ui.udpportSpinBox->setValue(m_srvprop.nUdpPort);
    ui.usertimeoutSpinBox->setValue(m_srvprop.nUserTimeout);
    ui.autosaveBox->setChecked(m_srvprop.bAutoSave);
    ui.createchannelsBox->setChecked(m_srvprop.uUserRights & USERRIGHT_CHANNEL_CREATION);
    ui.becomeoperatorsBox->setChecked(m_srvprop.uUserRights & USERRIGHT_CHANNEL_OPERATORS);
    ui.guestloginBox->setChecked(m_srvprop.uUserRights & USERRIGHT_GUEST_LOGIN);
    ui.multiloginBox->setChecked(m_srvprop.uUserRights & USERRIGHT_DOUBLE_LOGIN);
    ui.viewallusersBox->setChecked(m_srvprop.uUserRights & USERRIGHT_VIEW_ALL_USERS);
    ui.viewallusersBox->setEnabled(false); //this cannot be changed while the server is running
    ui.clientbroadcastBox->setChecked(m_srvprop.uUserRights & USERRIGHT_CLIENT_BROADCAST);
    ui.chancmdBox->setChecked(m_srvprop.uUserRights & USERRIGHT_CHANNEL_COMMANDS);
    ui.subscriptionsBox->setChecked(m_srvprop.uUserRights & USERRIGHT_SUBSCRIPTIONS);
    ui.forwardaudioBox->setChecked(m_srvprop.uUserRights & USERRIGHT_FORWARD_AUDIO);
    ui.forwardvideoBox->setChecked(m_srvprop.uUserRights & USERRIGHT_FORWARD_VIDEO);
    ui.forwarddesktopBox->setChecked(m_srvprop.uUserRights & USERRIGHT_FORWARD_DESKTOP);
    ui.forwarddesktopaccessBox->setChecked(m_srvprop.uUserRights & USERRIGHT_FORWARD_DESKTOPINPUT);
    ui.forwarddesktopBox->setEnabled(VersionSameOrLater(_Q(m_srvprop.szServerVersion), TT_SERVER_DESKTOP_MIN));
    ui.maxloginattemptsSpinBox->setValue(m_srvprop.nMaxLoginAttempts);
    ui.maxloginattemptsSpinBox->setEnabled(VersionSameOrLater(_Q(m_srvprop.szServerProtocolVersion), "4.6"));
    ui.maxiploginsSpinBox->setValue(m_srvprop.nMaxLoginsPerIPAddress);
    ui.maxiploginsSpinBox->setEnabled(VersionSameOrLater(_Q(m_srvprop.szServerProtocolVersion), "4.6"));
    ui.audmaxbpsSpinBox->setValue(m_srvprop.nAudioCodecBpsLimit / 1000);
    ui.audtxSpinBox->setValue(m_srvprop.nMaxAudioTxPerSecond/1024);
    ui.videotxSpinBox->setValue(m_srvprop.nMaxVideoTxPerSecond/1024);
    ui.desktoptxSpinBox->setValue(m_srvprop.nMaxDesktopTxPerSecond/1024);
    ui.totaltxSpinBox->setValue(m_srvprop.nMaxTotalTxPerSecond/1024);
    ui.serverversionEdit->setText(_Q(m_srvprop.szServerVersion));
    if(!VersionSameOrLater(_Q(m_srvprop.szServerVersion), TT_SERVER_DESKTOP_MIN))
        ui.motdChkBox->hide();


    if(!admin)
    {
        ui.servernameEdit->setEnabled(false);
        ui.serverpasswdEdit->setEnabled(false);
        ui.maxusersSpinBox->setEnabled(false);
        ui.motdTextEdit->setEnabled(false);
        ui.tcpportSpinBox->setEnabled(false);
        ui.udpportSpinBox->setEnabled(false);
        ui.usertimeoutSpinBox->setEnabled(false);
        ui.autosaveBox->setEnabled(false);
        ui.createchannelsBox->setEnabled(false);
        ui.becomeoperatorsBox->setEnabled(false);
        ui.guestloginBox->setEnabled(false);
        ui.multiloginBox->setEnabled(false);
        ui.viewallusersBox->setEnabled(false);
        ui.clientbroadcastBox->setEnabled(false);
        ui.chancmdBox->setEnabled(false);
        ui.subscriptionsBox->setEnabled(false);
        ui.forwardaudioBox->setEnabled(false);
        ui.forwardvideoBox->setEnabled(false);
        ui.forwarddesktopBox->setEnabled(false);
        ui.forwarddesktopaccessBox->setEnabled(false);
        ui.maxloginattemptsSpinBox->setEnabled(false);
        ui.maxiploginsSpinBox->setEnabled(false);
        ui.audmaxbpsSpinBox->setEnabled(false);
        ui.audtxSpinBox->setEnabled(false);
        ui.videotxSpinBox->setEnabled(false);
        ui.desktoptxSpinBox->setEnabled(false);
        ui.totaltxSpinBox->setEnabled(false);
        ui.motdChkBox->hide();
    }
}

void ServerPropertiesDlg::slotAccepted()
{
    COPY_TTSTR(m_srvprop.szServerName, ui.servernameEdit->text());
    COPY_TTSTR(m_srvprop.szServerPasswd, ui.serverpasswdEdit->text());
    m_srvprop.nMaxUsers = ui.maxusersSpinBox->value();
    m_srvprop.nMaxLoginAttempts = ui.maxloginattemptsSpinBox->value();
    m_srvprop.nMaxLoginsPerIPAddress = ui.maxiploginsSpinBox->value();
    if(_Q(m_srvprop.szMOTDRaw).size() &&
       _Q(m_srvprop.szMOTDRaw) != ui.motdTextEdit->toPlainText())
    {
        if(QMessageBox::question(this,  tr("Message of the day"), 
                    tr("Change message of the day?"), 
                    QMessageBox::Yes | QMessageBox::No) == QMessageBox::Yes)
           COPY_TTSTR(m_srvprop.szMOTD, ui.motdTextEdit->toPlainText());
        else
           COPY_TTSTR(m_srvprop.szMOTD, _Q(m_srvprop.szMOTDRaw));
    }
    else
        COPY_TTSTR(m_srvprop.szMOTD, ui.motdTextEdit->toPlainText());
    m_srvprop.nTcpPort = ui.tcpportSpinBox->value();
    m_srvprop.nUdpPort = ui.udpportSpinBox->value();
    m_srvprop.nUserTimeout = ui.usertimeoutSpinBox->value();
    m_srvprop.bAutoSave = ui.autosaveBox->isChecked();
    m_srvprop.nAudioCodecBpsLimit = ui.audmaxbpsSpinBox->value() * 1000;
    m_srvprop.nMaxAudioTxPerSecond = ui.audtxSpinBox->value()*1024;
    m_srvprop.nMaxVideoTxPerSecond = ui.videotxSpinBox->value()*1024;
    m_srvprop.nMaxDesktopTxPerSecond = ui.desktoptxSpinBox->value()*1024;
    m_srvprop.nMaxTotalTxPerSecond = ui.totaltxSpinBox->value()*1024;

    if(ui.createchannelsBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_CHANNEL_CREATION;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_CHANNEL_CREATION;
    if(ui.becomeoperatorsBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_CHANNEL_OPERATORS;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_CHANNEL_OPERATORS;
    if(ui.guestloginBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_GUEST_LOGIN;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_GUEST_LOGIN;
    if(ui.multiloginBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_DOUBLE_LOGIN;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_DOUBLE_LOGIN;
    if(ui.viewallusersBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_VIEW_ALL_USERS;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_VIEW_ALL_USERS;
    if(ui.clientbroadcastBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_CLIENT_BROADCAST;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_CLIENT_BROADCAST;
    if(ui.chancmdBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_CHANNEL_COMMANDS;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_CHANNEL_COMMANDS;
    if(ui.subscriptionsBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_SUBSCRIPTIONS;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_SUBSCRIPTIONS;
    if(ui.forwardaudioBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_FORWARD_AUDIO;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_FORWARD_AUDIO;
    if(ui.forwardvideoBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_FORWARD_VIDEO;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_FORWARD_VIDEO;
    if(ui.forwarddesktopBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_FORWARD_DESKTOP;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_FORWARD_DESKTOP;
    if(ui.forwarddesktopaccessBox->isChecked())
        m_srvprop.uUserRights |= USERRIGHT_FORWARD_DESKTOPINPUT;
    else
        m_srvprop.uUserRights &= ~USERRIGHT_FORWARD_DESKTOPINPUT;
    TT_DoUpdateServer(ttInst, &m_srvprop);
}

void ServerPropertiesDlg::slotShowMOTDVars(bool checked)
{
    if(checked)
        ui.motdTextEdit->setPlainText(_Q(m_srvprop.szMOTDRaw));
    else
        ui.motdTextEdit->setPlainText(_Q(m_srvprop.szMOTD));
}
