/*
 * Copyright (c) 2005-2014, BearWare.dk
 * 
 * Contact Information:
 *
 * Bjoern D. Rasmussen
 * Skanderborgvej 40 4-2
 * DK-8000 Aarhus C
 * Denmark
 * Email: contact@bearware.dk
 * Phone: +45 20 20 54 59
 * Web: http://www.bearware.dk
 *
 * This source code is part of the TeamTalk 4 SDK owned by
 * BearWare.dk. All copyright statements may not be removed 
 * or altered from any source distribution. If you use this
 * software in a product, an acknowledgment in the product 
 * documentation is required.
 *
 */

#include "streammediafiledlg.h"
#include "mainwindow.h"
#include "preferencesdlg.h"

#include <QFileDialog>

extern QSettings* ttSettings;
extern TTInstance* ttInst;

StreamMediaFileDlg::StreamMediaFileDlg(QWidget* parent/* = 0*/)
    : QDialog(parent, Qt::Dialog | Qt::WindowTitleHint)
{
    ui.setupUi(this);
    setWindowIcon(QIcon(APPICON));

    ui.mediafileEdit->setText(ttSettings->value(SETTINGS_GENERAL_STREAMMEDIA_FILENAME).toString());

    ui.vidcodecBox->addItem("Theora", THEORA_CODEC);
    ui.vidqualitySlider->setValue(ttSettings->value(SETTINGS_GENERAL_THEORA_QUALITY,
                                                    DEFAULT_THEORA_QUALITY).toInt());
    ui.vidbitrateSpinBox->setValue(ttSettings->value(SETTINGS_GENERAL_THEORA_BITRATE,
                                                     DEFAULT_THEORA_BITRATE).toInt());
    slotVideoQualityChanged(ui.vidqualitySlider->value());

    connect(this, SIGNAL(accepted()), SLOT(slotAccepted()));
    connect(ui.toolButton, SIGNAL(clicked()), SLOT(slotSelectFile()));
    connect(ui.vidqualitySlider, SIGNAL(valueChanged(int)), 
            SLOT(slotVideoQualityChanged(int)));

    showMediaFormatInfo();
}

void StreamMediaFileDlg::slotAccepted()
{
    ttSettings->setValue(SETTINGS_GENERAL_STREAMMEDIA_FILENAME,
                         ui.mediafileEdit->text());
    ttSettings->setValue(SETTINGS_GENERAL_STREAMMEDIA_AUDIO,
                         ui.audioChkBox->isChecked());
    ttSettings->setValue(SETTINGS_GENERAL_STREAMMEDIA_VIDEO,
                         ui.videoChkBox->isChecked());
    ttSettings->setValue(SETTINGS_GENERAL_STREAMMEDIA_LOOP,
                         ui.loopChkBox->isChecked());
    ttSettings->setValue(SETTINGS_GENERAL_THEORA_QUALITY, ui.vidqualitySlider->value());
    ttSettings->setValue(SETTINGS_GENERAL_THEORA_BITRATE, ui.vidbitrateSpinBox->value());
}

void StreamMediaFileDlg::slotSelectFile()
{
    QFileInfo fileinfo(QDir::fromNativeSeparators(ui.mediafileEdit->text()));

    QString fileName = QFileDialog::getOpenFileName(this,
                        tr("Open Media File"),
                        QDir::toNativeSeparators(fileinfo.dir().absolutePath()),
                        tr("Media files (*.*)"));

    fileName = QDir::toNativeSeparators(fileName);
    ui.mediafileEdit->setText(fileName);

    showMediaFormatInfo();
}

void StreamMediaFileDlg::showMediaFormatInfo()
{
    QString filename = ui.mediafileEdit->text();
    QString audio, video;

    MediaFileInfo mediaFile;
    ZERO_STRUCT(mediaFile);

    if(filename.size() &&
       TT_GetMediaFileInfo(_W(filename), &mediaFile))
    {
        if(mediaFile.audioFmt.nAudioFmt != AFF_NONE)
        {
            QString channels;
            if(mediaFile.audioFmt.nChannels == 2)
                channels = tr("Stereo");
            else if(mediaFile.audioFmt.nChannels == 1)
                channels = tr("Mono");
            else
                channels = tr("%1 audio channels").arg(mediaFile.audioFmt.nChannels);

            audio = tr("%1 Hz, %2").arg(mediaFile.audioFmt.nSampleRate).arg(channels);
        }
        else
            audio = tr("Unknown format");

        double fps = (double)mediaFile.videoFmt.nFPS_Numerator / (double)mediaFile.videoFmt.nFPS_Denominator;
        if(mediaFile.videoFmt.picFourCC != FOURCC_NONE)
            video = tr("%1x%2 %3 FPS").arg(mediaFile.videoFmt.nWidth).arg(mediaFile.videoFmt.nHeight).arg(fps, 0, 'f', 1);
        else
            video = tr("Unknown format");

        int duration_sec = mediaFile.uDurationMSec / 1000;
        ui.durationLabel->setText(QString("%1:%2:%3")
                                  .arg(duration_sec / 3600, 2 , 10, QChar('0'))
                                  .arg((duration_sec % 3600) / 60, 2 , 10, QChar('0'))
                                  .arg(duration_sec % 60, 2 , 10, QChar('0')));
        ui.vidcodecBox->setEnabled(mediaFile.videoFmt.picFourCC != FOURCC_NONE);
        ui.vidqualitySlider->setEnabled(mediaFile.videoFmt.picFourCC != FOURCC_NONE);
        ui.vidbitrateSpinBox->setEnabled(mediaFile.videoFmt.picFourCC != FOURCC_NONE);

    }

    ui.audioLabel->setText(audio);
    ui.videoLabel->setText(video);

    ui.audioChkBox->setEnabled(mediaFile.audioFmt.nAudioFmt != AFF_NONE);
    ui.audioChkBox->setChecked(mediaFile.audioFmt.nAudioFmt != AFF_NONE);

    ui.videoChkBox->setEnabled(mediaFile.videoFmt.picFourCC != FOURCC_NONE);
    ui.videoChkBox->setChecked(mediaFile.videoFmt.picFourCC != FOURCC_NONE);

    ui.okBtn->setEnabled(mediaFile.audioFmt.nAudioFmt != AFF_NONE ||
                         mediaFile.videoFmt.picFourCC != FOURCC_NONE);
    
    //TT cannot stream video if webcam is already running
    if(mediaFile.videoFmt.picFourCC != FOURCC_NONE &&
       TT_GetFlags(ttInst) & CLIENT_VIDEO_READY)
    {
        ui.videoChkBox->setEnabled(false);
        ui.videoChkBox->setChecked(false);
    }
}

void StreamMediaFileDlg::slotVideoQualityChanged(int value)
{
    ui.vidqualityLabel->setText(QString::number(value));
}

