using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using BearWare;

namespace TeamTalk.NET
{
    public partial class DesktopDlg : Form
    {
        TeamTalk4 ttclient;
        int userid;
        int sessionid = 0;
        Bitmap bmp;
        int unsubscribe_cmdid = 0;

        public DesktopDlg(TeamTalk4 tt, int userid)
        {
            ttclient = tt;
            this.userid = userid;

            InitializeComponent();
            this.CenterToScreen();
            
            ttclient.OnUserDesktopWindow += new TeamTalk4.NewDesktopWindow(ttclient_OnUserDesktopWindow);
        }

        private void DesktopDlg_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (unsubscribe_cmdid == 0)
            {
                if (MessageBox.Show("Do you wish to stop receiving desktops from this user?", "Close Desktop",
                    MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    //we wait for the server to repond to us that we'll no 
                    //longer receive video from this user.
                    ttclient.OnCmdProcessing += new TeamTalk4.CommandProcessing(ttclient_OnCmdProcessing);

                    //figure out how we're subscribing to video
                    Subscription subtype = Subscription.SUBSCRIBE_NONE;
                    User user;
                    if (ttclient.GetUser(this.userid, out user))
                    {
                        if ((user.uLocalSubscriptions & Subscription.SUBSCRIBE_DESKTOP) == Subscription.SUBSCRIBE_DESKTOP)
                            subtype = Subscription.SUBSCRIBE_DESKTOP;
                        else if ((user.uLocalSubscriptions & Subscription.SUBSCRIBE_INTERCEPT_DESKTOP) == Subscription.SUBSCRIBE_INTERCEPT_DESKTOP)
                            subtype = Subscription.SUBSCRIBE_INTERCEPT_DESKTOP;
                    }

                    //store the command ID we get back from the client instance
                    unsubscribe_cmdid = ttclient.DoUnsubscribe(this.userid, subtype);
                    e.Cancel = true;
                }
                else
                    e.Cancel = true;
            }
            else
                ttclient.OnUserDesktopWindow -= ttclient_OnUserDesktopWindow;
        }

        void ttclient_OnCmdProcessing(int nCmdID, bool bComplete)
        {
            if (nCmdID == unsubscribe_cmdid && bComplete)
            {
                //now that the 'DoUnsubscribe' command has completed we can close the dialog.
                this.Close();
            }
        }

        void ttclient_OnUserDesktopWindow(int nUserID, int nSessionID)
        {
            if (nUserID != userid)
                return;
        }

        protected override void OnPaintBackground(PaintEventArgs e)
        {
            if (bmp != null)
            {
                e.Graphics.DrawImage(bmp, 0, 0, this.ClientRectangle.Width,
                    this.ClientRectangle.Height);
            }
            else
            {
                base.OnPaintBackground(e);
            }
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            //query bitmap size
            int bmpsize = 0;
            DesktopWindow deskwnd = new DesktopWindow();
            if (!ttclient.GetUserDesktopWindow(userid, IntPtr.Zero, ref bmpsize, ref deskwnd))
            {
                this.Close();
                return;
            }

            if (ClientSize.Height != deskwnd.nHeight || ClientSize.Width != deskwnd.nWidth)
            {
                ClientSize = new Size(deskwnd.nWidth, deskwnd.nHeight);
                //int width_diff = deskwnd.nWidth - ClientSize.Width;
                //int height_diff = deskwnd.nHeight - ClientSize.Height;
                //this.Size.Width += width_diff;
                //this.Size.Height += height_diff;
            }

            //session id changed, so we need to create new bitmap.
            if(deskwnd.nSessionID != sessionid)
            {
                if (bmp != null)
                    bmp.Dispose();
                bmp = null;

                switch(deskwnd.bmpFormat)
                {
                    case BitmapFormat.BMP_RGB8_PALETTE:
                        bmp = new Bitmap(deskwnd.nWidth, deskwnd.nHeight, PixelFormat.Format8bppIndexed);
                        ColorPalette pal = bmp.Palette;
                        for (int i = 0; i < 256; i++)
                            pal.Entries[i] = TeamTalk4.Palette_GetColorTable(BitmapFormat.BMP_RGB8_PALETTE, i);
                        bmp.Palette = pal;
                        break;
                    case BitmapFormat.BMP_RGB16_555:
                        bmp = new Bitmap(deskwnd.nWidth, deskwnd.nHeight, PixelFormat.Format16bppRgb555);
                        break;
                    case BitmapFormat.BMP_RGB24 :
                        bmp = new Bitmap(deskwnd.nWidth, deskwnd.nHeight, PixelFormat.Format24bppRgb);
                        break;
                    case BitmapFormat.BMP_RGB32:
                        bmp = new Bitmap(deskwnd.nWidth, deskwnd.nHeight, PixelFormat.Format32bppRgb);
                        break;
                }
            }

            if (bmp == null)
                return;

            // Lock the bitmap's bits.  
            System.Drawing.Rectangle rect = new System.Drawing.Rectangle(0, 0, bmp.Width, bmp.Height);
            BitmapData bmpData = bmp.LockBits(rect, ImageLockMode.ReadWrite, bmp.PixelFormat);

            // Get the address of the first line.
            IntPtr ptr = bmpData.Scan0;

            Debug.Assert(bmpsize == Math.Abs(bmpData.Stride) * bmp.Height);
            bmpsize = Math.Abs(bmpData.Stride) * bmp.Height;
            
            if (ttclient.GetUserDesktopWindow(userid, ptr, ref bmpsize, ref deskwnd))
                Invalidate();
            bmp.UnlockBits(bmpData);

        }

        DesktopInput ToTTKeyCode(MouseEventArgs e, DesktopKeyState state)
        {
            DesktopInput input;
            switch (e.Button)
            {
                case MouseButtons.Left:
                    input.uKeyCode = TeamTalk4.TT_DESKTOPINPUT_KEYCODE_LMOUSEBTN;
                    break;
                case MouseButtons.Right:
                    input.uKeyCode = TeamTalk4.TT_DESKTOPINPUT_KEYCODE_RMOUSEBTN;
                    break;
                case MouseButtons.Middle:
                    input.uKeyCode = TeamTalk4.TT_DESKTOPINPUT_KEYCODE_MMOUSEBTN;
                    break;
                default :
                    input.uKeyCode = TeamTalk4.TT_DESKTOPINPUT_KEYCODE_IGNORE;
                    break;
            }
            input.uKeyState = state;
            input.uMousePosX = (ushort)e.X;
            input.uMousePosY = (ushort)e.Y;
            return input;
        }

        DesktopInput ToTTKeyCode(KeyEventArgs e, DesktopKeyState state)
        {
            DesktopInput input;
            //TODO: convert to virtual key.
            input.uKeyCode = (uint)e.KeyValue;
            input.uKeyState = state;
            input.uMousePosX = input.uMousePosY = TeamTalk4.TT_DESKTOPINPUT_MOUSEPOS_IGNORE;
            return input;
        }

        DesktopInput[] ToTTKey(DesktopInput input)
        {
            DesktopInput[] inputs = new DesktopInput[1];
            inputs[0] = input;
            DesktopInput[] translated;
            WindowsHelper.KeyTranslate(TTKeyTranslate.TTKEY_WINKEYCODE_TO_TTKEYCODE,
                                       inputs, out translated);
            return translated;
        }

        private void DesktopDlg_MouseDown(object sender, MouseEventArgs e)
        {
            DesktopInput input = ToTTKeyCode(e, DesktopKeyState.DESKTOPKEYSTATE_DOWN);
            ttclient.SendDesktopInput(userid, ToTTKey(input));
        }

        private void DesktopDlg_MouseMove(object sender, MouseEventArgs e)
        {
            DesktopInput input = ToTTKeyCode(e, DesktopKeyState.DESKTOPKEYSTATE_NONE);
            ttclient.SendDesktopInput(userid, ToTTKey(input));
        }

        private void DesktopDlg_MouseUp(object sender, MouseEventArgs e)
        {
            DesktopInput input = ToTTKeyCode(e, DesktopKeyState.DESKTOPKEYSTATE_UP);
            ttclient.SendDesktopInput(userid, ToTTKey(input));
        }

        private void DesktopDlg_KeyDown(object sender, KeyEventArgs e)
        {
            DesktopInput input = ToTTKeyCode(e, DesktopKeyState.DESKTOPKEYSTATE_DOWN);
            ttclient.SendDesktopInput(userid, ToTTKey(input));
        }

        private void DesktopDlg_KeyUp(object sender, KeyEventArgs e)
        {
            DesktopInput input = ToTTKeyCode(e, DesktopKeyState.DESKTOPKEYSTATE_UP);
            ttclient.SendDesktopInput(userid, ToTTKey(input));
        }
    }
}